unit module_matrix;

interface

uses

  System.Math, System.SysUtils, module_defs,

  KnobsUtils, KnobsConversions;

type

  EMatrix                = class( Exception);
  EMatrixDimError        = class( EMatrix  );
  EMatrixIndex           = class( EMatrix  );
  EMatrixNotSquare       = class( EMatrix  );
  EDeterminantZero       = class( EMatrix  );
  EMatrixNotDiagonizable = class( EMatrix);

  TMatrixSizes = array[ 0 .. 1] of Word;

  TSignalMatrix = class
  private
    FSizes : TMatrixSizes;
    FData  : TSignalArray;
  public
    class function MakeSizes( aRows, aCols: Word): TMatrixSizes;
    class function MakeNullSizes: TMatrixSizes;
    class function MakeIdentity( aRows, aCols: Word): TSignalMatrix;
  private
    procedure   FixDataSize;
    function    GetRowCount: Word;
    procedure   SetRowCount( aValue: Word);
    function    GetColCount: Word;
    procedure   SetColCount( aValue: Word);
    function    GetRow( aRow: Word): TSignalMatrix;
    procedure   SetRow( aRow: Word; const aValue: TSignalMatrix);
    function    GetCol( aCol: Word): TSignalMatrix;
    procedure   SetCol( aCol: Word; const aValue: TSignalMatrix);
    function    GetElementCount: Cardinal;
    function    GetElement( aRow, aCol: Word): TSignal;
    procedure   SetElement( aRow, aCol: Word; aValue: TSignal);
  public
    constructor Create;                                                         overload;
    constructor Create( const aSizes: TMatrixSizes);                            overload;
    constructor Create( const aSizes: TMatrixSizes; const aData: TSignalArray); overload;
    constructor Create( const aMatrix: TSignalMatrix);                          overload;
    constructor Create( const aVector: TSignalArray);                           overload;
    procedure   Normalize;
    procedure   PopulateVector( var aVector: TSignalArray);
    procedure   Clear;
    procedure   Assign( const aSizes: TMatrixSizes; const aData: TSignalArray); overload;
    procedure   Assign( const aValue: TSignalMatrix);                           overload;
    function    CheckIndex( aRow, aCol: Word): Boolean;
    function    IsSquare: Boolean;
    function    HasSameDimensions         ( const aMatrix: TSignalMatrix): Boolean;
    function    HasTransposedDimensions   ( const aMatrix: TSignalMatrix): Boolean;
    procedure   AssertIndexValid( aRow, aCol: Word);
    procedure   AssertSquare;
    procedure   AssertSameDimensions      ( const aMatrix: TSignalMatrix);
    procedure   AssertTransposedDimensions( const aMatrix: TSignalMatrix);
    function    MaxDim: Word;
    function    MinDim: Word;
    procedure   Add( const aValue: TSignalMatrix);                              overload;
    procedure   Add( aValue: TSignal);                                          overload;
    procedure   Mult( const aValue: TSignalMatrix);                             overload;
    procedure   Mult( aValue: TSignal);                                         overload;
    procedure   Divide( const aValue: TSignalMatrix);                           overload;
    procedure   Divide( aValue: TSignal);                                       overload;
    procedure   Transpose;
    function    SubMatrix( aRow, aCol: Word): TSignalMatrix;
    function    Determinant: TSignal;
    procedure   Invert;
    function    IsEqual( const aValue: TSignalMatrix): Boolean;
    function    IsOrthogonal: Boolean;
    function    IsIDentity: Boolean;
    function    IsCol: Boolean;
    function    IsRow: Boolean;
    procedure   SwapElements( r, c: Word);
    procedure   SwapRows( aRow1, aRow2: Word);
    procedure   SwapCols( aCol1, aCol2: Word);
    procedure   Diagonalize;
    function    Trace: TSignal;
    function    InnerProduct( aValue: TSignalMatrix): TSignal;
  public
    property    RowCount                   : Word          read GetRowCount     write SetRowCount;
    property    ColCount                   : Word          read GetColCount     write SetColCount;
    property    Row[ aRow: Word]           : TSignalMatrix read GetRow          write SetRow;
    property    Col[ aCol: Word]           : TSignalMatrix read GetCol          write SetCol;
    property    ElementCount               : Cardinal      read GetElementCount;
    property    Element[ aRow, aCol: Word] : TSignal       read GetElement      write SetElement; default;
  end;





implementation



{ ========
  TSignalMatrix = class
  private
    FSizes      : TMatrixSizes;
    FData       : TSignalArray;
  public
    property    RowCount                   : Word          read GetRowCount     write SetRowCount;
    property    ColCount                   : Word          read GetColCount     write SetColCount;
    property    Row[ aRow: Word]           : TSignalMatrix read GetRow          write SetRow;
    property    Col[ aCol: Word]           : TSignalMatrix read GetCol          write SetCol;
    property    ElementCount               : Cardinal      read GetElementCount;
    property    Element[ aRow, aCol: Word] : TSignal       read GetElement      write SetElement; default;
  public
}

    class function TSignalMatrix.MakeSizes( aRows, aCols: Word): TMatrixSizes;
    begin
      Result[ 0] := aRows;
      Result[ 1] := aCols;
    end;

    class function TSignalMatrix.MakeNullSizes: TMatrixSizes;
    begin
      Result := MakeSizes( 0, 0);
    end;

    class function TSignalMatrix.MakeIdentity( aRows, aCols: Word): TSignalMatrix;
    var
      r : Integer;
      c : Integer;
    begin
      if aRows = aCols
      then begin
        Result := TSignalMatrix.Create( MakeSizes( aRows, aCols));
        with Result
        do begin
          for r := 0 to RowCount - 1
          do begin
            for c := 0 to ColCount - 1
            do begin
              if r = c
              then Element[ r, c] := 1
              else Element[ r, c] := 0;
            end;
          end;
        end;
      end
      else raise EMatrixNotSquare.Create( 'Non square matrix specified for MakeIdentity');
    end;

// private

    procedure   TSignalMatrix.FixDataSize;
    begin
      SetLength( FData, RowCount * ColCount);
    end;

    function    TSignalMatrix.GetRowCount: Word;
    begin
      Result := FSizes[ 0];
    end;

    procedure   TSignalMatrix.SetRowCount( aValue: Word);
    begin
      FSizes[ 0] := aValue;
      FixDataSize;
    end;

    function    TSignalMatrix.GetColCount: Word;
    begin
      Result := FSizes[ 1];
    end;

    procedure   TSignalMatrix.SetColCount( aValue: Word);
    begin
      FSizes[ 1] := aValue;
      FixDataSize;
    end;

    function    TSignalMatrix.GetRow( aRow: Word): TSignalMatrix;
    var
      c : Word;
    begin
      Result := TSignalMatrix.Create( MakeSizes( 1, ColCount));
      for c := 0 to ColCount - 1
      do Result[ 0, c] := Element[ aRow, c];
    end;

    procedure   TSignalMatrix.SetRow( aRow: Word; const aValue: TSignalMatrix);
    var
      c : Word;
    begin
      if Assigned( aValue) and aValue.IsRow and ( aValue.ColCount = ColCount)
      then begin
        for c := 0 to ColCount - 1
        do Element[ aRow, c] := aValue[ 0, c];
      end
      else raise EMatrixDimError.Create( 'aValue is not a row, or column count mismatch');
    end;

    function    TSignalMatrix.GetCol( aCol: Word): TSignalMatrix;
    var
      r : Word;
    begin
      Result := TSignalMatrix.Create( MakeSizes( RowCount, 1));
      for r := 0 to RowCount - 1
      do Result[ r, 0] := Element[ r, aCol];
    end;

    procedure   TSignalMatrix.SetCol( aCol: Word; const aValue: TSignalMatrix);
    var
      r : Word;
    begin
      if Assigned( aValue) and aValue.IsCol and ( aValue.RowCount = RowCount)
      then begin
        for r := 0 to RowCount - 1
        do Element[ r, aCol] := aValue[ r, 0];
      end
      else raise EMatrixDimError.Create( 'aValue is not a col, or row count mismatch');
    end;

    function    TSignalMatrix.GetElementCount: Cardinal;
    begin
      Result := Length( FData);
      Assert( Result = RowCount * ColCount, 'inconsistancy in element count');
    end;

    function    TSignalMatrix.GetElement( aRow, aCol: Word): TSignal;
    begin
      AssertIndexValid( aRow, aCol);
      Result := FData[ aRow * ColCount + aCol]
    end;

    procedure   TSignalMatrix.SetElement( aRow, aCol: Word; aValue: TSignal);
    begin
      AssertIndexValid( aRow, aCol);
      FData[ aRow * ColCount + aCol] := aValue;
    end;

//  public

    constructor TSignalMatrix.Create; // overload;
    begin
      Create( MakeNullSizes)
    end;

    constructor TSignalMatrix.Create( const aSizes: TMatrixSizes); // overload;
    begin
      inherited Create;
      RowCount := aSizes[ 0];
      ColCount := aSizes[ 1];
    end;

    constructor TSignalMatrix.Create( const aSizes: TMatrixSizes; const aData: TSignalArray); // overload;
    var
      i       : Integer;
      aLength : Integer;
    begin
      Create( aSizes);
      if Assigned( aData)
      then aLength := Min( Length( aData), ElementCount)
      else aLength := 0;
      for i := 0 to aLength - 1
      do FData[ i] := aData[ i];
    end;

    constructor TSignalMatrix.Create( const aMatrix: TSignalMatrix); // overload;
    begin
      if Assigned( aMatrix)
      then Create( aMatrix.FSizes, aMatrix.FData)
      else Create( MakeNullSizes, nil);
    end;

    constructor TSignalMatrix.Create( const aVector: TSignalArray); // overload;
    var
      r : Integer;
    begin
      Create( MakeSizes( Length( aVector), 0));
      for r := 0 to RowCount - 1
      do Element[ r, 0] := aVector[ r];
    end;

    procedure   TSignalMatrix.Normalize;
    var
      i : Integer;
    begin
      for i := 0 to ElementCount - 1
      do PNormalize( FData[ i]);
    end;

    procedure   TSignalMatrix.PopulateVector( var aVector: TSignalArray);
    var
      r : Integer;
    begin
      Assert(( ColCount = 1) and ( RowCount = Length( aVector)));
      for r := 0 to RowCount - 1
      do aVector[ r] := Element[ r, 0];
    end;

    procedure   TSignalMatrix.Clear;
    var
      i : Integer;
    begin
      for i := 0 to ElementCount - 1
      do FData[ i] := 0;
    end;

    procedure   TSignalMatrix.Assign( const aSizes: TMatrixSizes; const aData: TSignalArray); // overload;
    var
      i : Integer;
    begin
      ColCount := aSizes[ 0];
      RowCount := aSizes[ 1];
      for i := 0 to ElementCount - 1
      do FData[ i] := aData[ i];
    end;

    procedure   TSignalMatrix.Assign( const aValue: TSignalMatrix); // overload;
    begin
      if Assigned( aValue)
      then Assign( aValue.FSizes, aValue.FData)
      else Clear;
    end;

    function    TSignalMatrix.CheckIndex( aRow, aCol: Word): Boolean;
    begin
      Result := ( aRow < RowCount) and ( aCol < ColCount);
    end;

    function    TSignalMatrix.IsSquare: Boolean;
    begin
      Result := RowCount = ColCount;
    end;

    function    TSignalMatrix.HasSameDimensions( const aMatrix: TSignalMatrix): Boolean;
    begin
      Result := Assigned( aMatrix) and ( aMatrix.RowCount = RowCount) and ( aMatrix.ColCount = ColCount);
    end;

    function    TSignalMatrix.HasTransposedDimensions( const aMatrix: TSignalMatrix): Boolean;
    begin
      Result := Assigned( aMatrix) and ( aMatrix.RowCount = ColCount) and ( aMatrix.ColCount = RowCount);
    end;

    procedure   TSignalMatrix.AssertIndexValid( aRow, aCol: Word);
    begin
      if not CheckIndex( aRow, aCol)
      then raise EMatrixIndex.Create( 'Index out of bounds');
    end;

    procedure   TSignalMatrix.AssertSquare;
    begin
      if not IsSquare
      then raise EMatrixNotSquare.Create( 'Matrix is not square');
    end;

    procedure   TSignalMatrix.AssertSameDimensions( const aMatrix: TSignalMatrix);
    begin
      if not HasSameDimensions( aMatrix)
      then raise EMatrixDimError.Create( 'dimensions differ');
    end;

    procedure   TSignalMatrix.AssertTransposedDimensions( const aMatrix: TSignalMatrix);
    begin
      if not HasTransposedDimensions( aMatrix)
      then raise EMatrixDimError.Create( 'dimensions not mirrored');
    end;

    function    TSignalMatrix.MaxDim: Word;
    begin
      Result := Max( RowCount, ColCount);
    end;

    function    TSignalMatrix.MinDim: Word;
    begin
      Result := Min( RowCount, ColCount);
    end;

    procedure   TSignalMatrix.Add( const aValue: TSignalMatrix);
    var
      i : Integer;
    begin
      AssertSameDimensions( aValue);
      for i := 0 to ElementCount - 1
      do FData[ i] := FData[ i] + aValue.FData[ i];
    end;

    procedure   TSignalMatrix.Add( aValue: TSignal); // overload;
    var
      i : Integer;
    begin
      for i := 0 to ElementCount - 1
      do FData[ i] := FData[ i] + aValue;
    end;

    procedure   TSignalMatrix.Mult( const aValue: TSignalMatrix); //                              overload;
    var
      i    : Integer;
      j    : Integer;
      k    : Integer;
      aTmp : TSignalMatrix;
      aSum : TSignal;
    begin
      AssertTransposedDimensions( aValue);
      aTmp := TSignalMatrix.Create( MakeSizes( aValue.MaxDim, MaxDim));
      for i := 0 to RowCount - 1
      do begin
        aSum := 0.0;
        for j := 0 to ColCount - 1
        do begin
          for k := 0 to aValue.ColCount - 1
          do begin
            aSum := aSum + Element[ k, j] * aValue[ i, k];
          end;
          aTmp[ i, j] := aSum;
        end;
      end;
      Assign( aTmp);
    end;

    procedure   TSignalMatrix.Mult( aValue: TSignal); // overload;
    var
      i : Integer;
    begin
      for i := 0 to ElementCount - 1
      do FData[ i] := FData[ i] * aValue;
    end;

    procedure   TSignalMatrix.Divide( const aValue: TSignalMatrix); // overload;
    var
      aTmp: TSignalMatrix;
    begin
      if Assigned( aValue)
      then begin
        aTmp := TSignalMatrix.Create( aValue);
        try
          aTmp.Invert;
          Mult( aTmp);
        finally
          aTmp.Free;
        end;
      end;
    end;

    procedure   TSignalMatrix.Divide( aValue: TSignal); // overload;
    begin
      Mult( 1 / aValue);
    end;

    procedure   TSignalMatrix.Transpose;
    var
      aTmp : TSignalMatrix;
      r    : Word;
      c    : Word;
    begin
      aTmp := TSignalMatrix.Create( MakeSizes( ColCount, RowCount));
      for r := 0 to RowCount - 1
      do begin
        for c := 0 to ColCount - 1
        do aTmp[ c, r] := Element[ r, c];
      end;
      Assign( aTmp);
    end;

    function    TSignalMatrix.SubMatrix( aRow, aCol: Word): TSignalMatrix;
    // Returns a new sub matrix by eliminating row aRow and col aCol from self
    var
      r : Word;
      c : Word;
    begin
      AssertIndexValid( aRow, aCol);
      Result := TSignalMatrix.Create( MakeSizes( RowCount - 1, ColCount - 1));
      for r := 0 to RowCount - 1
      do begin
        if r < aRow
        then begin
          for c := 0 to ColCount - 1
          do begin
            if c < aCol
            then Result[ r, c    ] := Element[ r, c]
            else if c > aCol
            then Result[ r, c - 1] := Element[ r, c];
          end
        end
        else if r > aRow
        then begin
          for c := 0 to ColCount - 1
          do begin
            if c < aCol
            then Result[ r - 1, c    ] := Element[ r, c]
            else if c > aCol
            then Result[ r - 1, c - 1] := Element[ r, c];
          end
        end
      end;
    end;

    function    TSignalMatrix.Determinant: TSignal;
    var
      i          : Integer;
      aSign      : TSignal;
      aSubMatrix : TSignalMatrix;
    begin
      AssertSquare;
      if ColCount = 0
      then Result := 1
      else begin
        Result := 0;
        aSign   := 1;
        for i := 0 to ColCount - 1
        do begin
          aSubMatrix := SubMatrix( 0, i);
          try
            Result := Result + Element[ 0, i] * aSign * aSubMatrix.Determinant;
            aSign  := - aSign;
          finally
            aSubMatrix.Free;
          end;
        end;
      end;
    end;

    procedure   TSignalMatrix.Invert;
    var
      D     : TSignal;
      aTmp  : TSignalMatrix;
      aSub  : TSignalMatrix;
      r     : Integer;
      c     : Integer;
      aSign : TSignal;
    begin
      AssertSquare;
      D := Determinant;
      if D <> 0
      then begin
        aTmp  := TSignalMatrix.Create( MakeSizes( RowCount, ColCount));
        aSign := 1;
        for r := 0 to RowCount - 1
        do begin
          for c := 0 to ColCount - 1
          do begin
            aSub := SubMatrix( r, c);
            try
              aTmp[ r, c] := aSign * aSub.Determinant / D;
              aSign       := - aSign;
            finally
              aSub.Free;
            end;
          end;
        end;
      end
      else raise EDeterminantZero.Create( 'Determinant is zero');
    end;

    function    TSignalMatrix.IsEqual( const aValue: TSignalMatrix): Boolean;
    var
      i : Integer;
    begin
      Result := Assigned( aValue);
      if Result
      then begin
        Result := HasSameDimensions( aValue);
        if Result
        then begin
          for i := 0 to ElementCount - 1
          do begin
            if FData[ i] <> aValue.FData[ i]
            then begin
              Result := False;
              Break;
            end;
          end;
        end;
      end;
    end;

    function    TSignalMatrix.IsOrthogonal: Boolean;
    var
      aTransposed : TSignalMatrix;
      anInverted  : TSignalMatrix;
    begin
      AssertSquare;
      aTransposed := TSignalMatrix.Create( Self);
      try
        aTransposed.Transpose;
        anInverted := TSignalMatrix.Create( Self);
        try
          anInverted.Invert;
          Result := anInverted.IsEqual( aTransposed)
        finally
          anInverted.Free;
        end;
      finally
        aTransposed.Free;
      end;
    end;

    function    TSignalMatrix.IsCol: Boolean;
    begin
      Result := ColCount = 1;
    end;

    function    TSignalMatrix.IsRow: Boolean;
    begin
      Result := RowCount = 1;
    end;

    function    TSignalMatrix.IsIDentity: Boolean;
    var
      r : Word;
      c : Word;
    begin
      AssertSquare;
      Result := True;
      for r := 0 to RowCount - 1
      do begin
        for c := 0 to ColCount - 1
        do begin
          if r = c
          then begin
            if Element[ r, c] <> 1
            then begin
              Result := False;
              Break;
            end
          end
          else begin
            if Element[ r, c] <> 0
            then begin
              Result := False;
              Break;
            end
          end;
        end;
      end;
    end;

    procedure   TSignalMatrix.SwapElements( r, c: Word);
    var
      aTmp : TSignal;
    begin
      AssertIndexValid( r, c);
      AssertIndexValid( c, r);
      aTmp           := Element[ r, c];
      Element[ r, c] := Element[ c, r];
      Element[ c, r] := aTmp;
    end;

    procedure   TSignalMatrix.SwapRows( aRow1, aRow2: Word);
    var
      aTmp : TSignal;
      c    : Word;
    begin
      AssertIndexValid( aRow1, 0);
      AssertIndexValid( aRow2, 0);
      for c := 0 to ColCount - 1
      do begin
        aTmp               := Element[ aRow1, c];
        Element[ aRow1, c] := Element[ aRow2, c];
        Element[ aRow2, c] := aTmp;
      end;
    end;

    procedure   TSignalMatrix.SwapCols( aCol1, aCol2: Word);
    var
      aTmp : TSignal;
      r    : Word;
    begin
      AssertIndexValid( 0, aCol1);
      AssertIndexValid( 0, aCol2);
      for r := 0 to RowCount - 1
      do begin
        aTmp               := Element[ r, aCol1];
        Element[ r, aCol1] := Element[ r, aCol2];
        Element[ r, aCol2] := aTmp;
      end;
    end;

    procedure   TSignalMatrix.Diagonalize;
    var
      r    : Word;
      c    : Word;
      aTmp : TSignal;
    begin
      AssertSquare;
      if Element[ 0, 0] <> 0
      then begin
        for c := 1 to ColCount - 1
        do begin
          aTmp := Element[ 0, c] / Element[ 0, 0];
          for r := 0 to RowCount - 1
          do Element[ r, c] := Element[ r, c] + aTmp * Element[ r, 0];
        end;
        for r := 1 to RowCount - 1
        do Element[ r, 0] := 0;
      end
      else raise EMatrixNotDiagonizable.Create( 'Not diagonizable');
    end;

    function    TSignalMatrix.Trace: TSignal;
    var
      i : Word;
    begin
      AssertSquare;
      Result := 0;
      for i := 0 to ColCount - 1
      do Result := Result + Element[ i, i];
    end;

    function    TSignalMatrix.InnerProduct( aValue: TSignalMatrix): TSignal;
    var
      i : Integer;
    begin
      Result := 0;
      AssertSameDimensions( aValue);
      if IsCol and aValue.IsCol
      then begin
        for i := 0 to RowCount - 1
        do Result := Result + Element[ 0, i] * aValue[ 0, i];
      end
      else if isRow and aValue.IsRow
      then begin
        for i := 0 to ColCount - 1
        do Result := Result + Element[ i, 0] * aValue[ i, 0];
      end
      else raise EMatrixDimError.Create( 'Inner product needs two vectors')
    end;




end.

