unit KnobsForth;

{

   COPYRIGHT 1999 .. 2014 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.


  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

  Jittik is in floeistof dy't foar it grutste part t jittiksoer (CH3CO2H) en
  wetter bestiet; it jittiksoer nstiet troch fermintaasje fan alkohol troch
  jittiksoerbaktearjen (Acetobacter). [http://fy.wikipedia.org/wiki/Jittik]

  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

  KnobsForth is a Forth like threaded intepretative language, there are a
  couple of differences .. some items:

  - There is no user interface, KnobsForth is embeddable and as such the
    program using it can send text to it which then will be interpreted. It is
    also possible to throw a text file at it to be interpreted. All response
    will be in the form of the Say() callback, or a possible exception thrown.

  - The system is reentrant, while text is being interpreted new text can be
    sent to it. This will not be executed in parallel, but when the new program
    finishes the old one will continue. To have multiple threads execute within
    the same engine will need more work.

  - Stack items need not be numbers, they can be strings or objects too, in
    fact they often will not be numbers. There can even be stacks on stacks,
    which seems useful for continuations. For most simple types there will be
    proper overloading, like two strings can be added like numbers (although it
    may be needed at times to explicitly convert a number to a string with ($).

  - There is some limited support for lists. lists are not passive objects, but
    they can be executed. There is some conversion stuff for stacks into lists
    and the other way around.

  - System objects are reference counted, when they go out of scope they will
    be deleted. There are restrictions on recursion. The simple case of a colon
    definition being self recursive is handled. Recursive lists or stacks will
    most likely fail to recover all memory.

  - the create does> system works different than usually. The execution of the
    does> bit gets an index into code space, and in that space objects will be
    compiled - which need an extra dereference in order to access them. For
    this the word pfa@ wwas defined. As an example the basics for defining a
    complex number extension:

      : i@ ( pfa -- re im ) dup >r 1 + pfa@ @ r> pfa@ @ ;          // fetch a complex pair
      : i! ( re im pfa -- ) dup >r pfa@ ! r> 1 + pfa@ ! ;          // store a complex pair
      : i. ( re im -- ) ($) " i" + swap ($) "  + " + swap + . ;    // print a complex pair like 're + iim'
      : iconstant ( re im -- ) create , , does> i@ ;               // complex constant
      : ivariable ( -- ) create 2 allot does> ;                    // complex variable
      : i+ ( re1 im1 re2 im2 -- re im ) swap >r + swap r> + swap ; // complex addition

    i* is a bit more involved ... like ...

      : i* ( re1 im1 re2 im2 -- re im )

        3 pick >r
        1 pick r> * >r       // ac
        2 pick >r
        0 pick r> * r>       // bd
        swap -               // ac - bd
        3 pick >r
        0 pick >r * >r       // ad
        2 pick >r
        1 pick r> * r>       // bc
        + >r                 // ad + bc
        drop drop drop drop  // clean up
        r> r> swap           // ac-bd ad+bc
      ;

    something like that prolly, not tested .. division is a more fun even:

      (a+bi)/(c+di) = ((ac+bd)+(bc-ad)i)/(aa+cc)

    :-)

    anyways, create does> behaves a lil odd, and complex numbers could use some
    built-in support I guess.

  -
}


interface

uses

  WinApi.Windows, WinApi.ShellApi, System.Types, System.SysUtils, System.DateUtils, System.Classes, System.Contnrs,
  System.Variants, System.Math, Vcl.Forms,

  KnobsUtils, KnobsConversions, KnobsFiles;

const

  NoName       = '';
  NoNumber     = $deadbeefbabeface;
  Nothing      = '<nothing-50E13785-E2DF-4ADB-BE9F-2931DB2C4809>'; // magic cookie

  CBeginUntilWhile =  1;
  CWhileRepeat     =  2;
  CDoOd            =  3;
  CIfElseFi        =  4;
  CColonSemi       =  5;
  CCaseEsac        =  6;
  COfFo            =  7;
  CInNi            =  8;
  CPastLast        =  9;

  ControlNames : array[ CBeginUntilWhile .. CPastLast - 1] of string = (
    '''until'' or ''while''',
    '''repeat''',
    '''od'' or ''+od''',
    '''else'', ''fi''',
    ''';''',
    '''esac''',
    '''fo''',
    '''ni'''
  );
  ControlIntros : array[ CBeginUntilWhile .. CPastlast - 1] of string = (
    '''begin''',
    '(''begin'' ..) ''while''',
    '''do'' or ''?do''',
    '''if'' or ''else''',
    ''':'' (colon)',
    '''case''',
    '''of''',
    '''in'''
  );



type

  TCFA      = Integer;
  TVariants = array of Variant;

  TOnForthSay       = procedure( aSender: TObject; const aMsg: string)                              of object;
  TOnBye            = procedure( aSender: TObject)                                                  of object;
  TOnForthBreak     = procedure( aSender: TObject; const aMsg: string)                              of object;
  TOnForthLog       = procedure( aSender: TObject; const aMsg: string)                              of object;
  TOnStackUnderflow = procedure( aSender: TObject; const aMsg: string)                              of object;
  TOnUnknownSymbol  = function ( aSender: TObject; const aToken: string): Boolean                   of object;
  TOnDataOut        = procedure( aSender: TObject; const aChannel: string; const aData: TVariants)  of object;
  TOnAddScale       = procedure( aSender: TObject; const aName: string; const aScale: TMidiFilter)  of object;
  TOnGetScaleNames  = function ( aSender: TObject): string                                          of object;
  TOnClearScales    = procedure( aSender: TObject)                                                  of object;


  EKnobsForth      = class( Exception  );
  EMemoryViolation = class( EKnobsForth);
  ECrash           = class( EKnobsForth);
  TSymbol          = class;
  TSymbolClass     = class of TSymbol;
  TForthList       = class;
  TVocabulary      = class;
  TKnobsForth      = class;
  TInternalProc    = procedure( aSystem: TKnobsForth);

  TSymbol = class
  private
    FSystem    : TKnobsForth;
    FRefCount  : Cardinal;
    FName      : string;       // Name of the symbol (can be '')
    FComment   : string;       // Comment for symbol, probably a stack image
    FDocs      : string;       // Additional user generated comment for the symbol
    FPrevious  : TSymbol;      // link to previously defined symbol
    FVocLink   : TVocabulary;  // Link to vocabulary this symbol is in
    FCrash     : Boolean;      // True when execution is forbidden - rases an exception
    FImmediate : Boolean;      // True when executing in compilation mode
    FVisible   : Boolean;      // True when symbol is findable
    FSysSymbol : Boolean;      // True when this is a system symbol
    FIsSemi    : Boolean;      // True when this symbol ends a definition e.g. ';' (and possible friends)
  public
    constructor Create( aSystem: TKnobsForth; const aName: string = NoName);                                    virtual;
    destructor  Destroy;                                                                                       override;
    procedure   AddRef;                                                                                         virtual;
    procedure   ReleaseRef;                                                                                     virtual;
    procedure   EndCreate;                                                                                      virtual;
    procedure   CheckCrash;
    procedure   Interpret;                                                                                      virtual;
    procedure   Hide;                                                                                           virtual;
    procedure   Reveal;                                                                                         virtual;
    function    See: string;                                                                                    virtual;
    function    AppendFlags( const S: string): string;                                                          virtual;
    procedure   AddDocs( const aValue: string);                                                                 virtual;
    procedure   Dump       ( const aStringList: TStringList; anInd: Cardinal);                                  virtual;
    procedure   DumpLimited( const aStringList: TStringList; anInd: Cardinal);                                  virtual;
    procedure   DumpDetails( const aStringList: TStringList; anInd: Cardinal);                                  virtual;
    function    AsList: TForthList;                                                                             virtual;
  public
    class function TypeName: string;
  public
    property    RefCount  : Cardinal read FRefCount;
    property    Name      : string   read FName;
    property    Comment   : string   read FComment   write FComment;
    property    Docs      : string   read FDocs      write FDocs;
    property    Crash     : Boolean  read FCrash     write FCrash;
    property    Immediate : Boolean  read FImmediate write FImmediate;
    property    Visible   : Boolean  read FVisible   write FVisible;
    property    SysSymbol : Boolean  read FSysSymbol write FSysSymbol;
    property    IsSemi    : Boolean  read FIsSemi;
  end;

  TSymbols = class( TList)
  private
    FIndex  : TStringList;
    FSystem : TKnobsForth;
  private
    function    GetSymbol( anIndex: TCFA): TSymbol;
  public
    constructor Create( aSystem: TKnobsForth);
    destructor  Destroy;                                                                                       override;
    procedure   Clear;                                                                                         override;
    function    AddSymbol( const aName: string; aSymbolClass: TSymbolClass; const aVoc: TVocabulary): TSymbol; overload;
    function    AddSymbol( aSymbol: TSymbol; const aVoc: TVocabulary): TSymbol;                                overload;
    function    RemoveSymbol( const aName: string): TSymbol;
    function    FindToken( const aToken: string): TSymbol;
    function    CfaToSymbol( aCFA: TCFA): TSymbol;
    function    Symbols: string;
    procedure   Dump( const aStringList: TStringList; anInd: Cardinal);
  public
    property    Symbol[ anIndex: TCFA]: TSymbol read GetSymbol; default;
  end;

  TStack = class( TSymbol)
  private
    FData        : array of Variant;
    FOnUnderflow : TOnStackUnderflow;
  private
    function    GetDepth: Cardinal;
  public
    destructor  Destroy;                                                                                       override;
    procedure   Push( aValue: Variant);
    procedure   PushSymbol( const aValue: TSymbol);
    function    Pop: Variant;
    function    PopSymbol: TSymbol;
    function    Index( anIndex: Cardinal): Variant;
    procedure   Put( anIndex: Cardinal; aValue: Variant);
    function    Fetch: Variant;
    procedure   Swap;
    procedure   Clear;
    procedure   Dump( const aStringList: TStringList; anInd: Cardinal);                                        override;
    function    AsList: TForthList;                                                                            override;
  public
    property    Depth       : Cardinal          read GetDepth;
    property    OnUnderflow : TOnStackUnderflow read FOnUnderflow write FOnUnderflow;
  end;

  TForthList = class( TSymbol)
  strict private
    FValue    : TList;
  strict private
    function    GetCount: Integer;
    procedure   SetCount( aValue: Integer);
    function    GetSymbol( anIndex: TCFA): TSymbol;
  public
    constructor Create( aSystem: TKnobsForth; const aName: string = NoName);                                   override;
    destructor  Destroy;                                                                                       override;
    procedure   ReleaseRef;                                                                                    override;
    function    Copy: TForthList;
    procedure   Clear;
    procedure   Interpret;                                                                                     override;
    procedure   Call;
    function    See: string;                                                                                   override;
    procedure   Append( aValue: TSymbol);
    procedure   AcceptToken( const aToken: string);
    procedure   CompileVariant( const aValue: Variant);
    procedure   CompileLiteral( const aValue: Variant);
    procedure   CompileSymbol( const aSymbol: TSymbol);
    procedure   CompileToken( const aName: string);                                                            overload;
    procedure   CompileToken( const aName: string; const aSymbolClass: TSymbolClass);                          overload;
    function    Here: TCFA;
    procedure   SetHere( aValue: TCFA);
    procedure   Reverse;
    function    First: TSymbol;
    procedure   Rest;
    procedure   Concat( const aValue: TForthList);
    procedure   Dump( const aStringList: TStringList; anInd: Cardinal);                                        override;
    function    AsList: TForthList;                                                                            override;
  public
    property    Count                  : Integer read GetCount  write SetCount;
    property    Symbol[ anIndex: TCFA] : TSymbol read GetSymbol;
  end;

  TScale = class( TSymbol)
  private
    FScaleData: TMidiFilter;
  protected
    function    AsString: string;
  public
    constructor Create( aSystem: TKnobsForth; const aName: string = NoName);                                   override;
    procedure   Interpret;                                                                                     override;
    function    See: string;                                                                                   override;
    procedure   AddItem( anItem: TMidiSeptet);
    procedure   AddList( const aValue: Variant);
  public
    property ScaleData: TMidiFilter read FScaleData write FScaleData;
  end;

  TLiteral = class( TSymbol)
  private
    FValue : Variant;
  private
    procedure   SetValue( const aValue: Variant);
  public
    procedure   Interpret;                                                                                     override;
    function    See: string;                                                                                   override;
    procedure   DumpDetails( const aStringList: TStringList; anInd: Cardinal);                                 override;
    function    AsList: TForthList;                                                                            override;
  public
    property    Value: Variant read FValue write SetValue;
  end;

  TConstant = class( TLiteral)
  public
    procedure   Interpret;                                                                                     override;
    function    See: string;                                                                                   override;
  end;

  TVariable = class( TConstant)
  public
    procedure   Interpret;                                                                                     override;
    function    See: string;                                                                                   override;
  end;

  TVocabulary = class( TSymbol)
  private
    FSymbols : TSymbols; // Link to a symbol list
  public
    constructor Create( aSystem: TKnobsForth; const aName: string = NoName);                                   override;
    destructor  Destroy;                                                                                       override;
    procedure   Interpret;                                                                                     override;
  public
    function    AddSymbol( const aName: string; aSymbolClass: TSymbolClass): TSymbol;                          overload;
    function    AddSymbol( aSymbol: TSymbol): TSymbol;                                                         overload;
    function    RemoveSymbol( const aName: string): TSymbol;
    function    FindToken( const aToken: string): TSymbol;
    function    CfaToSymbol( aCFA: TCFA): TSymbol;
    function    SymbolList: string;
    procedure   DumpDetails( const aStringList: TStringList; anInd: Cardinal);                                 override;
  public
    property    Symbols : TSymbols read FSymbols;
  end;


  TAlias = class( TSymbol)
  private
    FAliased : TSymbol;
  public
    procedure   Interpret;                                                                                     override;
    function    See: string;                                                                                   override;
    procedure   DumpDetails( const aStringList: TStringList; anInd: Cardinal);                                 override;
  public
    property    Aliased: TSymbol read FAliased write FAliased;
  end;

  TDefer = class( TAlias)
  public
    procedure   Resolve( aSymbol: TSymbol);
    function    See: string;                                                                                   override;
    procedure   DumpDetails( const aStringList: TStringList; anInd: Cardinal);                                 override;
  end;

  TExecutable = class( TSymbol)
  end;


  TInternal = class( TExecutable)
  private
    FInternalProc: TInternalProc;
  public
    procedure   Interpret;                                                                                     override;
  public
    property    InternalProc: TInternalProc read FInternalProc write FInternalProc;
  end;

  TColon = class( TExecutable)
  public
    FCFA : TCFA;     // Location in code space
  public
    destructor  Destroy;                                                                                       override;
    procedure   ReleaseRef;                                                                                    override;
    procedure   Interpret;                                                                                     override;
    function    See: string;                                                                                   override;
    procedure   DumpDetails ( const aStringList: TStringList; anInd: Cardinal);                                override;
    function    AsList: TForthList;                                                                            override;
    procedure   DumpLimited( const aStringList: TStringList; anInd: Cardinal);                                 override;
  public
    property    CFA : TCFA read FCFA write FCFA;
  end;

  TCreate = class( TColon)
  public
    procedure   Interpret;                                                                                     override;
  end;

  TKnobsForth = class( TComponent)
  private
    FOnSay            : TOnForthSay;
    FOnBye            : TOnBye;
    FOnBreak          : TOnForthBreak;
    FOnLog            : TOnForthLog;
    FOnReset          : TNotifyEvent;
    FOnUnknownSymbol  : TOnUnknownSymbol;
    FOnDataOut        : TOnDataOut;
    FOnAddScale       : TOnAddScale;
    FOnGetScalenames  : TOnGetScaleNames;
    FOnClearScales    : TOnClearScales;
    FOnClearScreen    : TNotifyEvent;
  private
    FOnLine           : TOnLine;              // Callback for line counter        - must be set by user App.
    FOnFileEnter      : TOnFileEnter;         // Callback for file counter        - must be set by user App.
    FOnFileLeave      : TOnFileLeave;         // Callback for file counter        - must be set by user App.
    FOnFileResume     : TOnFileResume;        // Callback for file counter        - must be set by user App.
  private
    FVocs             : array of TVocabulary; // List of all vocabularies
    FOnly             : TVocabulary;          // Startup symbol list Selected with only (pseudo Voc)
    FCurrent          : TVocabulary;          // Compile voc
    FContext          : array of TVocabulary; // Search voc (chain)
    FDataStack        : TStack;               // Stack for data transfer, mostly just called 'the stack'
    FReturnStack      : TStack;               // Stack used for keeping track of proper nesting of symbol invocations
    FControlStack     : TStack;               // Stack used for keeping track of proper nesting of control structures
    FCode             : TForthList;           // Compiled forth code
    FAnonymous        : TObjectList;          // List of all anonymous symbolss (so they can be freed)
    FFileStack        : TFileStack;           // Stack of source files
    FFiles            : array of string;
  private
    FCompiling        : Cardinal;             // We are compiling    when FCompiling    > 0
    FInterpreting     : Cardinal;             // We are interpreting when FInterpreting > 0 :: this is text interpreting
    FYieldCounter     : Integer;
    FIP               : TCFA;                 // Forth instruction pointer
    FW                : TSymbol;              // Forth code symbol - this is the symbol pointed to by FIP
    FLastDefined      : TSymbol;              // The last defined symbol - only valid after creation completed
    FCurrentColon     : TColon;               // The colon definition currently under construction - nil if no colon symbol is being defined
  private
    FStartupFile      : string;               // name for library code to load at reset
  private
    function    GetBase: Cardinal;
    procedure   SetBase( aValue: Cardinal);
    function    GetCompiling: Boolean;
    procedure   SetCompiling( aValue: Boolean);
    function    GetInterpreting: Boolean;
    procedure   SetInterpreting( aValue: Boolean);
  private
    procedure   Log( const aMsg: string);
    procedure   RaiseBreak( const aMsg: string);
    procedure   RaiseError( const aMsg: string);
    procedure   RaiseErrorFmt( const aFmt: string; const anArgs: array of const);
    procedure   ShowWarning( const aMsg: string);
    procedure   ShowWarningFmt( const aFmt: string; const anArgs: array of const);
    procedure   Say( const S: string);
    procedure   SayFmt( const aFmt: string; const anArgs: array of const);
    procedure   Yield;
    procedure   DoStackUnderflow( aSender: TObject; const aMsg: string);
    function    ReturnStackTrace: string;
    function    GetToken  ( aSeparator: Char; Extended: Boolean = False): string;
    function    FetchToken( aSeparator: Char; Extended: Boolean = False): string;
    procedure   PushLine( const aLine: string; const anID: string);
    procedure   PushLines( const aLines: TStrings; const anID: string);
    procedure   PushFile( const aFileName: string);
    procedure   SkipToEol;
    function    GetToEol: string;
    function    ReadLine: string;
    procedure   FinalizeCreation( const aSymbol: TSymbol; const aComment: string = '');
    function    Number               ( const aToken: string; out aNumber: Variant): Boolean;
    function    Float                ( const aToken: string; out aNumber: Variant): Boolean;
    function    MustFindToken        ( const aToken: string; aSymbolClass: TSymbolClass): TSymbol;
    function    MustFindSymbol       ( const aToken: string): TSymbol;
    procedure   MustBeCompilingFor   ( const aToken: string);
    procedure   MustNotBeCompilingFor( const aToken: string);
    function    FindToken( const aToken: string): TSymbol;
    procedure   Run;
    procedure   AcceptToken( const aToken: string);
    function    HandleUnknowSymbol( const aToken: string): Boolean;
    procedure   DataOut( const aChannel: string; const aData: TVariants);
    procedure   AddScale( const aName: string; const aScale: TMidiFilter);
    function    GetScaleNames: string;
    procedure   Explain( aSymbol: TSymbol);
    procedure   Call( aList: TForthList; anIP: TCFA);
    procedure   AddAlias( const anOldName, aNewName: string);
    procedure   CompileLiteral( const aValue: Variant);
    procedure   CompileSymbol( const aSymbol: TSymbol);
    procedure   CompileToken( const aName: string);                                                            overload;
    procedure   CompileToken( const aName: string; const aSymbolClass: TSymbolClass);                          overload;
    function    Here: TCFA;
    procedure   SetHere( aValue: TCFA);
    procedure   DataPush( aValue: Variant);
    procedure   DataPushSymbol( const aValue: TSymbol);
    function    DataPop: Variant;
    function    DataFetch: Variant;
    function    DataIndex( anIndex: Integer): Variant;
    procedure   DataPut( anIndex: Integer; aValue: Variant);
    function    DataPopSymbol: TSymbol;
    procedure   DataSwap;
    procedure   ReturnPush( aValue: Variant);
    function    ReturnPop: Variant;
    procedure   ControlPush( aValue: Variant);
    function    ControlPop: Variant;
    function    AddSymbol( const aName: string; aSymbolClass: TSymbolClass): TSymbol;
    procedure   Internal( const aName: string; aProc: TInternalProc; const aComments: string);
    procedure   SetupInternals;
    procedure   Crash;
    procedure   Immediate;
    procedure   Hide;
    procedure   Reveal;
    procedure   SysSymbol;
    procedure   ForgetAll;
    procedure   CheckControlStructure( aSeen: Cardinal);
    procedure   CheckDataStack   ( const aMsg: string);
    procedure   CheckControlStack( const aMsg: string);
    procedure   CheckReturnStack ( const aMsg: string);
    function    Anonymous: string;
    function    VocDump : string;
    function    CodeDump: string;
    procedure   AddFile( const aFileName: string);
    procedure   ClearFiles;
    function    Files: TForthList;
    function    IsIncluded( const aFileName: string): Boolean;
    procedure   Required( const aFileName: string);
    function    CfaToSymbol( aCFA: TCFA): TSymbol;
    procedure   Bye;
 private
    function    GetSourceName  : string;
    function    GetSourceLineNr: Integer;
    function    GetSourceOffset: Integer;
    function    GetContext: TVocabulary;
    procedure   SetContext( aValue: TVocabulary);
    function    AddVoc( aVoc: TVocabulary): TVocabulary;
    procedure   FSave( const aContents, aFileName: string);
    function    FLoad( const aFileName: string): string;
    procedure   Evaluate( const aText: string);
    procedure   ClearScreen;
    procedure   ClearScales;
 public
    procedure   DoLine         ( aSender: TObject; aLineNr, aTotalCount: Integer);
    procedure   DoFileEnter    ( aSender: TObject; const aFileName: string);
    procedure   DoFileLeave    ( aSender: TObject; const aFileName: string);
    procedure   DoFileResume   ( aSender: TObject; const aFileName: string);
  public
    constructor Create( anOwner: TComponent);                                                                  override;
    destructor  Destroy;                                                                                       override;
    procedure   Reset;
    procedure   DoBreak;
    procedure   CheckStacks( const aMsg: string);
    procedure   AcceptCode  ( const aText: string; const aData: TVariants = nil); // For interactive use - can use code fragments
    procedure   AcceptAction( const aWord: string; const aData: TVariants = nil); // For automated   use - supposed to be called with one word

    // ----
    function    LeaseStackImage  ( const aData: TVariants): TForthList;
    procedure   ReleaseStackImage( aFrame: TForthList);
    function    LeaseSymbol  ( const aName: string): TSymbol;
    procedure   ReleaseSymbol( const aSymbol: TSymbol);
    procedure   AcceptWithStackImage( const aSymbol: TSymbol; const aFrame: TForthList);
    // ----
  public
    property    Base            : Cardinal         read GetBase          write SetBase;
    property    Compiling       : Boolean          read GetCompiling     write SetCompiling;
    property    Interpreting    : Boolean          read GetInterpreting  write SetInterpreting;
    property    SourceName      : string           read GetSourceName;
    property    SourceLineNr    : Integer          read GetSourceLineNr;
    property    SourceOffset    : Integer          read GetSourceOffset;
    property    Current         : TVocabulary      read FCurrent         write FCurrent;    // Compile Voc
    property    Context         : TVocabulary      read GetContext       write SetContext;  // Search  Voc
  published
    property    StartupFile     : string           read FStartupFile     write FStartupFile;
  published
    property    OnSay           : TOnForthSay      read FOnSay           write FOnSay;
    property    OnBye           : TOnBye           read FOnBye           write FOnBye;
    property    OnBreak         : TOnForthBreak    read FOnBreak         write FOnBreak;
    property    OnLog           : TOnForthLog      read FOnLog           write FOnLog;
    property    OnReset         : TNotifyEvent     read FOnReset         write FOnReset;
    property    OnUnknownSymbol : TOnUnknownSymbol read FOnUnknownSymbol write FOnUnknownSymbol;
    property    OnDataOut       : TOnDataOut       read FOnDataOut       write FOnDataOut;
    property    OnAddScale      : TOnAddScale      read FOnAddScale      write FOnAddScale;
    property    OnGetScaleNames : TOnGetScaleNames read FOnGetScalenames write FOnGetScalenames;
    property    OnClearScales   : TOnClearScales   read FOnClearScales   write FOnClearScales;
    property    OnClearScreen   : TNotifyEvent     read FOnClearScreen   write FOnClearScreen;
  published
    property    OnLine          : TOnLine          read FOnLine          write FOnLine;
    property    OnFileEnter     : TOnFileEnter     read FOnFileEnter     write FOnFileEnter;
    property    OnFileLeave     : TOnFileLeave     read FOnFileLeave     write FOnFileLeave;
    property    OnFileResume    : TOnFileresume    read FOnFileResume    write FOnFileResume;
  end;




implementation

const

  WhiteSpace : set of AnsiChar = [ #$00 .. #$20, #$7f];

var

    AllSymbolsCounter : Integer;
    AnonymousCounter  : Integer;

    function SymbolToVariant( const aSymbol: TSymbol): Variant;
    begin
      Result := Integer( aSymbol);
      TVarData( Result).VType := varByRef;
    end;

    function VariantIsString( V: Variant): Boolean;
    begin
      Result := VarIsStr( V);
    end;

    function VariantIsSymbol( V: Variant): Boolean;
    var
      W : TObject;
    begin
      if TVarData( V).VType = varByRef
      then begin
        TVarData( V).VType := varInteger;
        W := TObject( Integer( V));
        Result := W is TSymbol;
      end
      else Result := False;
    end;

    function VariantIsList( V: Variant): Boolean;
    var
      W : TObject;
    begin
      if TVarData( V).VType = varByRef
      then begin
        TVarData( V).VType := varInteger;
        W := TObject( Integer( V));
        Result := W is TForthList;
      end
      else Result := False;
    end;

    function VariantToSymbol( V: Variant): TSymbol;
    begin
      if VariantIsSymbol( V)
      then begin
        TVarData( V).VType := varInteger;
        Result := TSymbol( Integer( V));
      end
      else Result := nil;
    end;

    function VariantToList( V: Variant): TForthList;
    begin
      if VariantIsList( V)
      then begin
        TVarData( V).VType := varInteger;
        Result := TForthList( Integer( V));
      end
      else Result := nil;
    end;

    procedure VariantRelease( V: Variant);
    var
      W : TSymbol;
    begin
      if ( V <> Null) and VariantIsSymbol( V)
      then begin
        W := VariantToSymbol( V);
        W.ReleaseRef;
      end;
    end;

    function  LongToBase( aValue, aBase: UInt64; aFieldWidth: Integer = 1): string;

      function ToDigit( aValue, aBase: UInt64): string;
      begin
        Assert( aValue < aBase, 'ToDigit: Digit conversion error, aValue >= aBase');
        if aBase <= 62
        then
          if aValue > 9
          then begin
            if aValue > 35
            then Result := Char( aValue + Ord( '0') + 13)
            else Result := Char( aValue + Ord( '0') +  7);
          end
          else Result := Char( aValue + Ord( '0'))
        else Result := Format( '<&%d>', [ aValue], AppLocale);
      end;

    begin
      Result := '';
      while aValue <> 0
      do begin
        Result := ToDigit( aValue mod aBase, aBase) + Result;
        aValue := aValue div aBase;
      end;
      if aFieldWidth > 0
      then begin
        while Length( Result) < aFieldWidth
        do Result := '0' + Result;
      end;
    end;

    function QuoteEscape( const S: string): string;
    var
      C : Char;
    begin
      Result := '';
      for C in S
      do begin
        if C = '"'
        then Result := Result + '\"'
        else if C = '\'
        then Result := Result + '\\'
        else Result := Result + C;
      end;
    end;

    function DataToString( aValue: Variant; aBase: Cardinal): string;
    // done : must this be recursive? (i.e. expanding aggregates) - yes
    var
      W : TSymbol;
    begin
      if VariantIsSymbol( aValue)
      then begin
        W      := VariantToSymbol( aValue);
        Result := W.See;
      end
      else if VarIsOrdinal( aValue)
      then begin
        if aValue < 0
        then Result := '-' + LongToBase( - aValue, aBase)
        else Result :=       LongToBase(   aValue, aBase);
      end
      else if VarIsFloat( aValue)
      then Result := FloatToStrF( Extended( aValue), ffGeneral, 15, 15)
      else if VarIsStr( aValue)
      then Result := Format( '" %s"', [ QuoteEscape( aValue)], AppLocale)
      else Result := aValue;
    end;

    function DataToStringSimple( aValue: Variant; aBase: Cardinal): string;
    var
      W : TSymbol;
    begin
      if VariantIsSymbol( aValue)
      then begin
        W  := VariantToSymbol( aValue);
        if W is TForthList
        then Result := W.See
        else Result := W.Name;
      end
      else if VarIsOrdinal( aValue)
      then begin
        if aValue < 0
        then Result := '-' + LongToBase( - aValue, aBase)
        else Result :=       LongToBase(   aValue, aBase);
      end
      else if VarIsFloat( aValue)
      then Result := FloatToStrF( Extended( aValue), ffGeneral, 15, 15)
      else if VarIsStr( aValue)
      then Result := Format( '" %s"', [ QuoteEscape( aValue)], AppLocale)
      else Result := aValue;
    end;

    function DataToStringText( aValue: Variant; aBase: Cardinal): string;
    var
      W : TSymbol;
    begin
      if VariantIsSymbol( aValue)
      then begin
        W  := VariantToSymbol( aValue);
        if W is TForthList
        then Result := W.See
        else Result := W.Name;
      end
      else if VarIsOrdinal( aValue)
      then begin
        if aValue < 0
        then Result := '-' + LongToBase( - aValue, aBase)
        else Result :=       LongToBase(   aValue, aBase);
      end
      else if VarIsFloat( aValue)
      then Result := FloatToStrF( Extended( aValue), ffGeneral, 15, 15)
      else if VarIsStr( aValue)
      then Result := Format( '%s', [ aValue], AppLocale)
      else Result := aValue;
    end;

    procedure StringTofile( const S: string; var aFile: TextFile); overload;
    begin
      Rewrite( aFile);
      Write( aFile, S);
    end;

    procedure StringTofile( const S, aFileName: string); overload;
    var
      aFile: TextFile;
    begin
      AssignFile( aFile, afileName);
      try
        StringToFile( S, aFile);
      finally
        CloseFile( aFile);
      end;
    end;

    function FileToString( var aFile: TextFile): string; overload;
    var
      S : string;
    begin
      Result := '';
      Reset( aFile);
      while not Eof( aFile)
      do begin
        ReadLn( aFile, S);
        Result := Result + CarriageReturn + LineFeed + S;
      end;
    end;

    function FileToString( const aFileName: string): string; overload;
    var
      aFile : TextFile;
    begin
      AssignFile( aFile, aFileName);
      try
        FileToString( aFile);
      finally
        CloseFile( aFile);
      end;
    end;

    function  CleanupFilePath( const aPath: string): string;
    begin
      Result := ExpandFileName( aPath);
    end;

    procedure _also( aSystem: TKnobsForth);
    // also ( -- )
    // Dup top of search stack
    var
      L : Cardinal;
    begin
      with aSystem
      do begin
        L := Length( FContext) + 1;
        SetLength( FContext, L);
        FContext[ L - 1] := FContext[ L - 2];
      end;
    end;

    procedure _only( aSystem: TKnobsForth);
    // only ( -- )
    // pseudo voc containing voc control words
    begin
      with aSystem
      do begin
        SetLength( FContext, 1);  // Drop everything
        Context := FOnly;         // Set search  Voc
        Current := FOnly;         // Set compile Voc
      end;
    end;

    procedure _vocabulary( aSystem: TKnobsForth);
    // vocabulary <XXXX> ( -- )
    // to define new vocs
    begin
      with aSystem
      do begin
        MustNotBeCompilingFor( 'vocabulary');
        AddVoc( TVocabulary( AddSymbol( FetchToken( ' '), TVocabulary)));
      end;
    end;

    procedure _definitions( aSystem: TKnobsForth);
    // definitions ( -- )
    // set the top of the context stack as the current (compile to) vocabulary
    begin
      with aSystem
      do Current := Context;
    end;

    procedure _dotvocs( aSystem: TKnobsForth);
    // .vocs ( -- )
    // show list of all vocabularies
    var
      i : Cardinal;
      S : string;
    begin
      with aSystem
      do begin
        S := '';
        for i := 0 to Length( FVocs) - 1
        do begin
          if S = ''
          then S := FVocs[ i].Name
          else S := S + ' ' + FVocs[ i].Name;
        end;
        DataPush( S);
        MustFindToken( '.', TInternal).Interpret;
      end;
    end;

    procedure _dropvoc( aSystem: TKnobsForth);
    // dropvoc ( -- )
    // removes top of search voc stack
    var
      L : Cardinal;
    begin
      // drop a vocabulary, but don't drop the root vocabulary
      with aSystem
      do begin
        L := Length( FContext);
        if L > 1
        then begin
          L := L - 1;
          SetLength( FContext, L);
        end;
      end;
    end;

    procedure _dotorder( aSystem: TKnobsForth);
    // .order ( -- )
    // show search order
    var
      i : Cardinal;
      S : string;
    begin
      with aSystem
      do begin
        for i := Low( FContext) to High( FContext)
        do begin
          if S = ''
          then S := FContext[ i].Name
          else S := S + ' ' + FContext[ i].Name;
        end;
        if S = ''
        then S := '<nil>';
        if Assigned( FCurrent)
        then SayFmt( 'current: %s', [ FCurrent.Name])
        else Say   ( 'current: <nil>');
        SayFmt     ( 'context: %s', [ S]);
      end;
    end;

    procedure _reset( aSystem: TKnobsForth);
    // reset ( -- )
    begin
      aSystem.Reset;
    end;

    procedure _bye( aSystem: TKnobsForth);
    // bye ( -- )
    begin
      aSystem.Say( 'bye ...');
      aSystem.Bye;
    end;

    procedure _fload( aSystem: TKnobsForth);
    // fload ( "filename" -- "string" )
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( FLoad( V));
        VariantRelease( V);
      end;
    end;

    procedure _fsave( aSystem: TKnobsForth);
    // fsave ( contents "filename" -- )
    var
      aContents : string;
      aFileName : string;
      V         : Variant;
    begin
      with aSystem
      do begin
        aFileName := DataPop;
        V         := DataPop;
        aContents := DataToString( V, Base);
        FSave( aContents, aFileName);
        VariantRelease( V);
        VariantRelease( aFileName);
      end;
    end;

    procedure _cls( aSystem: TKnobsForth);
    // cls ( -- )
    begin
      with aSystem
      do begin
        ClearScreen;
      end;
    end;

    procedure _dataOut( aSystem: TKnobsForth);
    // dataout ( n1 .. nn n channel -- )
    var
      C : variant;
      N : variant;
      V : variant;
      D : TVariants;
      i : Integer;
    begin
      With aSystem
      do begin
        C := DataPop;
        N := DataPop;
        SetLength( D, Cardinal( N));
        for i := N - 1 downto 0
        do begin
          V := DataPop;
          D[ i] := V;
          VariantRelease( V);
        end;
        DataOut( C, D);
        VariantRelease( N);
        VariantRelease( C);
      end;
    end;

    procedure _dotwords( aSystem: TKnobsForth);
    // .words ( -- ) // show the symbol list
    var
      i : Integer;
    begin
      with aSystem
      do begin
        for i := 0 to Length( FContext) - 1
        do begin
          SayFmt( '%s:', [ FContext[ i].Name]);
          Say( '');
          Say( FContext[ i].SymbolList);
          if i <> Length( FContext) - 1
          then Say( '');
        end;
      end;
    end;

    procedure _anonymous( aSystem: TKnobsForth);
    // anonymous ( -- "string" ) // get anonymous symbol info
    begin
      with aSystem
      do DataPush( Anonymous);
    end;

    procedure _dump( aSystem: TKnobsForth);
    // dump ( -- ) // make a code dump
    begin
      with aSystem
      do CodeDump;
    end;

    procedure _vocdump( aSystem: TKnobsForth);
    // vocdump ( -- ) // make a voc based code dump
    begin
      with aSystem
      do VocDump;
    end;

    procedure _toname( aSystem: TKnobsForth);
    // >name ( <exec-token> -- "string") // Find a name for an execution token
    var
      W : TSymbol;
    begin
      with aSystem
      do begin
        W := DataPopSymbol;
        DataPush( W.Name);
        W.ReleaseRef;
      end;
    end;

    procedure _cfatoname( aSystem: TKnobsForth);
    // >name ( cfa -- "string") // Find a name for a CFA
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( CfaToSymbol( V).Name);
        VariantRelease( V);
      end;
    end;

    procedure _included( aSystem: TKnobsForth);
    // included ( -- ) // ( "fname" -- ) // include the file fname, and add the name to the files list
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        PushFile( ExpandFileName( V));
        VariantRelease( V);
      end;
    end;

    procedure _files( aSystem: TKnobsForth);
    // files ( -- "string") // return a string listing all included files
    var
      W : TForthList;
    begin
      with aSystem
      do begin
        W := Files;
        DataPushSymbol( W);
        W.ReleaseRef;
      end;
    end;

    procedure _includedq( aSystem: TKnobsForth);
    // included? ( "fname" -- f) // return true when fname is in files
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( IsIncluded( V));
        VariantRelease( V);
      end;
    end;

    procedure _required( aSystem: TKnobsForth);
    // required ( "fname" -- ) // when fname is not in the files list include the file named fname
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        Required( V);
        VariantRelease( V);
      end;
    end;


    procedure _eof( aSystem: TKnobsForth);
    // ;s ( -- ) // To stop on end of file, can be used to break out of a source file
    begin
      with aSystem
      do Interpreting := False;
    end;

    procedure _h_if( aSystem: TKnobsForth);
    // #if ... ( f -- ) immediate
    // conditional compilation
    var
      V      : Variant;
      aToken : string;
    begin
      with aSystem
      do begin
        V := DataPop;
        if V = 0
        then begin
          aToken := FetchToken( ' ');
          while
            ( not SameText( aToken, '#else'  )) and
            ( not SameText( aToken, '#fi'    )) and
            ( not SameText( aToken, endoffile))
          do aToken := FetchToken( ' ');

          if SameText( aToken, endoffile)
          then RaiseError( 'Unexpected end of file while processing "#if" or "#ifdef" directive.');
        end;
        VariantRelease( V);
      end;
    end;

    procedure _h_else( aSystem: TKnobsForth);
    // #else ... ( -- ) immediate
    // conditional compilation
    var
      aToken : string;
    begin
      with aSystem
      do begin
        aToken := FetchToken( ' ');
        while
          ( not SameText( aToken, '#fi'    )) and
          ( not SameText( aToken, endoffile))
        do aToken := FetchToken( ' ');

        if SameText( aToken, endoffile)
        then RaiseError( 'Unexpected end of file while processing "#else" directive.');
      end;
    end;

    procedure _h_fi( aSystem: TKnobsForth);
    // #fi ( -- ) immediate
    // conditional compilation
    begin
      // no code needed, just a place holder, see:
      // _h_if (and _ifdef) (for #if and #ifdef)
    end;

    procedure _h_define( aSystem: TKnobsForth);
    // #define XXXX ( -- ) immediate
    // defines a symbol without any semantics.
    // symbols without semantics can be checked by #defined or #ifdef
    // for conditional interpretation steering and they can serve
    // as a symbol - symbols push themselves on the stack when executed
    // symbols can not serve as a variable tho.
    begin
      with aSystem
      do FinalizeCreation( AddSymbol( FetchToken( ' '), TSymbol), '( -- ) // a #defined, does nothing');
    end;

    procedure _h_defined( aSystem: TKnobsForth);
    // #defined XXXX ( -- f ) immediate
    // check if XXXX was defined
    begin
      with aSystem
      do begin
        if FindToken( FetchToken( ' ')) = nil
        then DataPush(   0)
        else DataPush( - 1);
      end;
    end;

    procedure _h_ifdef( aSystem: TKnobsForth);
    // #ifdef XXXX ... ( -- ) immediate
    // check if XXXX was defined, combined wit #if
    begin
      with aSystem
      do begin
        if FindToken( FetchToken( ' ')) = nil
        then DataPush(   0)
        else DataPush( - 1);
      end;
      _h_if( aSystem);
    end;

    procedure _h_ifndef( aSystem: TKnobsForth);
    // #ifndef XXXX ... ( -- ) immediate
    // check if next inline symbol is not defined, combined wit #if
    begin
      with aSystem
      do begin
        if Assigned( FindToken( FetchToken( ' ')))
        then DataPush(   0)
        else DataPush( - 1);
      end;
      _h_if( aSystem);
    end;

    procedure _getmem( aSystem: TKnobsForth);
    // getmem ( size -- pointer ) // allocate memory from the system heap
    var
      P : Pointer;
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        GetMem( P, Cardinal( V));
        DataPush( Cardinal( P));
        VariantRelease( V);
      end;
    end;

    procedure _freemem( aSystem: TKnobsForth);
    // freemem ( pointer -- ) // free previously allocated memory
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        FreeMem( Pointer( Cardinal( V)));
        VariantRelease( V);
      end;
    end;

    procedure _reallocmem( aSystem: TKnobsForth);
    // reallocmem ( pointer size -- pointer ) // reallocate previously allocated memory
    var
      aSize : Variant;
      V     : Variant;
      P     : Pointer;
    begin
      with aSystem
      do begin
        aSize := DataPop;
        V     := DataPop;
        P     := Pointer( Cardinal( V));
        ReallocMem( P, Cardinal( aSize));
        DataPush( Cardinal( P));
        VariantRelease( V);
        VariantRelease( aSize);
      end;
    end;

    procedure _memfetch( aSystem: TKnobsForth);
    // mem@ ( pointer offset n -- n1 .. nn )  // read n cell sized items from an offset on pointer
    type
      TCardinals = array[ 0 .. Maxint div SizeOf( Cardinal) - 1] of Cardinal;
      PCardinals = ^TCardinals;
    var
      aSize    : Variant;
      anOffset : Variant;
      V        : Variant;
      aPointer : PCardinals;
      i        : Cardinal;
    begin
      with aSystem
      do begin
        aSize    := DataPop;
        anOffset := DataPop;
        V        := DataPop;
        aPointer := Pointer( Cardinal( V));
        for i := 0 to aSize - 1
        do DataPush( aPointer^[ Cardinal( anOffset) + i]);
        DataPush( aSize);
        VariantRelease( V       );
        VariantRelease( anOffset);
        VariantRelease( aSize   );
      end;
    end;

    procedure _memstore( aSystem: TKnobsForth);
    // mem! ( n1 .. nn n pointer offset -- ) // store n cell sized items to an offset on pointer
    type
      TCardinals = array[ 0 .. Maxint div SizeOf( Cardinal) - 1] of Cardinal;
      PCardinals = ^TCardinals;
    var
      aSize    : Variant;
      V        : Variant;
      anOffset : Variant;
      aPointer : PCardinals;
      i        : Cardinal;
    begin
      with aSystem
      do begin
        anOffset := DataPop;
        V        := DataPop;
        aPointer := Pointer( Cardinal( V));
        aSize    := DataPop;
        VariantRelease( V);
        for i := 0 to aSize - 1
        do begin
          V := DataPop;
          aPointer^[ Cardinal( anOffset) + i] := V;
          VariantRelease( V);
        end;
        VariantRelease( aSize   );
        VariantRelease( anOffset);
      end;
    end;

    procedure _tostring( aSystem: TKnobsForth);
    // ($) ( n -- str )
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( DataToStringSimple( V, Base));
        VariantRelease( V);
      end;
    end;

    procedure _format( aSystem: TKnobsForth);
    // format ( n fmt -- "string" )
    var
      aFmt : Variant;
      aVal : Variant;
    begin
      with aSystem
      do begin
        aFmt := DataPop;
        aVal := DataPop;
        DataPush( Format( aFmt, [ Double( aVal)], AppLocale));
        VariantRelease( aVal);
        VariantRelease( aFmt);
      end;
    end;

    procedure _dot( aSystem: TKnobsForth);
    // . ( n -- ) // print n to console
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        Say( DataToStringText( V, Base));
        VariantRelease( V);
      end;
    end;

    procedure _dotdebug( aSystem: TKnobsForth);
    // .debug ( n -- ) // pring n to debug output
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        Log( DataToString( DataPop, Base));
        VariantRelease( V);
      end;
    end;

    procedure _dots( aSystem: TKnobsForth);
    // .s ( -- ) // show data stack on console
    var
      aStringList : TStringList;
      i           : Integer;
    begin
      with aSystem
      do begin
        aStringList := TStringList.Create;
        try
          FDataStack.Dump( aStringList, 0);
          for i := 0 to aStringList.Count - 1
          do Say( aStringList[ i]);
        finally
          aStringList.Free;
        end;
      end;
    end;

    procedure _dotr( aSystem: TKnobsForth);
    // .r ( -- ) // show return stack on console
    var
      aStringList : TStringList;
      i           : Integer;
    begin
      with aSystem
      do begin
        aStringList := TStringList.Create;
        try
          FReturnStack.Dump( aStringList, 0);
          for i := 0 to aStringList.Count - 1
          do Say( aStringList[ i]);
        finally
          aStringList.Free;
        end;
      end;
    end;

    procedure _dotc( aSystem: TKnobsForth);
    // .c ( -- ) // show control stack on console
    var
      aStringList : TStringList;
      i           : Integer;
    begin
      with aSystem
      do begin
        aStringList := TStringList.Create;
        try
          FControlStack.Dump( aStringList, 0);
          for i := 0 to aStringList.Count - 1
          do Say( aStringList[ i]);
        finally
          aStringList.Free;
        end;
      end;
    end;

    procedure _dotstacks( aSystem: TKnobsForth);
    // .stacks ( n -- ) // show all stacks on console
    begin
      _dots( aSystem);
      _dotr( aSystem);
      _dotc( aSystem);
    end;

    procedure _decimal( aSystem: TKnobsForth); // Base := D'10'
    // decimal ( -- )
    // det base to &10
    begin
      aSystem.Base := 10;
    end;

    procedure _hex( aSystem: TKnobsForth); // Base := D'16'
    // hex ( -- )
    // set base to &16
    begin
      aSystem.Base := 16;
    end;

    procedure _binary( aSystem: TKnobsForth); // Base := D'2'
    // binary ( -- )
    // set base to &2
    begin
      aSystem.Base := 2;
    end;

      function ScanDoc( aSystem: TKnobsForth): string;
      var
        aToken : string;
        p      : Integer;
        S      : String;
      begin
        Result := '';
        with aSystem
        do begin
          aToken := GetToEol;
          p      := Pos( '#', LowerCase( aToken));
          while ( not SameText( aToken, endoffile)) and ( p = 0)
          do begin
            S := Trim( aToken);
            if ( Result = '')
            then begin
              if S <> ''
              then Result := Result + S + ^M;
            end
            else Result := Result + S + ^M;
            aToken := ReadLine;
            p      := Pos( '#cod', LowerCase( aToken));
          end;
          if SameText( aToken, endoffile)
          then raiseError( 'Unexpected end of file in "#doc (or #document xxxx) .... #cod" definition');
          SkipToEol;
          if p > 0
          then begin
            // Get last bit into result
            S := Trim( Copy( aToken, 1, p - 1));
            if S <> ''
            then Result := Result + S;
            // Push back the remainder of the line containing #cod to later scan the remaining chars.
            S := Copy( atoken, p + Length( '#cod'), Length( atoken));
            if S <> ''
            then PushLine( S, 'scandoc');
          end;
        end;
      end;

    procedure _cod( aSystem: TKnobsForth);
    // #cod ( -- )
    // just a dummy, ends #doc
    begin
    end;

    procedure _doc( aSystem: TKnobsForth);
    // #doc ... #cod ( -- ) Immediate
    // documents the last symbol used
    begin
      with aSystem
      do FLastDefined.AddDocs( ScanDoc( aSystem));
    end;

    procedure _doceol( aSystem: TKnobsForth);
    // #// ... ( -- ) Immediate
    // documents the last symbol used, documentation stops at end of line
    begin
      with aSystem
      do FLastDefined.AddDocs( GetToEol);
    end;

    procedure _document( aSystem: TKnobsForth);
    // #document XXXX ... #cod ( -- )
    // documents XXXX
    begin
      with aSystem
      do MustFindToken( FetchToken( ' '), TSymbol).AddDocs( ScanDoc( aSystem));
    end;

    procedure _documenteol( aSystem: TKnobsForth);
    // ##// ... ( -- )
    begin
      with aSystem
      do MustFindToken( FetchToken( ' '), TSymbol).AddDocs( GetToEol);
    end;

    procedure _lparen( aSystem: TKnobsForth);
    // ( ( -- )
    // comment up to )
    var
      S : string;
    begin
      with aSystem
      do begin
        S := FetchToken( ')');
        if SameText( S, endoffile)
        then
          RaiseError(
            'unterminated comment, missing ")" in "("'
          );
        if FLastDefined is TColon
        then FLastDefined.Comment := S;
      end;
    end;

    procedure _lcurly( aSystem: TKnobsForth);
    // { ( -- {list} )
    // construct a space separated list up to }
    // done : support recursive list elements
    // done : must lists be compiled? - erm depends, but list elementas must
    //        be compiled into the list.

    var
      S : string;
      L : TForthList;
    begin
      with aSystem
      do begin
        L := TForthList.Create( aSystem);
        try
          S := FetchToken( ' ');
          while not SameText( S, '}') and not SameText( S, endoffile)
          do begin
            L.AcceptToken( S);
            S := FetchToken( ' ');
          end;

          if SameText( S, endoffile)
          then begin
            RaiseError(
              'unterminated list, missing "}" in "{"'
            );
          end;

          if Compiling
          then CompileSymbol ( L)
          else DataPushSymbol( L);
          L.ReleaseRef;
        except
          L.ReleaseRef;
        end;
      end;
    end;

    procedure _rcurly( aSystem: TKnobsForth);
    // } ( -- ) // A noop to keep the list parser happy
    begin
      // just a noop to keep the list parser happy
    end;


    procedure _tolist( aSystem: TKnobsForth);
    // >list ( <exec-token> -- {list} ) // convert execution token into a quotation');
    var
      V : Variant;
      W : TSymbol;
      L : TForthList;
    begin
      with aSystem
      do begin
        V := DataPop;
        if VariantIsSymbol( V)
        then begin
          W := VariantToSymbol( V);
          DataPushSymbol( W.AsList);
          W.ReleaseRef;
        end
        else begin
          L := TForthList.Create( aSystem);
          L.CompileVariant( V);
          VariantRelease( V);
        end;
      end;
    end;

    procedure _call( aSystem: TKnobsForth);
    // call ( {list} -- )
    // executes (interprets) the list - should behave like the joy i combinator
    // if not a list it will just interpret the symbol
    // if not a symbol an error will be raised
    var
      W : TSymbol;
    begin
      with aSystem
      do begin
        W := DataPopSymbol;
        if W is TForthList
        then TForthList( W).Call
        else W.Interpret;
        W.ReleaseRef;
      end;
    end;

    procedure _stack( aSystem: TKnobsForth);
    // stack ( x y z -- x y z { x y z } ) // Push the current stack as a list on top of itself
    var
      W: TSymbol;
    begin
      with aSystem
      do begin
        W := FDataStack.AsList;   // Makes refcount on W = 1
        DataPushSymbol( W);       // Adds a ref, while we actually only have one
        W.ReleaseRef;             // So remove that ref again
      end;
    end;

    procedure _unstack( aSystem: TKnobsForth);
    // unstack ( {list}or"string" -- n1 ... nn )
    var
      aVal  : Variant;
      aCode : TForthList;
      i     : Cardinal;
      S     : string;
    begin
      with aSystem
      do begin
        aVal := DataPop;
        if VariantIsString( aVal)
        then begin
          S := aVal;
          for i := Length( S) downto 1
          do DataPush( S[ i]);
        end
        else if VariantIsList( aVal)
        then begin
          aCode := VariantToList( aVal);
          for i := aCode.Count - 1 downto 0
          do DataPushSymbol( aCode.Symbol[ i]);
          aCode.ReleaseRef;
        end
        else begin
          VariantRelease( aVal);
          RaiseError( 'list or string expected for "unstack"');
        end;
      end;
    end;

    procedure _swons( aSystem: TKnobsForth);
    // swons ( {list}or"string" n -- {list}or"string" )
    // append n to list or string
    var
      aVal  : Variant;
      bVal  : Variant;
      aCode : TForthList;
    begin
      with aSystem
      do begin
        aVal := DataPop;
        bVal := DataPop;
        try
          if VariantIsString( bVal)
          then DataPush( bVal + aVal)
          else if VariantIsList( bVal)
          then begin
            aCode := VariantToList( bVal).Copy;
            aCode.CompileLiteral( aVal);
            DataPushSymbol( aCode);
          end
          else RaiseError( 'list or string expected for "swons"');
        finally
          VariantRelease( aVal);
          VariantRelease( bVal);
        end;
      end;
    end;

    procedure _first( aSystem: TKnobsForth);
    // first ( {list}or"string" -- n )
    // get head of list or string
    var
      aVal  : Variant;
      aCode : TForthList;
      S     : string;
    begin
      with aSystem
      do begin
        aVal := DataPop;
        if VariantIsString( aVal)
        then begin
          S := aVal;
          if Length( S) > 0
          then DataPush( S[ 1])
          else RaiseError( 'empty string has no first item in "first"');
        end
        else if VariantIsList( aVal)
        then begin
          aCode := VariantToList( aVal);
          if aCode.Count > 0
          then begin
            DataPushSymbol( aCode.First);
            aCode.ReleaseRef;
          end
          else begin
            aCode.ReleaseRef;
            RaiseError( 'empty list has no first item in "first"');
          end;
        end
        else DataPush( aVal)
      end;
    end;

    procedure _rest( aSystem: TKnobsForth);
    // rest ( {list}or"string" -- {list}or"string" )
    // remove first element from list or string
    var
      aVal  : Variant;
      aCode : TForthList;
      S     : string;
    begin
      with aSystem
      do begin
        aVal := DataPop;
        try
          if VariantIsString( aVal)
          then begin
            S := aVal;
            if Length( S) > 0
            then DataPush( Copy( S, 2, Length( S) - 1))
            else RaiseError( 'empty string has no rest in "rest"');
          end
          else if VariantIsList( aVal)
          then begin
            aCode := VariantToList( aVal).Copy;
            if aCode.Count > 0
            then begin
              aCode.Rest;
              DataPushSymbol( aCode);
            end
            else RaiseError( 'empty list has no rest in "rest"');
          end
          else DataPush( aVal)
        finally
          VariantRelease( aVal);
        end;
      end;
    end;

    procedure _concat( aSystem: TKnobsForth);
    // concat ( {list}or"string" {list}or"string" -- {list}or"string" )
    // concatenate two lists or two strings
    // the + operator works for strings too, but not for lists
    var
      aValA  : Variant;
      aValB  : Variant;
      aCodeA : TForthList;
      aCodeB : TForthList;
    begin
      with aSystem
      do begin
        aValA := DataPop;
        aValB := DataPop;
        try
          if( VariantIsString( aValA) and VariantIsString( aValB))
          then DataPush( aValA + aValB)
          else if( VariantIsList( aValA) and VariantIsList( aValB))
          then begin
            aCodeA := VariantToList( aValA).Copy;
            aCodeB := VariantToList( aValB);
            aCodeA.Concat( aCodeB);
            DataPushSymbol( aCodeA);
          end
          else RaiseError( 'two list or two string objects expected for "concat"');
        finally
          VariantRelease( aValA);
          VariantRelease( aValB);
        end;
      end;
    end;

    procedure _dotlparen( aSystem: TKnobsForth);
    // .( ( -- )
    // print up to )
    var
      S : string;
    begin
      with aSystem
      do begin
        S := FetchToken( ')');
        if SameText( S, endoffile)
        then
          RaiseError(
            'missing ")" in ".("'
          );
        if Compiling
        then begin
          CompileLiteral( S);
          CompileToken( '.');
        end
        else Say( S);
      end;
    end;

    procedure _dotquote( aSystem: TKnobsForth);
    // ." ( -- )
    // print up to "
    var
      S : string;
    begin
      with aSystem
      do begin
        S := FetchToken( '"');
        if SameText( S, endoffile)
        then
          RaiseError(
            'missing """ in ".""'
          );
        if Compiling
        then begin
          CompileLiteral( S);
          CompileToken( '.');
        end
        else Say( S);
      end;
    end;

    procedure _slashslash( aSystem: TKnobsForth);
    // // ( -- )
    // comment up to EOL
    begin
      aSystem.SkipToEol;
    end;

    procedure _dotslashslash( aSystem: TKnobsForth);
    // .// ( -- )
    // print up to EOL
    var
      S : string;
    begin
      with aSystem
      do begin
        S := aSystem.GetToEol;
        if Compiling
        then begin
          CompileLiteral( S);
          CompileToken( '.');
        end
        else Say( S);
      end;
    end;

    procedure _swap( aSystem: TKnobsForth);
    // swap ( n1 n2 -- n2 n1 )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        B := DataPop;
        DataPush( A);
        DataPush( B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _at( aSystem: TKnobsForth);
    // at ( {list}or"string" index -- item ) // Get list or string item at index
    var
      aVal  : Variant;
      aCode : TForthList;
      p     : Variant;
    begin
      with aSystem
      do begin
        _swap( aSystem);
        aVal := DataPop;
        p    := DataPop;
        if VariantIsList( aVal)
        then begin
          aCode := VariantToList( aVal);
          DataPushSymbol( aCode.Symbol[ p]);
          aCode.ReleaseRef;
          VariantRelease( p);
        end
        else if VariantIsString( aVal)
        then begin
          DataPush( Copy( aVal, Cardinal( p) + 1, 1));
          VariantRelease( p);
        end
        else begin
          VariantRelease( aVal);
          VariantRelease( p);
          RaiseError( 'string or list object expected for "at"');
        end;
      end;
    end;

    procedure _count( aSystem: TKnobsForth);
    // count ( {list}or"string" -- count ) // Get string length or nr of items in a list
    var
      aVal  : Variant;
      aCode : TForthList;
    begin
      with aSystem
      do begin
        aVal := DataPop;
        if VariantIsList( aVal)
        then begin
          aCode := VariantToList( aVal);
          DataPush( aCode.Count);
          aCode.ReleaseRef;
        end
        else if VariantIsString( aVal)
        then DataPush( Length( aVal))
        else begin
          VariantRelease( aVal);
          RaiseError( 'string or list object expected for "count"');
        end;
      end;
    end;

    procedure _map( aSystem: TKnobsForth);
    // map ( {list} {quotation} -- {list} )
    var
      Quotation  : TForthList;
      List       : TForthList;
      ResultList : TForthList;
      i          : Integer;
    begin
      with aSystem
      do begin
        Quotation := DataPopSymbol as TForthList;
        List      := DataPopSymbol as TForthList;

        ResultList := TForthList.Create( aSystem);

        for i := 0 to List.Count - 1
        do begin
          List.Symbol[ i].Interpret;
          Quotation.Interpret;
          _call( aSystem);
          ResultList.CompileVariant( DataPop);
        end;
        DataPushSymbol( ResultList);

        ResultList.ReleaseRef;
        List      .ReleaseRef;
        Quotation .ReleaseRef;
      end;
    end;

    procedure _step( aSystem: TKnobsForth);
    // step ( {list} {quotation} -- )
    var
      Quotation : TForthList;
      List      : TForthList;
      i         : Integer;
    begin
      with aSystem
      do begin
        Quotation := DataPopSymbol as TForthList;
        List      := DataPopSymbol as TForthList;

        for i := 0 to List.Count - 1
        do begin
          List.Symbol[ i].Interpret;
          Quotation.Interpret;
          _call( aSystem);
        end;

        List      .ReleaseRef;
        Quotation .ReleaseRef;
      end;
    end;

    procedure _filter( aSystem: TKnobsForth);
    // filter ( {list} {quotation} -- {list} )
    var
      Quotation  : TForthList;
      List       : TForthList;
      ResultList : TForthList;
      W          : TSymbol;
      i          : Integer;
      V          : Variant;
    begin
      with aSystem
      do begin
        Quotation  := DataPopSymbol as TForthList;
        List       := DataPopSymbol as TForthList;
        ResultList := TForthList.Create( aSystem);

        for i := 0 to List.Count - 1
        do begin
          W := List.Symbol[ i];
          W.Interpret;
          Quotation.Interpret;
          _call( aSystem);    // Dequote the Quotation, that is: run it
          V := DataPop;
          if V
          then ResultList.CompileVariant( TLiteral( W).Value);
        end;

        DataPushSymbol( ResultList);
        ResultList.ReleaseRef;
        List      .ReleaseRef;
        Quotation .ReleaseRef;
      end;
    end;

    procedure _reduce( aSystem: TKnobsForth);
    // reduce ( {list} n {quotation} -- n )
    var
      Quotation : TForthList;
      List      : TForthList;
      V         : Variant;
      i         : Integer;
    begin
      with aSystem
      do begin
        Quotation := DataPopSymbol as TForthList;
        V         := DataPop;
        List      := DataPopSymbol as TForthList;

        DataPush( V);
        for i := 0 to List.Count - 1
        do begin
          List.Symbol[ i].Interpret;
          Quotation.Interpret;
          _call( aSystem);
        end;

        List     .ReleaseRef;
        Quotation.ReleaseRef;
      end;
    end;

    procedure _explode( aSystem: TKnobsForth);
    // explode ( "string" "splitter" -- {list} )
    var
      aStringList : TStringList;
      aString     : string;
      aSplitter   : string;
      aResult     : TForthList;
      i           : Integer;
    begin
      with aSystem
      do begin
        aSplitter   := DataPop;
        aString     := DataPop;
        aStringList := Explode( aString, aSplitter);
        try
          aResult := TForthList.Create( aSystem);
          for i := 0 to aStringList.Count - 1
          do aResult.CompileVariant( aStringList[ i]);
          DataPushSymbol( aResult);
          aResult.ReleaseRef;
        finally
          aStringList.Free;
        end;
      end;
    end;

    procedure _implode( aSystem: TKnobsForth);
    // implode ( {list} "joiner" -- "string" )
    var
      aJoiner : string;
      aList   : TForthList;
      aResult : string;
      i       : Integer;
      S       : string;
    begin
      with aSystem
      do begin
        aJoiner := DataPop;
        aList   := DataPopSymbol as TForthList;
        aResult := '';
        for i := 0 to aList.Count - 1
        do begin
          aList.Symbol[ i].Interpret;
          S := DataPop;
          if aResult = ''
          then aResult := S
          else aResult := aResult + aJoiner + S;
        end;
        DataPush( aResult);
        aList.ReleaseRef;
      end;
    end;


    procedure _plus( aSystem: TKnobsForth);
    // + ( n n -- n )
    var
      a1, a2 : Variant;
    begin
      with aSystem
      do begin
          // Weird, but makes string addition to be performed left to right
          // and for integer addition it does not matter ...
        a1 := DataPop;
        a2 := DataPop;
        DataPush( a2 + a1);
        VariantRelease( a1);
        VariantRelease( a2);
      end;
    end;

    procedure _minus( aSystem: TKnobsForth);
    // - ( n n -- n )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        B := DataPop;
        A := DataPop;
        DataPush( A - B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _times( aSystem: TKnobsForth);
    // * ( n n -- n )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        B := DataPop;
        A := DataPop;
        DataPush( A * B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _div( aSystem: TKnobsForth);
    // / ( n n - n )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        B := DataPop;
        A := DataPop;
        if VarIsFloat( A) or VarIsFloat( B)
        then DataPush( A / B)
        else DataPush( A div B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _divMod( aSystem: TKnobsForth);
    // /mod ( n n -- n n )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        B := DataPop;
        A := DataPop;
        if VarIsFloat( A) or VarIsFloat( B)
        then DataPush( A / B)
        else DataPush( A div B);
        DataPush( A mod B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _int( aSystem: TKnobsForth);
    // int ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Int( V));
        VariantRelease( V);
      end;
    end;

    procedure _frac( aSystem: TKnobsForth);
    // frac ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Frac( V));
        VariantRelease( V);
      end;
    end;

    procedure _power( aSystem: TKnobsForth);
    // ^ ( n n -- n )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        B := DataPop;
        A := DataPop;
        DataPush( Power( A, B));
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _exp( aSystem: TKnobsForth);
    // exp ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Exp( V));
        VariantRelease( V);
      end;
    end;

    procedure _ln( aSystem: TKnobsForth);
    // ln ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Ln( V));
        VariantRelease( V);
      end;
    end;

    procedure _rad2deg( aSystem: TKnobsForth);
    // rad2deg ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( RadToDeg( V));
        VariantRelease( V);
      end;
    end;

    procedure _deg2rad( aSystem: TKnobsForth);
    // deg2rad ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( DegToRad( V));
        VariantRelease( V);
      end;
    end;

    procedure _sin( aSystem: TKnobsForth);
    // sin ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Sin( V));
        VariantRelease( V);
      end;
    end;

    procedure _asin( aSystem: TKnobsForth);
    // asin ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcSin( V));
        VariantRelease( V);
      end;
    end;

    procedure _cos( aSystem: TKnobsForth);
    // cos ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Cos( V));
        VariantRelease( V);
      end;
    end;

    procedure _acos( aSystem: TKnobsForth);
    // acos ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcCos( V));
        VariantRelease( V);
      end;
    end;

    procedure _tan( aSystem: TKnobsForth);
    // tan ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Tan( V));
        VariantRelease( V);
      end;
    end;

    procedure _atan( aSystem: TKnobsForth);
    // atan ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcTan( V));
        VariantRelease( V);
      end;
    end;

    procedure _atan2( aSystem: TKnobsForth);
    // atan2 ( n1 n2 -- n )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        B := DataPop;
        DataPush( ArcTan2( B, A));
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _cosh( aSystem: TKnobsForth);
    // cosh ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Cosh( V));
        VariantRelease( V);
      end;
    end;

    procedure _sinh( aSystem: TKnobsForth);
    // sinh ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Sinh( V));
        VariantRelease( V);
      end;
    end;

    procedure _tanh( aSystem: TKnobsForth);
    // tanh ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Tanh( V));
        VariantRelease( V);
      end;
    end;

    procedure _atanh( aSystem: TKnobsForth);
    // atanh ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcTanh( V));
        VariantRelease( V);
      end;
    end;

    procedure _sech( aSystem: TKnobsForth);
    // sech ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( SecH( V));
        VariantRelease( V);
      end;
    end;

    procedure _csch( aSystem: TKnobsForth);
    // csch ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( CscH( V));
        VariantRelease( V);
      end;
    end;

    procedure _acot( aSystem: TKnobsForth);
    // acot ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcCot( V));
        VariantRelease( V);
      end;
    end;

    procedure _asec( aSystem: TKnobsForth);
    // asec ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcSec( V));
        VariantRelease( V);
      end;
    end;

    procedure _acsc( aSystem: TKnobsForth);
    // acsc ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcCsc( V));
        VariantRelease( V);
      end;
    end;

    procedure _acosh( aSystem: TKnobsForth);
    // acosh ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcCosh( V));
        VariantRelease( V);
      end;
    end;

    procedure _asinh( aSystem: TKnobsForth);
    // asinh ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcSinh( V));
        VariantRelease( V);
      end;
    end;

    procedure _acoth( aSystem: TKnobsForth);
    // acoth ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcCotH( V));
        VariantRelease( V);
      end;
    end;

    procedure _asech( aSystem: TKnobsForth);
    // asech ( n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcSecH( V));
        VariantRelease( V);
      end;
    end;

    procedure _acsch( aSystem: TKnobsForth);
    // acsch ( n n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( ArcCscH( V));
        VariantRelease( V);
      end;
    end;

    procedure _sqrt( aSystem: TKnobsForth);
    // sqrt ( n n -- n )
    var
      V : variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Sqrt( V));
        VariantRelease( V);
      end;
    end;

    procedure _and( aSystem: TKnobsForth);
    // and ( n n -- n )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        B := DataPop;
        DataPush( A and B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _or( aSystem: TKnobsForth);
    // or ( n n -- n )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        B := DataPop;
        DataPush( A or B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _xor( aSystem: TKnobsForth);
    // xor ( n n -- n )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        A := DataPop;
        DataPush( A xor B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _constant( aSystem: TKnobsForth);
    // constant XXXX ( n -- )
    // define a constant
    var
      W : TConstant;
      V : Variant;
    begin
      with aSystem
      do begin
        W := AddSymbol( FetchToken( ' '), TConstant) as TConstant;
        V := DataPop;
        W.Value := V;
        VariantRelease( V);
        FinalizeCreation( W);
      end;
    end;

    procedure _variable( aSystem: TKnobsForth);
    // variable XXXX ( -- )
    // define a variable
    begin
      with aSystem
      do FinalizeCreation( TVariable( AddSymbol( FetchToken( ' '), TVariable)));
    end;

    procedure _scale( aSystem: TKnobsForth);
    // { n1 n2 .. nn } makescale [name]
    // scale ( {list} -- )
    var
      W : TScale;
      V : Variant;
    begin
      with aSystem
      do begin
        W := AddSymbol( FetchToken( ' '), TScale) as TScale;
        V := DataPop;
        W.addList( V);
        VariantRelease( V);
        FinalizeCreation( W);
        W.Interpret;
      end;
    end;

    procedure _scalenames( aSystem: TKnobsForth);
    // scaleneames ( -- string )
    begin
      with aSystem
      do DataPush( GetScaleNames);
    end;

    procedure _fetch( aSystem: TKnobsForth);
    // @ ( var-token -- val )
    var
      W : TSymbol;
    begin
      with aSystem
      do begin
        W := DataPopSymbol;
        DataPush(( W as TLiteral).Value);
        W.ReleaseRef;
      end;
    end;

    procedure _store( aSystem: TKnobsForth);
    // ! ( val var-token -- )
    var
      aVar : TLiteral;
    begin
      with aSystem
      do begin
        aVar := DataPopSymbol as TLiteral;
        aVar.Value := DataPop;
        aVar.ReleaseRef;
      end;
    end;

    procedure _plusStore( aSystem: TKnobsForth);
    // +! ( val var-token -- )
    var
      aVar : TVariable;
    begin
      with aSystem
      do begin
        aVar := DataPopSymbol as TVariable;
        aVar.Value := aVar.Value + DataPop;
        aVar.ReleaseRef;
      end;
    end;

    procedure _minStore( aSystem: TKnobsForth);
    // -! ( val var-token -- )
    var
      aVar : TVariable;
    begin
      with aSystem
      do begin
        aVar := DataPopSymbol as TVariable;
        aVar.Value := aVar.Value - DataPop;
        aVar.ReleaseRef;
      end;
    end;

    procedure _drop( aSystem: TKnobsForth);
    // drop ( n -- )
    begin
      with aSystem
      do VariantRelease( DataPop);
    end;

    procedure _dup( aSystem: TKnobsForth);
    // dup ( n -- n n )
    var
      A : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        DataPush( A);
        DataPush( A);
        VariantRelease( A);
      end;
    end;

    procedure _true( aSystem: TKnobsForth);
    // true ( -- f )
    begin
      with aSystem
      do DataPush( True);
    end;

    procedure _false( aSystem: TKnobsForth);
    // false ( -- f )
    begin
      with aSystem
      do DataPush( False);
    end;

    procedure _e( aSystem: TKnobsForth);
    // e ( -- n )
    begin
      with aSystem
      do DataPush( Exp( 0));
    end;

    procedure _pi( aSystem: TKnobsForth);
    // pi ( -- n )
    begin
      with aSystem
      do DataPush( pi);
    end;

    procedure _space( aSystem: TKnobsForth);
    // space ( -- n )
    begin
      with aSystem
      do DataPush( ' ');
    end;

    procedure _negate( aSystem: TKnobsForth);
    // negate ( n -- -n )
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( - V);
        VariantRelease( V);
      end;
    end;

    procedure _not( aSystem: TKnobsForth);
    // not ( n -- !n )
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( not DataPop);
        VariantRelease( V);
      end;
    end;

    procedure _equal( aSystem: TKnobsForth);
    // = ( n1 n2 -- f )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        B := DataPop;
        DataPush( A = B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _unequal( aSystem: TKnobsForth);
    // <> ( n1 n2 -- f )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        B := DataPop;
        A := DataPop;
        DataPush( A <> B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _smaller( aSystem: TKnobsForth);
    // < ( n1 n2 -- f )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        B := DataPop;
        A := DataPop;
        DataPush( A < B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _greater( aSystem: TKnobsForth);
    // > ( n1 n2 -- f )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        B := DataPop;
        A := DataPop;
        DataPush( A > B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _smallerEqual( aSystem: TKnobsForth);
    // <= ( n1 n2 -- f )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        B := DataPop;
        A := DataPop;
        DataPush( A <= B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _greaterEqual( aSystem: TKnobsForth);
    // >= ( n1 n2 -- f )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        B := DataPop;
        A := DataPop;
        DataPush( A >= B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _random( aSystem: TKnobsForth);
    // random ( n1 n2 -- n3 )
    var
      Start  : Variant;
      Finish : Variant;
    begin
      with aSystem
      do begin
        Finish := DataPop;
        Start  := DataPop;
        DataPush( Start + Random( Finish - Start + 1));
        VariantRelease( Start );
        VariantRelease( Finish);
      end;
    end;

    procedure _now( aSystem: TKnobsForth);
    // now ( -- n )
    begin
      with aSystem
      do DataPush( Now);
    end;

    procedure _utc( aSystem: TKnobsForth);
    // utc ( -- n ) // push current UTC time as datetime
    begin
      with aSystem
      do DataPush( UTC);
    end;

    procedure _timetoosc( aSystem: TKnobsForth);
    // time>osc ( -- n ) // convert datetime to OSC time
    begin
      with aSystem
      do DataPush( DateTimeToOscTime( DataPop));
    end;

    procedure _osctotime( aSystem: TKnobsForth);
    // osc>time ( -- n ) // convert OSC time to datetime
    begin
      with aSystem
      do DataPush( OscTimeToDateTime( DataPop));
    end;

    procedure _oscnow( aSystem: TKnobsForth);
    // oscnow ( -- n ) // push current local time as OSC time
    begin
      with aSystem
      do DataPush( OscNow);
    end;

    procedure _format_time( aSystem: TKnobsForth);
    // format_time ( nt nf -- s )
    var
      F : variant;
    begin
      with aSystem
      do begin
        F := DataPop;
        DataPush( FormatDateTime( F, DataPop, AppLocale));
        VariantRelease( F);
      end;
    end;

    procedure _replace( aSystem: TKnobsForth);
    // replace ( "string" "search" "replace" -- "string" ) // replace all occurances of search in string with replace
    // in a case sensitive manner
    var
      aReplace : string;
      aSearch  : string;
      aString  : string;
      aResult  : string;
    begin
      with aSystem
      do begin
        aReplace := DataPop;
        aSearch  := DataPop;
        aString  := DataPop;
        aResult  := StringReplace( aString, aSearch, aReplace, [ rfReplaceAll]);
        DataPush( aResult);
      end
    end;

    procedure _replacei( aSystem: TKnobsForth);
    // replacei ( "string" "search" "replace" -- "string" ) // replace all occurances of search in string with replace
    // in a case insensitive manner
    var
      aReplace : string;
      aSearch  : string;
      aString  : string;
      aResult  : string;
    begin
      with aSystem
      do begin
        aReplace := DataPop;
        aSearch  := DataPop;
        aString  := DataPop;
        aResult  := StringReplace( aString, aSearch, aReplace, [ rfReplaceAll, rfIgnoreCase]);
        DataPush( aResult);
      end
    end;

    procedure _replacefirst( aSystem: TKnobsForth);
    // replacefirst ( "string" "search" "replace" -- "string" ) // replace the first occurance of search in string with replace
    // in a case sensitive manner
    var
      aReplace : string;
      aSearch  : string;
      aString  : string;
      aResult  : string;
    begin
      with aSystem
      do begin
        aReplace := DataPop;
        aSearch  := DataPop;
        aString  := DataPop;
        aResult  := StringReplace( aString, aSearch, aReplace, [ ]);
        DataPush( aResult);
      end
    end;

    procedure _replacefirsti( aSystem: TKnobsForth);
    // replacefirsti ( "string" "search" "replace" -- "string" ) // replace the first occurance of search in string with replace
    // in a case insensitive manner
    var
      aReplace : string;
      aSearch  : string;
      aString  : string;
      aResult  : string;
    begin
      with aSystem
      do begin
        aReplace := DataPop;
        aSearch  := DataPop;
        aString  := DataPop;
        aResult  := StringReplace( aString, aSearch, aReplace, [ rfIgnoreCase]);
        DataPush( aResult);
      end
    end;

    procedure _questiondup( aSystem: TKnobsForth);
    // ?dup ( false -- false | true -- true true )
    var
      A : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        DataPush( A);
        VariantRelease( A);
        if A <> 0
        then DataPush( A);
      end;
    end;

    procedure _2drop( aSystem: TKnobsForth);
    // 2drop ( n n -- )
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        VariantRelease( V);
        V := DataPop;
        VariantRelease( V);
      end;
    end;

    procedure _over( aSystem: TKnobsForth);
    // over ( n1 n2 -- n1 n2 n1 )
    var
      A, B : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        B := DataPop;
        DataPush( B);
        DataPush( A);
        DataPush( B);
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _2dup( aSystem: TKnobsForth);
    // 2dup ( n1 n2 -- n1 n2 n1 n2 )
    begin
      with aSystem
      do begin
        _over( aSystem);
        _over( aSystem);
      end;
    end;

    procedure _2swap( aSystem: TKnobsForth);
    // 2swap ( n1 n2 n3 n4 -- n3 n4 n1 n2 )
    var
      t1, t2, t3, t4 : Variant;
    begin
      with aSystem
      do begin
        t4 := DataPop;
        t3 := DataPop;
        t2 := DataPop;
        t1 := DataPop;

        DataPush( t3);
        DataPush( t4);
        DataPush( t2);
        DataPush( t1);

        VariantRelease( t1);
        VariantRelease( t2);
        VariantRelease( t3);
        VariantRelease( t4);
      end;
    end;

    procedure _lmshift( aSystem: TKnobsForth);
    // << ( n1 n2 -- n3 )
    // shift n1 left n2 bit
    var
      aCnt : Variant;
      aVal : Variant;
    begin
      with aSystem
      do begin
        aCnt := DataPop;
        aVal := DataPop;
        DataPush( aVal shl aCnt);
        VariantRelease( aCnt);
        VariantRelease( aVal);
      end;
    end;

    procedure _rmshift( aSystem: TKnobsForth);
    // >> ( n1 n2 -- n3 )
    // shift n1 right n2 bit
    var
      aCnt : Variant;
      aVal : Variant;
    begin
      with aSystem
      do begin
        aCnt := DataPop;
        aVal := DataPop;
        DataPush( aVal Shr aCnt);
        VariantRelease( aCnt);
        VariantRelease( aVal);
      end;
    end;

    procedure _2over( aSystem: TKnobsForth);
    // 2over ( n1 n2 n3 n4 -- n1 n2 n3 n4 n1 n2 )
    var
      A, B, C, D : Variant;
    begin
      with aSystem
      do begin
        D := DataPop;
        C := DataPop;
        B := DataPop;
        A := DataPop;
        DataPush( A);
        DataPush( B);
        DataPush( C);
        DataPush( D);
        DataPush( A);
        DataPush( B);
        VariantRelease( A);
        VariantRelease( B);
        VariantRelease( C);
        VariantRelease( D);
      end;
    end;

    procedure _rot( aSystem: TKnobsForth);
    // rot ( n1 n2 n3 -- n2 n3 n1 )
    var
      A, B, C : Variant;
    begin
      with aSystem
      do begin
        C := DataPop;
        B := DataPop;
        A := DataPop;
        DataPush( B);
        DataPush( C);
        DataPush( A);
        VariantRelease( A);
        VariantRelease( B);
        VariantRelease( C);
      end;
    end;

    procedure _plusrot( aSystem: TKnobsForth);
    // +rot ( n1 n2 n3 -- n2 n3 n1 )
    begin
      with aSystem
      do _rot( aSystem);
    end;

    procedure _minusrot( aSystem: TKnobsForth);
    // -rot ( n1 n2 n3 -- n3 n1 n2 )
    var
      A, B, C : Variant;
    begin
      with aSystem
      do begin
        C := DataPop;
        B := DataPop;
        A := DataPop;
        DataPush( C);
        DataPush( A);
        DataPush( B);
        VariantRelease( A);
        VariantRelease( B);
        VariantRelease( C);
      end;
    end;

    procedure _pick( aSystem: TKnobsForth);
    // Pick ( nn .. na .. n0 a -- nn .. na .. n0 na )
    begin
      with aSystem
      do DataPush( DataIndex( DataPop));
    end;

    procedure _put( aSystem: TKnobsForth);
    // Put ( nn .. na .. n0 x a -- nn .. x .. n0 )
    var
      aValue  : Variant;
      anIndex : Variant;
      V       : Variant;
    begin
      with aSystem
      do begin
        aValue  := DataPop;
        anIndex := DataPop;
        V       := Indent( anIndex);
        DataPut( anIndex, aValue);
        VariantRelease( V);
      end;
    end;

    procedure _abs( aSystem: TKnobsForth);
    // abs ( n1 -- n2 )
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPush( Abs( V));
        VariantRelease( V);
      end;
    end;

    procedure _max( aSystem: TKnobsForth);
    // max ( n1 n2 -- max(n1,n2) )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        B := DataPop;
        DataPush( Max( A, B));
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _min( aSystem: TKnobsForth);
    // min ( n1 n2 -- min(n1,n2) )
    var
      A : Variant;
      B : Variant;
    begin
      with aSystem
      do begin
        A := DataPop;
        B := DataPop;
        DataPush( Min( A, B));
        VariantRelease( A);
        VariantRelease( B);
      end;
    end;

    procedure _doubleQuote( aSystem: TKnobsForth);
    // " a string" ( -- <string> ) immediate
    // scan next token quote delimited from the input stream
    // and push it onto the stack as a string.
    var
      S : string;
    begin
      with aSystem
      do begin
        S := FetchToken( '"');
        if Compiling
        then CompileLiteral( S)
        else DataPush( S);
      end;
    end;

    procedure _singleQuote( aSystem: TKnobsForth);
    // '' a string' ( -- <string> ) immediate
    // scan next token quote delimited from the input stream
    // and push it onto the stack as a string.
    // same as " except the closure is ' in this case.
    var
      S : string;
    begin
      with aSystem
      do begin
        S := FetchToken( '''');
        if Compiling
        then CompileLiteral( S)
        else DataPush( S);
      end;
    end;

    procedure _colon( aSystem: TKnobsForth);
    // : ( -- n )
    // start a colon definition
    var
      W : TColon;
    begin
      with aSystem
      do begin
        MustNotBeCompilingFor( ':');
        Compiling     := True;
        W             := TColon( AddSymbol( FetchToken( ' '), TColon));
        FCurrentColon := W;
        FLastDefined  := W; // Delay FinalizeCreation until Semi or SemiNoReturn - but LastDefined needs be set for recursive calls to work
        ControlPush( CColonSemi);
      end;
    end;

    procedure _recurse( aSystem: TKnobsForth);
    // recurse ( -- ) immediate
    // call into self at run time;
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'recurse');
        CompileSymbol( FLastDefined);
      end;
    end;

    procedure _semi( aSystem: TKnobsForth);
    // ; ( x -- ) Immediate
    // terminate a colon definition
    begin
      with aSystem
      do begin
        if Compiling
        then begin
          CheckControlStructure( CColonSemi);
          CompileToken( ';');
          FinalizeCreation( FCurrentColon);
          FCurrentColon := nil;
          Compiling     := False;
        end
        else FIP := -1;
      end;
    end;

    procedure _token( aSystem: TKnobsForth);
    // ( bSep -- <string> )
    // read bSep separated token from input stream
    var
      S : Variant;
    begin
      with aSystem
      do begin
        S := DataPop;
        if Length( S) > 0
        then DataPush( FetchToken( string( S)[ 1], True)) // Uses extended token scan i.e. < ....> stuff
        else DataPush( '');
        VariantRelease( S);
      end;
    end;

    procedure _untoken( aSystem: TKnobsForth);
    // ( <string> bSep -- )
    // write bSep separated token from input stream
    var
      i       : Cardinal;
      S       : Variant;
      T       : Variant;
      aStrings: TStringList;
    begin
      with aSystem
      do begin
        S        := DataPop;
        S        := DataPop + S;
        aStrings := TStringList.Create;
        try
          T := '';
          for i := 1 to Length( S)
          do begin
            case string( S)[ i] of
              ^J : begin aStrings.Add( T); T := ''; end;
              ^M : ;
              else T := T + S[ i];
            end;
          end;
          if T <> ''
          then aStrings.Add( T);
          PushLines( aStrings, 'untoken');
        finally
          aStrings.Free;
        end;
        VariantRelease( S);
        VariantRelease( T);
      end;
    end;

    procedure _readLine( aSystem: TKnobsForth);
    // readline ( -- str )
    // read a line from the input and push it as a string
    var
      aLine : string;
    begin
      with aSystem
      do begin
        aLine := readLine;
        DataPush( aLine);
      end
    end;

    procedure _fmark( aSystem: TKnobsForth);
    // >mark ( -- n )
    // set up a mark to be resolved by >resolve
    // forward jump
    begin
      with aSystem
      do begin
        DataPush( Here);
        CompileLiteral( 0);         // To be resolved by >resolve
      end;
    end;

    procedure _fresolve( aSystem: TKnobsForth);
    // >resolve ( n -- )
    // resolve a >mark
    // forward jump
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        ( FCode.Symbol[ V] as TLiteral).Value := Here;   // Resolve >mark label
        VariantRelease( V);
      end;
    end;

    procedure _bmark( aSystem: TKnobsForth);
    // <mark ( -- n )
    // set up a mark to be resolved by <resolve
    // backward jump
    begin
      with aSystem
      do DataPush( Here);
    end;

    procedure _bresolve( aSystem: TKnobsForth);
    // <resolve ( n -- )
    // resolve a <mark
    // backward jump
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        CompileLiteral( V);
        VariantRelease( V);
      end;
    end;

    procedure _comma( aSystem: TKnobsForth);
    // , ( n -- )
    // add n to the end of the FCode space, as a TLiteral
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        CompileLiteral( V);
        VariantRelease( V);
      end;
    end;

    procedure _allot( aSystem: TKnobsForth);
    // allot ( n -- )
    // claim n cells of FCode space, compile literals into that space
    var
      i : Integer;
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        for i := 0 to V - 1
        do CompileLiteral( 0);
        VariantRelease( V);
      end;
    end;

    procedure _here( aSystem: TKnobsForth);
    // here ( -- n )
    // returns the index of the first free FCode space location
    begin
      with aSystem
      do DataPush( Here);
    end;

    procedure _create( aSystem: TKnobsForth);
    // create XXXX ( -- )
    // define XXXX as a new symbol that ccan have its run-time behaviour changed by does>
    // when XXXX is later executed it pushes it's CFA onto the stack.
    var
      W : TCreate;
    begin
      with aSystem
      do begin
        W     := TCreate( AddSymbol( FetchToken( ' '), TCreate));
        W.CFA := Here;
        CompileLiteral( 0);        // To be resolved by 'does>' it will compile the code index of the code after does>
        FinalizeCreation( W);
      end;
    end;

    procedure _pdoes( aSystem: TKnobsForth);
    // (does) ( -- pfa )
    // changes the last symbol defined with create to make it
    // execute the code that follows inline, it then exits.
    // this also terminates the does> code with a semis
    var
      W     : TCreate;
      aData : TLiteral;
    begin
      with aSystem
      do begin
        if FLastDefined is TCreate                      // Pick up symbol defined by 'Create'
        then begin
          W := TCreate( FLastDefined);
          with W
          do begin
            aData := TLiteral( FCode.Symbol[ CFA]);     // And make the data elt at CFA
            aData.Value := FIP;                         // point to the 'does' part, index of code after 'does>'
          end;
          CompileToken( ';');                           // Add a termination symbol for later cleanup
          FIP := ReturnPop;                             // Perform Semi semantics, to not execute the does> part
        end
        else
          RaiseError( 'LastDefined is not of class "Create", can''t use "(does)" to modify it''s semantics.');
      end;
    end;

    procedure _does( aSystem: TKnobsForth);
    // does> ( -- ) immediate
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'does>');
        CompileToken( '(does)', TInternal);
      end;
    end;

    procedure _ip( aSystem: TKnobsForth);
    // ip ( -- n ) // Push the current FIP value
    begin
      with aSystem
      do DataPush( FIP);
    end;

    procedure _pfafetch( aSystem: TKnobsForth);
    // pfa@ ( pfa -- <data-token> ) // pfa is an index in code space, pfa@ will fetch the TSymbol stored there
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPushSymbol( FCode.Symbol[ V]);
        VariantRelease( V);
      end;
    end;

    procedure _pfafetchfetch( aSystem: TKnobsForth);
    // pfa@@ ( pfa -- data pfa ) // pfa is an index in code space, pfa@@ will fetch the contents from the TSymbol
    //                              stored there and it will the repush the pfa
    // todo : this is not working yet ..
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPushSymbol( FCode.Symbol[ V]);
        _fetch( aSystem);
        DataPush( V);
        VariantRelease( V);
      end;
    end;

    procedure _pfafetchstore( aSystem: TKnobsForth);
    // pfa@! ( pfa data -- pfa ) // pfa is an index in code space, pfa@! will store data into the TSymbol there
    //                              and it will the repush the pfa
    // todo : this is not working yet .. nonsense even
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        DataPushSymbol( FCode.Symbol[ V]);
        _fetch( aSystem);
        DataPush( V);
        VariantRelease( V);
      end;
    end;

    procedure _branch( aSystem: TKnobsForth);
    // branch ( -- )
    // executes a branch
    begin
      with aSystem
      do FIP := ( FCode.Symbol[ FIP] as TLiteral).Value;
    end;

    procedure _fbranch( aSystem: TKnobsForth);
    // fbranch ( -- )
    // executes a conditional branch
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        if Boolean( V)
        then FIP := FIP + 1
        else FIP := ( FCode.Symbol[ FIP] as TLiteral).Value;
        VariantRelease( V);
      end;
    end;

    procedure _crash( aSystem: TKnobsForth);
    // crash ( -- )
    begin
      with aSystem
      do Crash;
    end;

    procedure _immediate( aSystem: TKnobsForth);
    // immediate ( -- )
    begin
      with aSystem
      do Immediate;
    end;

    procedure _sysword( aSystem: TKnobsForth);
    // sysword ( -- )
    begin
      with aSystem
      do SysSymbol;
    end;

    procedure _hide( aSystem: TKnobsForth);
    // hide ( -- )
    begin
      with aSystem
      do Hide;
    end;

    procedure _reveal( aSystem: TKnobsForth);
    // reveal ( -- )
    begin
      with aSystem
      do Reveal;
    end;

    procedure __begin( aSystem: TKnobsForth);
    // _begin ( -- )
    begin
      // Just a marker for debugging
    end;

    procedure __until( aSystem: TKnobsForth);
    // _until ( n -- ), alias for ?branch
    begin
      _fbranch( aSystem);
    end;

    procedure __while( aSystem: TKnobsForth);
    // _while ( n -- ), alias for bBranch
    begin
      _fbranch( aSystem);
    end;

    procedure __again( aSystem: TKnobsForth);
    // _again ( -- ) alias for branch
    begin
      _branch( aSystem);
    end;

    procedure __repeat( aSystem: TKnobsForth);
    // _repeat ( -- ) alias for branch
    begin
      _branch( aSystem);
    end;

    procedure __do( aSystem: TKnobsForth);
    // _do ( ns ne -- ) runtime handler for do
    begin
      with aSystem
      do begin
        ControlPush( DataPop);  // start value
        ControlPush( DataPop);  // end   value
        FIP := FIP + 1;        // Skip forward branch
      end;
    end;

    procedure __qdo( aSystem: TKnobsForth);
    // _?do ( ns ne -- ) runtime handler for ?do
    var
      S, E : Variant;
    begin
      with aSystem
      do begin
        S := DataPop;
        E := DataPop;
        if E = S
        then begin
          _branch( aSystem);
          VariantRelease( S);
          VariantRelease( E);
        end
        else begin
          ControlPush( S);  // start value
          ControlPush( E);  // end   value
          FIP := FIP + 1;  // skip forward branch
        end;
      end;
    end;

    procedure __od( aSystem: TKnobsForth);
    // _od ( -- ) runtime handler for od
    var
      S, E : Integer;
    begin
      with aSystem
      do begin
        E := ControlPop;
        S := ControlPop + 1;
        if S >= E
        then FIP := FIP + 1       // Skip branch
        else begin
          ControlPush( S);
          ControlPush( E);
          _branch( aSystem);      // Backward branch
        end;
      end;
    end;

    procedure __pod( aSystem: TKnobsForth);
    // _pod ( n -- ) runtime handler for +od
    var
      S, E : Variant;
      V    : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        E := ControlPop;
        S := ControlPop + V;
        VariantRelease( V);
        if S < E
        then begin
          ControlPush( S);
          ControlPush( E);
          _branch( aSystem);      // Backward branch
        end
        else begin
          VariantRelease( S);
          VariantRelease( E);
          FIP := FIP + 1;         // Skip branch
        end;
      end;
    end;

    procedure __break( aSystem: TKnobsForth);
    // _break ( -- ) runtime handler for break
    begin
      with aSystem
      do begin
        VariantRelease( ControlPop);
        VariantRelease( ControlPop);
      end;
    end;

    procedure __leave( aSystem: TKnobsForth);
    // _leave ( -- ) runtime handler for leave, non ANSI
    var
      E : Integer;
    begin
      with aSystem
      do begin
        E := ControlPop;           // Make loop terminate
             ControlPop;
        ControlPush( E - 1);
        ControlPush( E);
      end;
    end;

    procedure __if( aSystem: TKnobsForth);
    // _if ( n -- ) runtime handler for if
    begin
      with aSystem
      do _fbranch( aSystem);
    end;

    procedure __else( aSystem: TKnobsForth);
    // _else ( -- ) runtime handler for else
    begin
      with aSystem
      do _branch( aSystem);
    end;

    procedure __fi( aSystem: TKnobsForth);
    // _fi ( -- ) runtime handler for fi
    begin
      // just a marker for debugging
    end;

    procedure __case( aSystem: TKnobsForth);
    // _case ( -- ) runtime handler for case
    begin
      // just a marker for debugging
    end;

    procedure __default( aSystem: TKnobsForth);
    // _default ( -- ) runtime handler for default
    begin
      // just a marker for debugging
    end;

    procedure __esac( aSystem: TKnobsForth);
    // _esac ( -- ) runtime handler for esac
    begin
      // just a marker for debugging
    end;

    procedure __of( aSystem: TKnobsForth);
    // _of ( n1 n2 -- n1 | ) runtime handler for of
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        if V = DataFetch
        then begin
          VariantRelease( DataPop);
          FIP := FIP + 1;
        end
        else _branch( aSystem);
        VariantRelease( V);
      end;
    end;

    procedure __fo( aSystem: TKnobsForth);
    // _fo ( -- ) runtime handler for fo
    begin
      with aSystem
      do _branch( aSystem); // Jump to after esac
    end;

    procedure __in( aSystem: TKnobsForth);
    // _in ( n1 n2 -- n1 | ) runtime handler for in
    type
      TSmallSet = set of 0 .. 63;
      TSetConvertor = record
        case Boolean of
          False : ( int     : UInt64    );
          True  : ( smallset: TSmallSet );
      end;
    var
      aSet : TSetConvertor;
      V    : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        aSet.int := V;
        if Cardinal( DataFetch) in aSet.smallset
        then begin
          VariantRelease( DataPop);
          FIP := FIP + 1;
        end
        else _branch( aSystem);
        VariantRelease( V);
      end;
    end;

    procedure __ni( aSystem: TKnobsForth);
    // _ni ( -- ) runtime handler for ni
    begin
      with aSystem
      do _branch( aSystem); // Jump to after esac
    end;

    procedure __exit( aSystem: TKnobsForth);
    // _exit ( -- ) runtime handler for Exit
    begin
      with aSystem
      do FIP := ReturnPop;
    end;

    procedure _begin( aSystem: TKnobsForth);
    // begin ( -- n n )
    // begin .. again
    // begin .. until
    // begin .. while .. repeat
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'begin');
        CompileToken( '_begin');
        _bmark( aSystem);
        ControlPush( CBeginUntilWhile);
      end;
    end;

    procedure _until( aSystem: TKnobsForth);
    // until ( n n -- )
    // end a begin .. until loop
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'until');
        CheckControlStructure( CBeginUntilWhile);
        CompileToken( '_until');
        _bresolve( aSystem);
      end;
    end;

    procedure _again( aSystem: TKnobsForth);
    // again ( n -- )
    // end a begin .. again loop
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'again');
        CheckControlStructure( CBeginUntilWhile);
        CompileToken( '_again');
        _bresolve( aSystem);
      end;
    end;

    procedure _while( aSystem: TKnobsForth);
    // while ( n -- )
    // begin .. while .. repeat
    var
      aLabel : Variant;
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'while');
        CheckControlStructure( CBeginUntilWhile);
        CompileToken( '_while');

        aLabel := DataPop;
        DataPush( Here);

        DataPush( aLabel);
        CompileLiteral( 0);
        VariantRelease( aLabel);

        ControlPush( CWhileRepeat);
      end;
    end;

    procedure _repeat( aSystem: TKnobsForth);
    // repeat ( n n -- )
    // begin ... while ... repeat
    var
      aLabel : Variant;
      V      : Variant;
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'repeat');
        CheckControlStructure( CWhileRepeat);
        CompileToken( '_repeat');
        V := DataPop;
        CompileLiteral( V);
        VariantRelease( V);
        aLabel := Here;
        V := DataPop;
        ( FCode.Symbol[ V] as TLiteral).Value := aLabel; // Resolve While
        VariantRelease( V);
      end;
    end;

    procedure _do( aSystem: TKnobsForth);
    // do ( -- n )
    // start a do .. od or a do .. +od
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'do');
        CompileToken( '_do');
        _fmark( aSystem);
        _bmark( aSystem);
        ControlPush( CDoOd);
      end;
    end;

    procedure _qdo( aSystem: TKnobsForth);
    // ?do ( -- n )
    // start a ?do .. od or a ?do .. +od
    begin
      with aSystem
      do begin
        MustBeCompilingFor( '?do');
        CompileToken( '_?do');
        _fmark( aSystem);
        _bmark( aSystem);
        ControlPush( CDoOd);
      end;
    end;

    procedure _od( aSystem: TKnobsForth);
    // od ( n -- )
    // end a do .. od
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'od');
        CheckControlStructure( CDoOd);
        CompileToken( '_od');
        _bresolve( aSystem);
        _fresolve( aSystem);
      end;
    end;

    procedure _pod( aSystem: TKnobsForth);
    // +od ( n -- )
    // end a do .. +od
    begin
      with aSystem
      do begin
        MustBeCompilingFor( '+od');
        CheckControlStructure( CDoOd);
        CompileToken( '_pod');
        _bresolve( aSystem);
        _fresolve( aSystem);
      end;
    end;

    procedure _break( aSystem: TKnobsForth);
    // break ( -- )
    // drop loop stack entries
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'break');
        CompileToken( '_break');
      end;
    end;

    procedure _leave( aSystem: TKnobsForth);
    // ( -- )
    // leave the enclosing (?)do .. (+)od when the next
    // od or +od is reached (non ANSI)
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'leave');
        CompileToken( '_leave');
      end;
    end;

    procedure _i( aSystem: TKnobsForth);
    // ( -- n )
    // get innermost (?)do .. (+)od index
    begin
      with aSystem
      do DataPush( FControlStack.Index( 1));
    end;

    procedure _j( aSystem: TKnobsForth);
    // ( -- n )
    // get 2nd (?)do .. (+)od index
    begin
      with aSystem
      do DataPush( FControlStack.Index( 3));
    end;

    procedure _k( aSystem: TKnobsForth);
    // ( -- n )
    // get 3rd (?)do .. (+)od index
    begin
      with aSystem
      do DataPush( FControlStack.Index( 5));
    end;

    procedure _tor( aSystem: TKnobsForth);
    // ( n -- )
    // transfer n to return stack
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        ReturnPush( V);
        VariantRelease( V);
      end;
    end;

    procedure _rto( aSystem: TKnobsForth);
    // ( -- n )
    // get n from return stack
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := ReturnPop;
        DataPush( V);
        VariantRelease( V);
      end;
    end;

    procedure _if( aSystem: TKnobsForth);
    // if ( -- n )
    // start an if [else] fi conditional
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'if');
        CompileToken( '_if');
        _fmark( aSystem);
        ControlPush( CIfElseFi);
      end;
    end;

    procedure _else( aSystem: TKnobsForth);
    // else ( n -- n )
    // start an else part in a if else fi conditional
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'else');
        CheckControlStructure( CIfElseFi);
        CompileToken( '_else');
        _fmark( aSystem);
        DataSwap;
        _fresolve( aSystem);
        ControlPush( CIfElseFi);
      end;
    end;

    procedure _fi( aSystem: TKnobsForth);
    // fi ( n -- )
    // end an if [else] fi conditional
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'fi');
        CheckControlStructure( CIfElseFi);
        CompileToken( '_fi');
        _fresolve( aSystem);
      end;
    end;

    procedure _case( aSystem: TKnobsForth);
    // case ( -- 4 0 )
    // start a case default esac statement
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'case');
        ControlPush( CCaseEsac);
        CompileToken( '_case');
        DataPush( 0);  // End of resolve list marker - for esac
      end;
    end;

    procedure _default( aSystem: TKnobsForth);
    // default ( -- )
    // dummy in case default esac
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'default');
        CompileToken( '_default');
      end;
    end;

    procedure _esac( aSystem: TKnobsForth);
    // esac ( n0 n1 .. nn-1 0 4 -- )
    // Eend a case esac statement
    var
      aLabel : Variant;
      Loc    : Integer;
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'esac');
        CompileToken( '_esac');
        aLabel := Here;

        Loc := DataPop;
        while Loc <> 0
        do begin
          ( FCode.Symbol[ Loc] as TLiteral).Value := aLabel; // Resolve fo labels.
          VariantRelease( Loc);
          Loc := DataPop;
        end;
        VariantRelease( Loc);
        CheckControlStructure( CCaseEsac);
      end;
    end;

    procedure _of( aSystem: TKnobsForth);
    // of ( -- n 5 )
    // start an of section within a case .. esac construct
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'of');
        CompileToken( '_of');
        _fmark( aSystem);
        ControlPush( COfFo);
      end;
    end;

    procedure _fo( aSystem: TKnobsForth);
    // fo ( n1 5 -- n1 n2 )
    // end an of section
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'fo');
        CheckControlStructure( COfFo);
        CompileToken( '_fo');
        _fmark( aSystem);
        DataSwap;
        _fresolve( aSystem);
      end;
    end;

    procedure _in( aSystem: TKnobsForth);
    // in ( -- n 5 )
    // start an in section within a case .. esac construct
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'in');
        CompileToken( '_in');
        _fmark( aSystem);
        ControlPush( CInNi);
      end;
    end;

    procedure _ni( aSystem: TKnobsForth);
    // ni ( n1 5 -- n1 n2 )
    // end an in section
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'ni');
        CheckControlStructure( CInNi);
        CompileToken( '_ni');
        _fmark( aSystem);
        DataSwap;
        _fresolve( aSystem);
      end;
    end;

    procedure _exit( aSystem: TKnobsForth);
    // ( -- )
    // immediately leave the current colon definition returning to the caller
    begin
      with aSystem
      do begin
        MustBeCompilingFor( 'exit');
        CompileToken( '_exit');
      end;
    end;

    procedure _defer( aSystem: TKnobsForth);
    // defer <XXXX> ( -- )
    // make XXXX a symbol without semantics, to be resolved by Is or Implements
    // to later add sematics to it.
    begin
      with aSystem
      do begin
        MustNotBeCompilingFor( 'defer');
        FinalizeCreation( AddSymbol( FetchToken( ' '), TDefer));
      end;
    end;

    procedure _tick( aSystem: TKnobsForth);
    //  '  <XXXX> ( -- <exec-token> )
    // get next token from input stream, search it and leave it's execution
    // token on the stack, error if not found
    var
      W : TSymbol;
    begin
      with aSystem
      do begin
        W := MustFindSymbol( FetchToken( ' '));
        DataPushSymbol( W);
      end;
    end;

    procedure _bracketedtick( aSystem: TKnobsForth);
    // ['] <XXXX> ( -- ) immediate
    // get next token from input stream, search it and compile it's execution
    // token into the current definition. Error if not found, error if not compiling
    begin
      with aSystem
      do begin
        MustBeCompilingFor( '['']');
        CompileToken( FetchToken( ' '));
      end;
    end;

    procedure _LSBracket( aSystem: TKnobsForth);
    // [ ( -- ) immediate
    // turns compiler off, starts interpreter
    begin
      with aSystem
      do Compiling := False;
    end;

    procedure _RSBracket( aSystem: TKnobsForth);
    // ] ( -- )
    // turns interpreter off, compiler on
    begin
      with aSystem
      do Compiling := True;        // Turns compiler ON
    end;

    procedure _Literal( aSystem: TKnobsForth);
    // literal ( n -- ) Immediate
    // compile TOS as a literal, target only
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        CompileLiteral( DataPop);
        VariantRelease( V);
      end;
    end;

    procedure _qcomp( aSystem: TKnobsForth);
    // ?comp ( -- )
    // generate error when not compiling
    begin
      with aSystem
      do begin
        if not Compiling
        then RaiseError( 'Must be compiling ...');
      end;
    end;

    procedure _compilingq( aSystem: TKnobsForth);
    // compiling? ( -- n ) // Return true when compiling
    begin
      with aSystem
      do DataPush( Compiling);
    end;

    procedure _compile( aSystem: TKnobsForth);
    // compile   <XXXX> ( -- )
    // [compile] <XXXX> ( -- ) immediate
    // postpone  <XXXX> ( -- ) immediate
    // get next token from input stream, search it and compile it into
    // the current definition, error if not found
    begin
      with aSystem
      do CompileToken( ' ');
    end;

    procedure _alias( aSystem: TKnobsForth);
    // alias ( -- )
    // define a new (additional) name for an existing symbol
    // Used as 'Alias XXXX YYYY' to make XXXX also known as YYYY
    // or  alias old_name new_name
    var
      aNewName       : string;
      anExistingName : string;
    begin
      with aSystem
      do begin
        MustNotBeCompilingFor( 'alias');
        anExistingName := FetchToken( ' ');
        aNewName       := FetchToken( ' ');
        AddAlias( anExistingName, aNewName);
      end;
    end;

    procedure _is( aSystem: TKnobsForth);
    // is <XXXX>( <exec-token> -- ) immediate
    // used as ' YYYY Is XXXX, used to resolve a deferred symbol
    // to exec-token.
    // when XXXX is not a deferred symbol an error is generated
    var
      aName  : string;
      aDefer : TDefer;
      W      : TSymbol;
    begin
      with aSystem
      do begin
        MustNotBeCompilingFor( 'is');
        aName  := FetchToken( ' ');
        aDefer := TDefer( MustFindToken( aName, TDefer));
        try
          W := DataPopSymbol;
          aDefer.Resolve( W);
          W.ReleaseRef;
        except
          on E: Exception
          do
            aSystem.
              RaiseErrorFmt(
                'Resolve error for ''%s'' with message: %s in "is"',
                [
                  aName,
                  E.Message
                ]
              );
        end;
      end;
    end;

    procedure _lastdefined( aSystem: TKnobsForth);
    // lastDefined ( -- <exec-token> )
    // returns the execution token of the symbol that was last defined
    begin
      with aSystem
      do begin
        MustNotBeCompilingFor( 'lastdefined');
        DataPushSymbol( FLastDefined);
      end;
    end;

    procedure _implements( aSystem: TKnobsForth);
    // implements XXXX ( -- )
    // binds the last defined symbol to the deferred symbol XXXX
    // when XXXX is not a deferred symbol an error is generated.
    begin
      aSystem.MustNotBeCompilingFor( 'implements');
      _lastdefined( aSystem);
      _is         ( aSystem);
    end;

    procedure _execute( aSystem: TKnobsForth);
    // execute ( <exec-token> -- )
    // executes (interprets) exec-token.
    var
      W : TSymbol;
    begin
      with aSystem
      do begin
        W := DataPopSymbol;
        W.Interpret;
        W.ReleaseRef;
      end;
    end;

    procedure _perform( aSystem: TKnobsForth);
    // perform ( var-token -- )
    // @ execute
    begin
      with aSystem
      do begin
        _fetch  ( aSystem);
        _execute( aSystem);
      end;
    end;

    procedure _evaluate( aSystem: TKnobsForth);
    // evaluate ( $ -- )
    // interprets a string
    var
      V : Variant;
    begin
      with aSystem
      do begin
        V := DataPop;
        Evaluate( V);
        VariantRelease( V);
      end;
    end;

    procedure _explain( aSystem: TKnobsForth);
    // explain ( <exec-token> -- )
    // gives help for exec-token
    var
      W : TSymbol;
    begin
      with aSystem
      do begin
        W := DataPopSymbol;
        Explain( W);
        W.ReleaseRef;
      end;
    end;

    procedure _see( aSystem: TKnobsForth);
    // see ( -- ) eats the next token and explains it : see ' explain ;
    begin
      _tick   ( aSystem);
      _explain( aSystem);
    end;

    procedure _help( aSystem: TKnobsForth);
    // .help ( -- ) show very basic help
    const
      // string literals may have at mosst 255 elements
      part1 = 'Use "only also forth definitions" to get into into a reasonable search order, then use ".words" to see all defined symbols in the "forth" and "only" vocabularies.';
      part2 = 'Use "see XXXX" to get help on that particular XXXX; use ".vocs" to see all defined vocabularies.';
      part3 = 'Usually vocs will be defined in the "forth" vocabulary, using "also vocname" will then add a vocabulary "vocname" to the search order.';
      part4 = 'The symbol "definitions" will set the compile vocabulary to the topmost search vocabulary, not removing it from the search order.';
      part5 = 'Use "help" to see this again.';
    begin
      with aSystem
      do begin
        DataPush( part1);
        _dot( aSystem);
        DataPush( part2);
        _dot( aSystem);
        DataPush( part3);
        _dot( aSystem);
        DataPush( part4);
        _dot( aSystem);
        DataPush( part5);
        _dot( aSystem);
      end;
    end;


{ ========
  TStack = class
  private
    FSystem      : TKnobsForth;
    FName        : string;
    FData        : array of Variant;
    FOnUnderflow : TOnStackUnderflow;
  public
    property    Name        : string            read FName;
    property    Depth       : Cardinal          read GetDepth;
    property    OnUnderflow : TOnStackUnderflow read FOnUnderflow write FOnUnderflow;
  private
}

    function    TStack.GetDepth: Cardinal;
    begin
      Result := Length( FData);
    end;

//  public

    destructor  TStack.Destroy; // override;
    begin
      Clear;
      inherited;
    end;

    procedure   TStack.Push( aValue: Variant);
    begin
      if aValue <> Null
      then begin
        if VariantIsSymbol( aValue)
        then VariantToSymbol( aValue).AddRef;
        SetLength( FData, Length( FData) + 1);
        FData[ Length( FData) - 1] := aValue;
      end
      else FSystem.RaiseError( 'Pushing NULL value');
    end;

    procedure   TStack.PushSymbol( const aValue: TSymbol);
    begin
      Push( SymbolToVariant( aValue));
    end;

    function    TStack.Pop: Variant;
    begin
      if Depth > 0
      then begin
        Result := FData[ Length( FData) - 1];
        SetLength( FData, Length( FData) - 1);
      end
      else begin
        if Assigned( FOnUnderflow)
        then FOnUnderflow( Self, FName);
        Result := Null;
      end;
    end;

    function    TStack.PopSymbol: TSymbol;
    var
      V : Variant;
    begin
      Result := nil;
      V      := Pop;
      if VarIsType( V, varByRef)
      then begin
        TVarData( V).VType := varInteger;
        Result := TSymbol( Integer( V));
      end
      else FSystem.RaiseErrorFmt( 'expected to pop an object from "%s"', [ Name]);
    end;

    procedure   TStack.Swap;
    var
      A, B: Variant;
    begin
      A := Pop;
      B := Pop;
      Push( A);
      Push( B);
    end;

    function    TStack.Index( anIndex: Cardinal): Variant;
    begin
      if anIndex < Depth
      then Result := FData[ Depth - anIndex - 1]
    end;

    procedure   TStack.Put( anIndex: Cardinal; aValue: Variant);
    var
      p : Integer;
    begin
      if anIndex < Depth
      then begin
        p := Depth - anIndex + 1;
        FData[ p] := aValue;
      end;
    end;

    function    TStack.Fetch: Variant;
    begin
      Result := Index( 0);
    end;

    procedure   TStack.Clear;
    var
      V : Variant;
    begin
      while Depth > 0
      do begin
        V := Pop;
        VariantRelease( V);
      end;
    end;

    procedure   TStack.Dump( const aStringList: TStringList; anInd: Cardinal); // override;
    var
      i : Cardinal;
    begin
      if Assigned( aStringList)
      then begin
        if Depth = 0
        then aStringList.Add( Format( 'stack %s: empty', [ Name], AppLocale))
        else begin
          aStringList.Add( Format( 'stack %s:', [ Name], AppLocale));
          for i := Depth - 1 downto 0
          do aStringList.Add( DataToStringText( FData[ i], FSystem.Base));
        end;
      end;
    end;

    function    TStack.AsList: TForthList;
    var
      i : Integer;
    begin
      Result := TForthList.Create( FSystem);
      for i := 0 to Depth - 1
      do Result.CompileVariant( Index( i));
    end;


{ ========
  TSymbol = class
  private
    FSystem    : TKnobsForth;
    FRefCount  : Cardinal;
    FName      : string;       // Name of the symbol (can be '')
    FComment   : string;       // Comment for symbol, probably a stack image
    FDocs      : string;       // Additional user generated comment for the symbol
    FPrevious  : TSymbol;      // link to previously defined symbol
    FVocLink   : TVocabulary;  // Link to vocabulary this symbol is in
    FCrash     : Boolean;      // True when execution is forbidden - rases an exception
    FImmediate : Boolean;      // True when executing in compilation mode
    FVisible   : Boolean;      // True when symbol is findable
    FSysSymbol : Boolean;      // True when this is a system symbol
    FIsSemi    : Boolean;      // True when this symbol ends a definition e.g. ';' (and possible friends)
  public
    property    RefCount  : Cardinal read FRefCount;
    property    Name      : string   read FName;
    property    Comment   : string   read FComment   write FComment;
    property    Docs      : string   read FDocs      write FDocs;
    property    Crash     : Boolean  read FCrash     write FCrash;
    property    Immediate : Boolean  read FImmediate write FImmediate;
    property    Visible   : Boolean  read FVisible   write FVisible;
    property    SysSymbol : Boolean  read FSysSymbol write FSysSymbol;
    property    IsSemi    : Boolean  read FIsSemi;
  public
}

    constructor TSymbol.Create( aSystem: TKnobsForth; const aName: string = NoName); // virtual;
    begin
      inherited Create;
      FSystem   := aSystem;
      FName     := aName;
      FPrevious := FSystem.FLastDefined;
      FIsSemi   := SameText( aName, ';');
      Inc( AllSymbolsCounter);
      if SameText( aName, NoName) and Assigned( FSystem.FAnonymous) // Keep track of anonymous symbols
      then begin
        FSystem.FAnonymous.Add( Self);
        Inc( AnonymousCounter);
      end;
      AddRef;
    end;

    destructor  TSymbol.Destroy; // override;
    begin
      if FRefCount <= 0
      then begin
        Dec( AllSymbolsCounter);
        if SameText( Name, NoName)
        then begin
          if Assigned( FSystem.FAnonymous) // Keep track of anonymous symbols
          then begin
            FSystem.FAnonymous.Remove( Self);
            Dec( AnonymousCounter);
          end;
        end
        else begin
          if Assigned( FVocLink)
          then FVocLink.RemoveSymbol( Name);
        end;
        FSystem.FLastDefined := FPrevious;
        inherited;
      end
      else raise Exception.Create( 'symbol freed while ref count > 0');
    end;

    procedure   TSymbol.AddRef;
    begin
      Inc( FRefCount);
    end;

    procedure   TSymbol.ReleaseRef;
    begin
      if FRefCount > 0
      then Dec( FRefCount);
      if FRefCount <= 0
      then Free;
    end;

    procedure   TSymbol.EndCreate; // virtual;
    begin
      // Flags end of creation, not used here, it's for children
    end;

    procedure   TSymbol.CheckCrash;
    begin
      if Crash
      then raise ECrash.CreateFmt( 'the symbol "%s" has it''s crash bit set - this is meant to indicate that the implementation was not finished, as indicated by the "crash" directive used on it', [ Name]);
    end;

    procedure   TSymbol.Interpret; // virtual;
    begin
      CheckCrash;
      FSystem.DataPushSymbol( Self);
    end;

    procedure   TSymbol.Hide; // virtual;
    begin
      Visible := False;
    end;

    procedure   TSymbol.Reveal; // virtual;
    begin
      Visible := True;
    end;

    function    TSymbol.See: string; // virtual;
    begin
      if Docs = ''
      then Result := Format( '%s %s // %s'     , [ LowerCase( TypeName), Name, Comment      ], AppLocale)
      else Result := Format( '%s %s // %s | %s', [ LowerCase( TypeName), Name, Comment, Docs], AppLocale);
      Result := AppendFlags( Result);
    end;

    function    TSymbol.AppendFlags( const S: string): string; // virtual;
    begin
      Result := S;
      if Crash
      then Result := Result + ' crash';
      if Immediate
      then Result := Result + ' immediate';
      if SysSymbol
      then Result := Result + ' sysword';
      if not Visible
      then Result := Result + ' hide';
    end;

    procedure   TSymbol.AddDocs( const aValue: string); // virtual;
    begin
      if FDocs = ''
      then FDocs := aValue
      else FDocs := FDocs + ' | ' + aValue;
    end;

    procedure   TSymbol.Dump( const aStringList: TStringList; anInd: Cardinal); // virtual;
    var
      aPrefix : string;
    begin
      aPrefix := Indent( anInd);
      with aStringList
      do begin
        Add( Format( '%sType         = %s', [ aPrefix, TypeName], AppLocale));
        DumpDetails  ( aStringList, anInd + 1);
        Add( aPrefix + 'End');
      end;
    end;

    procedure   TSymbol.DumpLimited( const aStringList: TStringList; anInd: Cardinal); // virtual;
    var
      aPrefix : string;
    begin
      aPrefix := Indent( anInd);
      with aStringList
      do begin
        Add( Format( '%sType = %s'            , [ aPrefix, TypeName], AppLocale));
        Add( Format( '%s  Name         = ''%s''', [ aPrefix, Name    ], AppLocale));
        Add( Format( '%s  RefCount     = %d'    , [ aPrefix, RefCount], AppLocale));
        Add( Format( '%sEnd'                  , [ aPrefix          ], AppLocale));
      end;
    end;

    procedure   TSymbol.DumpDetails( const aStringList: TStringList; anInd: Cardinal); // virtual;
    var
      aPrefix : string;
    begin
      aPrefix := Indent( anInd);
      with aStringList
      do begin
        Add( Format( '%sName         = ''%s''', [ aPrefix, Name    ], AppLocale));
        Add( Format( '%sRefCount     = %d'    , [ aPrefix, RefCount], AppLocale));
      end;
    end;

    function    TSymbol.AsList: TForthList; // virtual;
    begin
      Result := TForthList.Create( FSystem);
      Result.CompileVariant( Name);
    end;

//  public

    class function TSymbol.TypeName   : string;
    begin
      Result := Copy( ClassName, 2, Length( ClassName) - 1);
    end;


{ ========
  TSymbols = class( TList)
  private
    FIndex  : TStringList;
    FSystem : TKnobsForth;
  public
    property    Symbol[ anIndex: TCFA]: TSymbol read GetSymbol; default;
  private
}

    function    TSymbols.GetSymbol( anIndex: TCFA): TSymbol;
    begin
      Result := TSymbol( Items[ anIndex]);
    end;

//  public

    constructor TSymbols.Create( aSystem: TKnobsForth);
    begin
      inherited Create;
      FSystem := aSystem;
      FIndex  := TStringList.Create;
      with FIndex
      do begin
        Sorted        := True;
        CaseSensitive := True;      // For faster compares
        Duplicates    := dupAccept;
      end;
    end;

    destructor  TSymbols.Destroy; // override;
    begin
      FreeAndNil( FIndex); // The index is owned and must be disposed off, the
                           //   objects in the index are not owned though
      inherited Destroy;   // This will call Clear whilst FIndex doesn't exist
                           //   anymore. Also see [ TSymbols.Clear ]
    end;

    procedure   TSymbols.Clear; //                                                 override;
    var
      W : TSymbol;
      i : Integer;
    begin
      if Assigned( FIndex) // Destroy calls clear, but kills FIndex first.
      then FIndex.Clear;   // Removes all strings from the index, frees no
                           // objects. All the symbols are owned and should be
                           // disposed off. Also see [ TSymbols.Destroy ]
      for i := Count - 1 downto 0
      do begin
        W := Symbol[ i];
        Delete( i);
        W.ReleaseRef;
      end;
      inherited;
    end;

    function    TSymbols.AddSymbol( const aName: string; aSymbolClass: TSymbolClass; const aVoc: TVocabulary): TSymbol; // overload;
    begin
      if SameText( aName, NoName)
      then FSystem.RaiseError( 'Not allowwed to use AddSymbol for anonymous symbols');
      Result := AddSymbol( aSymbolClass.Create( FSystem, aName), aVoc);
    end;

    function    TSymbols.AddSymbol( aSymbol: TSymbol; const aVoc: TVocabulary): TSymbol; // overload;
    begin
      if Assigned( aSymbol)
      then begin
        Add( aSymbol);
        aSymbol.FVocLink := aVoc;
        FIndex.AddObject( LowerCase( aSymbol.Name), aSymbol);
      end
      else FSystem.RaiseError( 'Compiler error, can''t add <nil> to a symbol list.');
      Result := aSymbol;
    end;

    function    TSymbols.RemoveSymbol( const aName: string): TSymbol;
    var
      i : Integer;
    begin
      i := FIndex.IndexOf( LowerCase( aName));
      if i >= 0
      then begin
        Result := TSymbol( FIndex.Objects[ i]); // Remove symbol name from index
        FIndex.Delete( i);
        Remove( Result);                        // Remove symbol from self
      end
      else Result := nil;
    end;

    function    TSymbols.FindToken( const aToken: string): TSymbol;
    // Make sure to find the latest definition
    var
      i : Integer;
      L : string;
      W : TSymbol;
    begin
      L := LowerCase( aToken);
      i := FIndex.IndexOf( L);
      if i >= 0
      then begin
        Result := TSymbol( FIndex.Objects[ i]);
        if not Result.Visible
        then begin
          Result := nil;
          for i := 0 to Count - 1
          do begin
            W := TSymbol( FIndex.Objects[ i]);
            if W.Visible
            then begin
              if SameText( W.Name, aToken)
              then begin
                Result := W;
                Break;
              end;
            end;
          end;
        end;
      end
      else Result := nil;
    end;

    function    TSymbols.CfaToSymbol( aCFA: TCFA): TSymbol;
    var
      i : Integer;
      W : TSymbol;
    begin
      Result := nil;
      for i := 0 to Count - 1
      do begin
        W := Symbol[ i];
        if ( W is TColon) and ( TColon( W).CFA = aCFA)
        then begin
          Result := W;
          Break;
        end;
      end;
    end;

    function    TSymbols.Symbols: string;
    var
      i : Integer;
    begin
      Result := '';
      for i := 0 to Count - 1
      do begin
        if not Symbol[ i].SysSymbol
        then begin
          if Result = ''
          then Result := Symbol[ i].Name
          else Result := Result + ' ' + Symbol[ i].Name;
        end;
      end;
    end;

    procedure   TSymbols.Dump( const aStringList: TStringList; anInd: Cardinal);
    var
      i : Integer;
      W : TSymbol;
    begin
      for i := 0 to Count - 1
      do begin
        W := Symbol[ i];
        if W is TVocabulary
        then W.DumpLimited( aStringList, anInd)
        else W.Dump       ( aStringList, anInd);
      end;
    end;


{ ========
  TForthList = class( TSymbol)
  strict private
    FValue    : TList;
  public
    property    Count                  : Integer read GetCount  write SetCount;
    property    Symbol[ anIndex: TCFA] : TSymbol read GetSymbol;
  strict private
}

    function    TForthList.GetCount: Integer;
    begin
      Result := FValue.Count;
    end;

    procedure   TForthList.SetCount( aValue: Integer);
    begin
      FValue.Count := aValue;
    end;

    function    TForthList.GetSymbol( anIndex: TCFA): TSymbol;
    begin
      Result := FValue[ anIndex];
    end;

//  public

    constructor TForthList.Create( aSystem: TKnobsForth; const aName: string = NoName); // override;
    begin
      inherited;
      FValue := TList.Create;
    end;

    destructor  TForthList.Destroy; // override;
    var
      i : TCFA;
      W : TSymbol;
    begin
      for i := Count - 1 downto 0
      do begin
        W := Symbol[ i];
        if W.RefCount > 0
        then W.ReleaseRef;
      end;
      FreeAndNil( FValue);
      inherited;
    end;

    procedure   TForthList.ReleaseRef; // override;
    begin
      inherited;
    end;

    function    TForthList.Copy: TForthList;
    var
      i : TCFA;
      W : TSymbol;
    begin
      Result := TForthList.Create( FSystem);
      for i := 0 to Count - 1
      do begin
        W := Symbol[ i];
        Result.CompileSymbol( W);
      end;
    end;

    procedure   TForthList.Clear;
    begin
      FValue.Clear;
    end;

    procedure   TForthList.Interpret; // override;
    begin
      CheckCrash;
      FSystem.DataPushSymbol( Self);
    end;

    procedure   TForthList.Call;
    // This should behave like the joy i combinator, or like the factor call combinator
    // or like forth docol for that matter
    begin
      with FSystem
      do Call( Self, 0);
    end;

    function    TForthList.See: string; // override;
    var
      i : TCFA;
    begin
      Result := '{';
      for i := 0 to Count - 1
      do Result := Result + ' ' + Symbol[ i].See;
      Result := Result + ' }';
    end;

    procedure   TForthList.Append( aValue: TSymbol);
    begin
      CompileSymbol( aValue);
    end;

    procedure   TForthList.AcceptToken( const aToken: string);
    var
      W       : TSymbol;
      aNumber : Variant;
    begin
      try
        W := nil;
        // Search for global definitions
        if not Assigned( W)
        then W := FSystem.FindToken( aToken);
        // When found let the symbol be interpreted or compiled depending
        // on immediateness  { and --> NOT <-- the current compile state }.
        if Assigned( W)
        then begin
          if W.Immediate { or not FSystem.Compiling }
          then begin
            W.Interpret;
            CompileVariant( FSystem.DataPop);
          end
          else CompileSymbol( W);
        end
        // Otherwise see if the token is a number
        else if FSystem.Number( aToken, aNumber)
        then CompileLiteral( aNumber)                         // Compile number literal
        // None worked, it's an error then.
        else FSystem.RaiseErrorFmt( '%s ?', [ aToken]);
      except
        // All errors are trapped here.
        on E: Exception
        do FSystem.Say( E.Message);
      end;
    end;

    procedure   TForthList.CompileVariant( const aValue: Variant);
    var
      W : TSymbol;
    begin
      if VariantIsSymbol( aValue)
      then begin
        // This makes nested lists work ok, or so it seems
        W := VariantToSymbol( aValue);
        CompileSymbol( W);
        W.ReleaseRef;
      end
      else begin
        // Not sure about this bit ...
        CompileLiteral( aValue);
        VariantRelease( aValue);
      end;
    end;

    procedure   TForthList.CompileLiteral( const aValue: Variant);
    var
      aData : TLiteral;
    begin
      aData := TLiteral.Create( FSystem);
      aData.Value := aValue;
      CompileSymbol( aData);
      aData.ReleaseRef;
    end;

    procedure   TForthList.CompileSymbol( const aSymbol: TSymbol);
    begin
      FValue.Add( aSymbol);
      aSymbol.AddRef;
    end;

    procedure   TForthList.CompileToken( const aName: string); // overload;
    begin
      CompileToken( aName, TSymbol);
    end;

    procedure   TForthList.CompileToken( const aName: string; const aSymbolClass: TSymbolClass); // overload;
    begin
      CompileSymbol( FSystem.MustFindToken( aName, aSymbolClass));
    end;

    function    TForthList.Here: TCFA;
    begin
      Result := Count;
    end;

    procedure   TForthList.SetHere( aValue: TCFA);
    begin
      Count := aValue;
    end;

    procedure   TForthList.Reverse;
    var
      Cnt : TCFA;
      i   : TCFA;
      T   : TSymbol;
    begin
      Cnt := Count div 2;
      for i := 0 to Cnt - 1
      do begin
        T                      := FValue[ i];
        FValue[ i]             := FValue[ Count - i - 1];
        FValue[ Count - i - 1] := T
      end;
    end;

    function    TForthList.First: TSymbol;
    begin
      Result := nil;
      if Count > 0
      then Result := Symbol[ 0]
      else FSystem.RaiseError( 'can''t get first on an empty list');
    end;

    procedure   TForthList.Rest;
    begin
      FValue.Delete( 0);
    end;

    procedure   TForthList.Concat( const aValue: TForthList);
    var
      i : TCFA;
    begin
      for i := 0 to aValue.Count
      do CompileSymbol( aValue.Symbol[ i]);
    end;

    procedure   TForthList.Dump( const aStringList: TStringList; anInd: Cardinal); // override;
    var
      i : TCFA;
      W : TSymbol;
    begin
      with FSystem
      do begin
        if Count = 0
        then aStringList.Add( Indent( anInd + 1) + '<empty>')
        else
          for i := 0 to Count - 1
          do begin
            W := CfaToSymbol( i);
            if Assigned( W)
            then begin
              aStringList.Add( ''); // Separator between definitions
              aStringList.Add(
                Indent( anInd + 1) +
                Format(
                  '%s [%d]',
                  [
                    W.Name,
                    W.RefCount
                  ],
                  AppLocale
                )
            );
            end;
            W := Symbol[ i];
            if Assigned( W)
            then
              with W
              do begin
                aStringList.Add(
                  Format(
                    '%s%.4d %s %s (%s) [%d]',
                    [
                      Indent( anInd + 1),
                      i,
                      TrimStr( Name    , 32),
                      TrimStr( TypeName, 15),
                      See,
                      RefCount
                    ],
                    AppLocale
                  )
                );
              end;
          end;
        aStringList.Add( '');
      end;
    end;

    function    TForthList.AsList: TForthList; // override;
    begin
      Result := Self;
    end;


{ ========
  TScale = class( TConstant)
  private
    FScaleData: TMidiFilter;
  public
    property ScaleData: TMidiFilter read FScaleData write FScaleData;
  protected
}

    function    TScale.AsString: string;
    var
      i : TMidiSeptet;
    begin
      Result := '{ ';
      for i := Low( TMidiSeptet) to High( TMidiSeptet)
      do begin
        if i in ScaleData
        then Result := Result + Format( '%d ', [ i], AppLocale);
      end;
      Result := Result + '}';
      Result := AppendFlags( Result);
    end;

//  public

    constructor TScale.Create( aSystem: TKnobsForth; const aName: string = NoName); // override;
    begin
      inherited;
      FScaleData := [];
    end;

    procedure   TScale.Interpret; // override;
    begin
      CheckCrash;
      if Assigned( FSystem)
      then FSystem.AddScale( Name, ScaleData)
    end;

    function    TScale.See: string; // override;
    begin
      Result := Format( '%s scale %s', [ AsString, Name], AppLocale);
    end;

    procedure   TScale.AddItem( anItem: TMidiSeptet);
    begin
      ScaleData := ScaleData + [ anItem];
    end;

    procedure   TScale.AddList( const aValue: Variant);
    var
      W     : TSymbol;
      aList : TForthList;
      anElt : TSymbol;
      anInt : Variant;
      i     : Integer;
    begin
      if VariantIsSymbol( aValue)
      then begin
        W := VariantToSymbol( aValue);
        if W is TForthList
        then begin
          aList := W as TForthList;
          for i := 0 to aList.Count - 1
          do begin
            anElt := aList.Symbol[ i];
            anElt.Interpret;
            anInt := FSystem.DataPop;
            if ( anInt >= Low( TMidiSeptet)) and ( anInt <= High( TMidiSeptet))
            then begin
              AddItem( anInt);
              VariantRelease( anInt);
            end
            else begin
              VariantRelease( anInt);
              FSystem.RaiseError( 'scales expect a list object with items that evaluate to valid MIDI data [$00,$7f] to be on the data stack');
            end;
          end;
        end
        else FSystem.RaiseError( 'scales expect a list object to be on the data stack');
      end
      else FSystem.RaiseError( 'scales expect an object to be on the data stack');
    end;


{ ========
  TLiteral = class( TSymbol)
  private
    FValue : Variant;
  public
    property    Value: Variant read FValue write SetValue;
  private
}

    procedure   TLiteral.SetValue( const aValue: Variant);
    begin
      if VariantIsSymbol( FValue)
      then VariantToSymbol( FValue).ReleaseRef;
      FValue := aValue;
      if VariantIsSymbol( aValue)
      then VariantToSymbol( aValue).AddRef;
    end;

//  public

    procedure   TLiteral.Interpret; // override;
    begin
      CheckCrash;
      FSystem.DataPush( FValue);
    end;

    function    TLiteral.See: string; // override;
    begin
      Result := Format( '%s', [ DataToString( FValue, FSystem.Base)], AppLocale);
    end;

    procedure   TLiteral.DumpDetails( const aStringList: TStringList; anInd: Cardinal); // override;
    begin
      inherited DumpDetails( aStringList, anInd);
      aStringList.Add( Indent( anInd) + Format( 'Value        = %s', [ Value], AppLocale));
    end;

    function    TLiteral.AsList: TForthList; // override;
    begin
      Result := TForthList.Create( FSystem);
      Result.CompileVariant( FValue);
    end;

{ ========
  TConstant = class( TData)
  public
}

    procedure   TConstant.Interpret; // override;
    begin
      CheckCrash;
      with FSystem
      do DataPush( Value);
    end;

    function    TConstant.See: string; // override;
    begin
      Result := Format( '%s constant %s', [ FValue, Name], AppLocale);
      Result := AppendFlags( Result);
    end;


{ ========
  TVariable = class( TConstant)
  public
}

    procedure   TVariable.Interpret; // override;
    begin
      CheckCrash;
      FSystem.DataPushSymbol( Self);
    end;

    function    TVariable.See: string; // override;
    begin
      Result := Format( 'variable %s', [ Name], AppLocale);
      Result := AppendFlags( Result);
    end;


{ ========
  TVocabulary = class( TSymbol)
  private
    FSymbols : TSymbols; // Link to a symbol list
  public
    property    Symbols : TSymbols read FSymbols;
  public
}

    constructor TVocabulary.Create( aSystem: TKnobsForth; const aName: string = NoName); // override;
    begin
      inherited;
      FSymbols  := TSymbols.Create( aSystem);
      Comment := '( -- ) // a vocabulary, at run time replaces the top of the context stack with itself';
    end;

    destructor  TVocabulary.Destroy; // override;
    begin
      FreeAndNil( FSymbols); // The symbol list is owned and must be disposed off
      inherited;
    end;

    procedure   TVocabulary.Interpret; // override;
    begin
      CheckCrash;
      FSystem.Context := Self;
    end;

//  public

    function    TVocabulary.AddSymbol( const aName: string; aSymbolClass: TSymbolClass): TSymbol; // overload;
    begin
      if Assigned( FSymbols)
      then Result := FSymbols.AddSymbol( aName, aSymbolClass, Self)
      else Result := nil;
    end;

    function    TVocabulary.AddSymbol( aSymbol: TSymbol): TSymbol; // overload;
    begin
      if Assigned( FSymbols)
      then Result := FSymbols.AddSymbol( aSymbol, Self)
      else Result := nil;
    end;

    function    TVocabulary.RemoveSymbol( const aName: string): TSymbol;
    begin
      if Assigned( FSymbols)
      then Result := FSymbols.RemoveSymbol( aName)
      else Result := nil;
    end;

    function    TVocabulary.FindToken( const aToken: string): TSymbol;
    begin
      if Assigned( FSymbols)
      then Result := FSymbols.FindToken( aToken)
      else Result := nil;
    end;

    function    TVocabulary.CfaToSymbol( aCFA: TCFA): TSymbol;
    begin
      if Assigned( FSymbols)
      then Result := FSymbols.CfaToSymbol( aCFA)
      else Result := nil;
    end;

    function    TVocabulary.SymbolList: string;
    begin
      Result := FSymbols.Symbols;
    end;

    procedure   TVocabulary.DumpDetails( const aStringList: TStringList; anInd: Cardinal); // override;
    var
      aPrefix : string;
    begin
      inherited;
      aPrefix := Indent( anInd);
      with aStringList
      do begin
        Add( aPrefix + 'Symbols');
        if Assigned( FSymbols)
        then FSymbols.Dump( aStringList, anInd + 1);
        Add( aPrefix + 'End');
      end;
    end;


{ ========
  TAlias = class( TSymbol)
  private
    FAliased : TSymbol;
  public
    property    Aliased: TSymbol read FAliased write FAliased;
  public
}

    procedure   TAlias.Interpret; // override;
    begin
      CheckCrash;
      if Assigned( Aliased)
      then begin
        try
          Aliased.Interpret;
        except
          on E: Exception
          do
            FSystem.
              RaiseErrorFmt(
                '"%s" is "%s", generated error "%s" during interpretation',
                [
                  Name,
                  Aliased.Name,
                  E.Message
                ]
              );
        end;
      end
      else FSystem.RaiseErrorFmt( '''%s'' is "%s", can not interpret that', [ Name, Nothing]);
    end;

    function    TAlias.See: string; // override;
    begin
      Result := Format( 'alias %s %s', [ Aliased.Name, Name], AppLocale);
      Result := AppendFlags( Result);
    end;

    procedure   TAlias.DumpDetails( const aStringList: TStringList; anInd: Cardinal); // override;
    var
      aPrefix: string;
    begin
      inherited;
      aPrefix := Indent( anInd);
      with aStringList
      do begin
        if Assigned( Aliased)
        then Add( Format( '%salias        = ''%s''', [ aPrefix, Aliased.Name], AppLocale))
        else Add(         '%salias        = ''<nothing>''');
      end;
    end;


{ ========
  TDefer = class( TAlias)
  end;
}

    procedure   TDefer.Resolve( aSymbol: TSymbol);
    begin
      if Assigned( FAliased)
      then FAliased.ReleaseRef;
      Aliased := aSymbol;
      Aliased.AddRef;
    end;

    function    TDefer.See: string; // override;
    begin
      Result := Format( 'defer %s '' %s is %s', [ Name, Aliased.Name, Name], AppLocale);
      Result := AppendFlags( Result);
    end;

    procedure   TDefer.DumpDetails( const aStringList: TStringList; anInd: Cardinal); // override;
    var
      aPrefix: string;
    begin
      inherited;
      aPrefix := Indent( anInd);
      with aStringList
      do begin
        if Assigned( Aliased)
        then Add( Format( '%sis           = ''%s''', [ aPrefix, Aliased.Name], AppLocale))
        else Add(         '%sis           = ''<nothing>''');
      end;
    end;


{ ========
  TExecutable = class( TSymbol)
  public
}

{
  TInternalProc = procedure( aSystem: TSystem);

  TInternal = class( TExecutable)
  private
    FInternalProc: TInternalProc;
  public
    property    InternalProc: TInternalProc read FInternalProc write FInternalProc;
  public
}

    procedure   TInternal.Interpret; // override;
    begin
      CheckCrash;
      InternalProc( FSystem);
    end;


{ ========
  TColon = class( TExecutable)
  public
    FCFA : TCFA;     // Location in code space
  public
    property    CFA : TCFA read FCFA write FCFA;
  public
}

    destructor  TColon.Destroy; // override;
    var
      IP : TCFA;
      W  : TSymbol;
    begin
      with FSystem
      do begin
        IP := CFA;
        W  := FCode.Symbol[ IP];
        while Assigned( W)
        do begin
          if W <> Self
          then W.ReleaseRef;
          if W.IsSemi
          then Break
          else begin
            Inc( IP);
            W := FCode.Symbol[ IP];
          end;
        end;
        FSystem.SetHere( CFA);
      end;
      inherited;
    end;

    procedure   TColon.ReleaseRef; // override;
    var
      IP : TCFA;
      W  : TSymbol;
    begin
      with FSystem
      do begin
        IP := CFA;
        W  := FCode.Symbol[ IP];
        while Assigned( W)
        do begin
          if W.IsSemi
          then Break
          else begin
            if W = Self
            then begin
              if FRefCount > 0
              then Dec( FRefCount);
            end;
            Inc( IP);
            W := FCode.Symbol[ IP];
          end;
        end;
      end;
      inherited;
    end;

    procedure   TColon.Interpret; // override;
    begin
      CheckCrash;
      with FSystem
      do Call( FCode, CFA);
    end;

    function    TColon.See: string; // override;
    var
      IP   : TCFA;
      W    : TSymbol;
      Val  : Variant;
      Skip : TCFA;
    begin
      Result := ': ' + Name + ' ';
      if Docs = ''
      then Result := Result + Format( '( %s ) '       , [ Comment      ], AppLocale)
      else Result := Result + Format( '( %s ) ( %s ) ', [ Comment, Docs], AppLocale);
      with FSystem
      do begin
        IP := CFA;
        W  := FCode.Symbol[ IP];
        while Assigned( W) and not W.IsSemi
        do begin
          Skip := 0;
          if W = Self
          then Result := Result + 'recurse'
          else if W is TLiteral and not ( W is TConstant)
          then begin
            Val := ( W as TLiteral).Value;
            if VariantIsString( Val)
            then Result := Result + Format( '" %s"', [ QuoteEscape( string( Val))], AppLocale)
            else Result := Result + Format( '%s'   , [ DataToString( Val, Base)  ], AppLocale)
          end
          else if ( W is TInternal) and ( Length( W.Name) > 0) and ( W.Name[ 1] = '_')
          then begin
            if SameText( W.Name, '_begin')
            then begin
              Result := Result + 'begin';             Skip := 0;
            end
            else if SameText( W.Name,'_until')
            then begin
              Result := Result + 'until';             Skip := 1;
            end
            else if SameText( W.Name,'_while')
            then begin
              Result := Result + 'while';             Skip := 1;
            end
            else if SameText( W.Name,'_again')
            then begin
              Result := Result + 'again';             Skip := 1;
            end
            else if SameText( W.Name,'_repeat')
            then begin
              Result := Result + 'repeat';            Skip := 1;
            end
            else if SameText( W.Name,'_do')
            then begin
              Result := Result + 'do';                Skip := 1;
            end
            else if SameText( W.Name,'_?do')
            then begin
              Result := Result + '?do';               Skip := 1;
            end
            else if SameText( W.Name,'_od')
            then begin
              Result := Result + 'od';                Skip := 1;
            end
            else if SameText( W.Name,'_break')
            then begin
              Result := Result + 'break';             Skip := 0;
            end
            else if SameText( W.Name,'_pod')
            then begin
              Result := Result + '+od';               Skip := 1;
            end
            else if SameText( W.Name,'_leave')
            then begin
              Result := Result + 'leave';             Skip := 0;
            end
            else if SameText( W.Name,'_if')
            then begin
              Result := Result + 'if';                Skip := 1;
            end
            else if SameText( W.Name,'_else')
            then begin
              Result := Result + 'else';              Skip := 1;
            end
            else if SameText( W.Name,'_fi')
            then begin
              Result := Result + 'fi';                Skip := 0;
            end
            else if SameText( W.Name,'_case')
            then begin
              Result := Result + 'case';              Skip := 0;
            end
            else if SameText( W.Name,'_default')
            then begin
              Result := Result + 'default';           Skip := 0;
            end
            else if SameText( W.Name,'_esac')
            then begin
              Result := Result + 'esac';              Skip := 0;
            end
            else if SameText( W.Name,'_of')
            then begin
              Result := Result + 'of';                Skip := 1;
            end
            else if SameText( W.Name,'_fo')
            then begin
              Result := Result + 'fo';                Skip := 1;
            end
            else if SameText( W.Name,'_in')
            then begin
              Result := Result + 'in';                Skip := 1;
            end
            else if SameText( W.Name,'_ni')
            then begin
              Result := Result + 'ni';                Skip := 1;
            end
            else if SameText( W.Name,'_exit')
            then begin
              Result := Result + 'exit';              Skip := 0;
            end
            else Result := Result + W.Name;
          end
          else Result := Result + W.Name;
          Result := Result + ' ';
          IP := IP + 1 + Skip;
          W  := FCode.Symbol[ IP];
        end;
        Result := Result + ';';
      end;
      Result := AppendFlags( Result);
    end;

    procedure   TColon.DumpDetails( const aStringList: TStringList; anInd: Cardinal); // override;
    begin
      inherited DumpDetails( aStringList, anInd);
    end;

    function    TColon.AsList: TForthList; // override;
    var
      IP : TCFA;
      W  : TSymbol;
    begin
      Result := TForthList.Create( FSystem);
      IP := CFA;
      with FSystem
      do begin
        W  := FCode.Symbol[ IP];
        while Assigned( W) and not W.IsSemi
        do begin
          Result.Append( W);
          Inc( IP);
          W := FCode.Symbol[ IP];
        end;
      end;
    end;

    procedure   TColon.DumpLimited( const aStringList: TStringList; anInd: Cardinal); // override;
    var
      aPrefix : string;
    begin
      aPrefix := Indent( anInd);
      with aStringList
      do begin
        Add( Format( '%sName         = ''%s''', [ aPrefix, Name    ], AppLocale));
        Add( Format( '%sCFA          = %d'    , [ aPrefix, CFA     ], AppLocale));
        Add( Format( '%sRefCount     = %d'    , [ aPrefix, RefCount], AppLocale));
      end;
    end;

{ ========
  TCreate = class( TColon)
  public
}

    procedure   TCreate.Interpret; // override;
    begin
      CheckCrash;
      with FSystem
      do begin
        DataPush( CFA + 1);                                         // Set up data index as CFA + 1, pints to the second TLiteral
        Call( FCode, ( FCode.Symbol[ Self.CFA] as TLiteral).Value); // Load IP from the TLiteral at CFA
      end;
    end;


{ ========
  TKnobsForth = class( TComponent)
  private
    FOnSay            : TOnForthSay;
    FOnBye            : TOnBye;
    FOnBreak          : TOnForthBreak;
    FOnLog            : TOnForthLog;
    FOnReset          : TNotifyEvent;
    FOnUnknownSymbol  : TOnUnknownSymbol;
    FOnDataOut        : TOnDataOut;
    FOnAddScale       : TOnAddScale;
    FOnGetScalenames  : TOnGetScaleNames;
    FOnClearScales    : TOnClearScales;
    FOnClearScreen    : TNotifyEvent;
  private
    FOnLine           : TOnLine;              // Callback for line counter        - must be set by user App.
    FOnFileEnter      : TOnFileEnter;         // Callback for file counter        - must be set by user App.
    FOnFileLeave      : TOnFileLeave;         // Callback for file counter        - must be set by user App.
    FOnFileResume     : TOnFileResume;        // Callback for file counter        - must be set by user App.
  private
    FVocs             : array of TVocabulary; // List of all vocabularies
    FOnly             : TVocabulary;          // Startup symbol list Selected with only (pseudo Voc)
    FCurrent          : TVocabulary;          // Compile voc
    FContext          : array of TVocabulary; // Search voc (chain)
    FDataStack        : TStack;               // Stack for data transfer, mostly just called 'the stack'
    FReturnStack      : TStack;               // Stack used for keeping track of proper nesting of symbol invocations
    FControlStack     : TStack;               // Stack used for keeping track of proper nesting of control structures
    FCode             : TForthList;           // Compiled forth code
    FAnonymous        : TObjectList;          // List of all anonymous symbolss (so they can be freed)
    FFileStack        : TFileStack;           // Stack of source files
    FFiles            : array of string;
  private
    FCompiling        : Cardinal;             // We are compiling    when FCompiling    > 0
    FInterpreting     : Cardinal;             // We are interpreting when FInterpreting > 0 :: this is text interpreting
    FYieldCounter     : Integer;
    FIP               : TCFA;                 // Forth instruction pointer
    FW                : TSymbol;              // Forth code symbol - this is the symbol pointed to by FIP
    FLastDefined      : TSymbol;              // The last defined symbol - only valid after creation completed
    FCurrentColon     : TColon;               // The colon definition currently under construction - nil if no colon symbol is being defined
  private
    FStartupFile      : string;               // name for library code to load at reset
  public
    property    Base            : Cardinal         read GetBase          write SetBase;
    property    Compiling       : Boolean          read GetCompiling     write SetCompiling;
    property    Interpreting    : Boolean          read GetInterpreting  write SetInterpreting;
    property    SourceName      : string           read GetSourceName;
    property    SourceLineNr    : Integer          read GetSourceLineNr;
    property    SourceOffset    : Integer          read GetSourceOffset;
    property    Current         : TVocabulary      read FCurrent         write FCurrent;    // Compile Voc
    property    Context         : TVocabulary      read GetContext       write SetContext;  // Search  Voc
  published
    property    StartupFile     : string           read FStartupFile     write FStartupFile;
  published
    property    OnSay           : TOnForthSay      read FOnSay           write FOnSay;
    property    OnBye           : TOnBye           read FOnBye           write FOnBye;
    property    OnBreak         : TOnForthBreak    read FOnBreak         write FOnBreak;
    property    OnLog           : TOnForthLog      read FOnLog           write FOnLog;
    property    OnReset         : TNotifyEvent     read FOnReset         write FOnReset;
    property    OnUnknownSymbol : TOnUnknownSymbol   read FOnUnknownSymbol write FOnUnknownSymbol;
    property    OnDataOut       : TOnDataOut       read FOnDataOut       write FOnDataOut;
    property    OnAddScale      : TOnAddScale      read FOnAddScale      write FOnAddScale;
    property    OnGetScaleNames : TOnGetScaleNames read FOnGetScalenames write FOnGetScalenames;
    property    OnClearScales   : TOnClearScales   read FOnClearScales   write FOnClearScales;
    property    OnClearScreen   : TNotifyEvent     read FOnClearScreen   write FOnClearScreen;
  published
    property    OnLine          : TOnLine          read FOnLine          write FOnLine;
    property    OnFileEnter     : TOnFileEnter     read FOnFileEnter     write FOnFileEnter;
    property    OnFileLeave     : TOnFileLeave     read FOnFileLeave     write FOnFileLeave;
    property    OnFileResume    : TOnFileresume    read FOnFileResume    write FOnFileResume;
  private
}

    function    TKnobsForth.GetBase: Cardinal;
    var
      W : TSymbol;
    begin
      W := FindToken( 'base');
      if Assigned( W) and ( W is TVariable)
      then begin
        W.Interpret;
        _fetch( Self);
        Result := DataPop;
      end
      else Result := 10;
    end;

    procedure   TKnobsForth.SetBase( aValue: Cardinal);
    var
      W : TSymbol;
    begin
      W := FindToken( 'base');
      if Assigned( W) and ( W is TVariable)
      then begin
        DataPush( aValue);
        W.Interpret;
        _store( Self);
      end;
    end;

    function    TKnobsForth.GetCompiling: Boolean;
    begin
      Result := FCompiling > 0;
    end;

    procedure   TKnobsForth.SetCompiling( aValue: Boolean);
    begin
      if aValue
      then Inc( FCompiling)
      else if FCompiling > 0
      then Dec( FCompiling);
    end;

    function    TKnobsForth.GetInterpreting: Boolean;
    begin
      Result := FInterpreting > 0;
    end;

    procedure   TKnobsForth.SetInterpreting( aValue: Boolean);
    begin
      if aValue
      then Inc( FInterpreting)
      else if FInterpreting > 0
      then Dec( FInterpreting);
    end;

//  public

    procedure   TKnobsForth.Log( const aMsg: string);
    begin
      if Assigned( FOnLog)
      then FOnLog( Self, aMsg);
    end;

    procedure   TKnobsForth.Say( const S: string);
    begin
      if Assigned( FOnSay)
      then FOnSay( Self, S)
      else RaiseErrorFmt( 'No forth say handler, message was: %s', [ S]);
    end;

    procedure   TKnobsForth.SayFmt( const aFmt: string; const anArgs: array of const);
    begin
      Say( Format( aFmt, anArgs, AppLocale));
    end;

    procedure   TKnobsForth.RaiseBreak( const aMsg: string);
    begin
      if Assigned( FOnBreak)
      then begin
        try
          FOnBreak( Self, aMsg)
        finally
          FCompiling    := 0;
          FInterpreting := 0;
          FIP           := -1;
          FDataStack   .Clear;
          FReturnStack .Clear;
          FControlStack.Clear;
          FFileStack   .Reset;
        end;
      end
      else RaiseErrorFmt( 'break:  %s', [ aMsg]);
    end;

    procedure   TKnobsForth.RaiseError( const aMsg: string);
    begin
      try
        raise EKnobsForth.CreateFmt( '%s  [line %d, column %d, file "%s", trace { %s } ]', [ aMsg, SourceLineNr, SourceOffset, SourceName, ReturnStackTrace]);
      finally
        FCompiling    :=  0;
        FInterpreting :=  0;
        FIP           := -1;
        FDataStack   .Clear;
        FReturnStack .Clear;
        FControlStack.Clear;
        FFileStack   .Reset;
      end;
    end;

    procedure   TKnobsForth.RaiseErrorFmt( const aFmt: string; const anArgs: array of const);
    begin
      RaiseError( Format( aFmt, anArgs, AppLocale));
    end;

    procedure   TKnobsForth.ShowWarning( const aMsg: string);
    begin
      Say( Format( 'warning: %s', [ aMsg], AppLocale));
    end;

    procedure   TKnobsForth.ShowWarningFmt( const aFmt: string; const anArgs: array of const);
    begin
      ShowWarning( Format( aFmt, anArgs, AppLocale));
    end;

    procedure   TKnobsForth.Yield;
    begin
      Inc( FYieldCounter);
      if FYieldCounter > 100
      then begin
        FYieldCounter := 0;
        Application.ProcessMessages;
      end;
    end;

    procedure   TKnobsForth.DoStackUnderflow( aSender: TObject; const aMsg: string);
    begin
      if Interpreting
      then RaiseErrorFmt( '"%s" stack underflow', [ aMsg]);
    end;

    function    TKnobsForth.ReturnStackTrace: string;
    var
      i     : Integer;
      anIP  : TCFA;
      W     : TSymbol;
      aName : string;
    begin
      Result := 'unknown';
      if Assigned( FW)
      then begin
        if ( FW is TInternal) and ( Length( FW.Name) > 0) and ( FW.Name[ 1] = '_')
        then Result := Copy( FW.Name, 2, Length( FW.Name) - 1)
        else Result := FW.Name;
      end;
      for i := 0 to FReturnStack.Depth - 1
      do begin
        anIP := FReturnStack.Index( i);
        if ( anIp >= 0) and ( anIP < FCode.Count)
        then begin
          W  := FCode.Symbol[ anIP];
          if Assigned( W)
          then begin
            aName := W.Name;
            if ( W is TInternal) and ( Length( aName) > 0) and ( aName[ 1] = '_')
            then aName := Copy( aName, 2, Length( aName) - 1);
            Result:= Result + Format( ' %s', [ aName], AppLocale);
          end;
        end;
        Result := Result + Format( ' [&%d $%x]', [ anIP, anIP], AppLocale);
      end;
    end;

    function    TKnobsForth.GetToken( aSeparator: Char; Extended: Boolean = False): string;
    begin
      Result := FFileStack.GetToken( aSeparator, Extended);
    end;

    function    TKnobsForth.FetchToken( aSeparator: Char; Extended: Boolean = False): string;
    begin
      Result := GetToken( aSeparator, Extended);
      if SameText( Result, endoffile)
      then RaiseError( 'Unexpected end of file');
    end;

    procedure   TKnobsForth.PushLine( const aLine: string; const anID: string);
    begin
      FFileStack.PushLine( aLine, anID);
    end;

    procedure   TKnobsForth.PushLines( const aLines: TStrings; const anID: string);
    begin
      FFileStack.PushLines( aLines, anID);
    end;

    procedure   TKnobsForth.PushFile( const aFileName: string);
    begin
      if FileExists( aFileName)
      then begin
        FFileStack.Push( aFileName);
        AddFile( aFileName);
      end
      else ShowWarningFmt( 'the included file "%s" does not exist', [ aFileName]);
    end;

    procedure   TKnobsForth.SkipToEol;
    begin
      FFileStack.SkipToEol;
    end;

    function    TKnobsForth.GetToEol: string;
    begin
      Result := FFileStack.GetToEol;
    end;

    function    TKnobsForth.ReadLine: string;
    begin
      Result := FFileStack.ReadLine;
    end;

    procedure   TKnobsForth.FinalizeCreation( const aSymbol: TSymbol; const aComment: string = '');
    begin
      FLastDefined := aSymbol;
      with FLastDefined
      do begin
        if aComment <> ''
        then Comment := aComment;
        Reveal;
        EndCreate;
      end;
    end;

    function    TKnobsForth.Number( const aToken: string; out aNumber: Variant): Boolean;
    const

      DigitValues : array[ AnsiChar] of Byte = (

      {          0    1    2    3     4    5    6    7      8    9    a    b     c    d    e    f}

      { 00 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255, // Ctrl
      { 10 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255,
      { 20 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255, // Specials
      { 30 }     0,   1,   2,   3,    4,   5,   6,   7,     8,   9, 255, 255,  255, 255, 255, 255, // Nums
      { 40 }   255,  10,  11,  12,   13,  14,  15,  16,    17,  18,  19,  20,   21,  22,  23,  24, // UpCase
      { 50 }    25,  26,  27,  28,   29,  30,  31,  32,    33,  34,  35, 255,  255, 255, 255, 255,
      { 60 }   255,  10,  11,  12,   13,  14,  15,  16,    17,  18,  19,  20,   21,  22,  23,  24, // LoCase
      { 70 }    25,  26,  27,  28,   28,  30,  31,  32,    33,  34,  35, 255,  255, 255, 255, 255,
      { 80 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255, // Extended ascii
      { 90 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255,
      { a0 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255,
      { b0 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255,
      { c0 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255,
      { d0 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255,
      { e0 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255,
      { f0 }   255, 255, 255, 255,  255, 255, 255, 255,   255, 255, 255, 255,  255, 255, 255, 255

      );

    var
      i        : Integer;                  // Loop index
      d        : Cardinal;                 // Current digit
      n        : Int64;                    // Number accumulator
      c        : AnsiChar;                 // aToken[ i] when within the for loop
      Neg      : Boolean;                  // True -> leading minus seen
      Asc      : Boolean;                  // True -> Ascii modifier seen
      LocBase  : Cardinal;                 // The current number base in which the string is to be interpreted
      DotCount : Cardinal;                 // Number of DOTs in number
      DotPos   : Integer;                  // Position of dot
    begin
      Result   := Length( aToken) > 0;     // Numbers must have length > 0
      n        := 0;                       // Initially the number is zero
      Neg      := False;                   //   and not negative
      Asc      := False;                   //   and not anAsc
      LocBase  := Base;                    // The default number base is the current system number base
      DotCount := 0;                       // We saw no dots yet
      DotPos   := 0;                       // So it's at the 0th position currently

      for i := 1 to Length( aToken)
      do begin
        c := AnsiChar( aToken[ i]);        // Setup c as part of the loop invariant
        if Asc                             // In Asc mode, special handling needed
        then begin
          {$Q-R-}                          // Allow for overflows and range violations
          n := 256 * n + Ord( c);          // Asc numbers are considerd to be in base &256
          {$Q+R+}
        end
        else begin                         // Not in Asc mode now
          case c of                        // Check for special chars
            '.' : begin
                Inc( DotCount);            // Float   modifier
                DotPos := i;
              end;
            '%' : LocBase :=  2;              // Binary  modifier
            '&' : LocBase := 10;              // Decimal modifier
            '$' : LocBase := 16;              // Hex     modifier
            '''' : begin                   // Asc     modifier
                if                         // Asc     modifier - must be 1st char or 1st after '-'
                  (  i = 1) or
                  (( i = 2) and Neg)
                then Asc := True           // Switch to Asc mode
                else begin
                  Result := False;         // Not a valid Asc modifier
                  Break;                   // Not a number
                end;
              end;
            '-' : begin                    // Negative number modifier, unary minus
                if i = 1                   // but only if '-' is the first char
                then Neg := True           // Indicate negative number
                else begin
                  Result := False;         // Not a valid unary minus
                  Break;                   // Not a number
                end;
              end;
            else begin                     // No special char, should be a digit then
              d := DigitValues[ c];        // Invalid digits return 255
              if d >= LocBase                 // Check for invalid digit
              then begin
                Result := False;           // Invalid digit
                Break;                     // Not a number
              end
              else begin                   // Valid digit within number base
                {$Q-R-}                    // Allow for overflows and range violations
                n := LocBase * n + d;         // Accumulate digit into number
                {$Q+R+}
              end;
            end;
          end; // Case
        end
      end; // For

      if DotCount > 1                      // no more than one dot in a number
      then Result := False;

      if Result                            // Most likely it's a number then ... sift out some possible mistakes
      then begin

        {$Q-R-}                            // Allow for overflows and range violations
        if Neg                             // Check if an unary minus was seen
        then aNumber := - Int64( n)        // and if so return -n
        else aNumber :=   Int64( n);       // otherwise just    n
        {$Q+R+}

        if DotCount = 1
        then aNumber := aNumber / Power( 10, Length( aToken) - DotPos);

      end;

      if not Result                        // if not a number grReturn $deadbeef
      then aNumber := NoNumber;            // as a probable invalid number indicator.
    end;

    function    TKnobsForth.Float( const aToken: string; out aNumber: Variant): Boolean;
    //
    // Only supports floats in decimal notation.
    //
    var
      aFloat: Extended;
    begin
      Result  := False;
      aNumber := $deadbeef;
      try
        if TextToFloat( PChar( aToken), aFloat, fvExtended)
        then begin
          aNumber := aFloat;
          Result  := True;
        end;
      except
        // Just ignorre any exceptions and return False instead.
      end;
    end;

    function    TKnobsForth.MustFindToken( const aToken: string; aSymbolClass: TSymbolClass): TSymbol;
    var
      W: TSymbol;
    begin
      Result := FindToken( aToken);
      if ( not Assigned( Result)) or not ( Result is aSymbolClass)
      then begin
        Result := nil;
        if aSymbolClass = TSymbol                     // The generic case should find all symbols
        then
          RaiseErrorFmt(
            '%s ?',
            [
              aToken
            ]
          )
        else begin                                    // and otherwise a specialised class was wanted
          W := FindToken( aToken);
          if Assigned( W)
          then
            RaiseErrorFmt(
              '"%s" is of type "%s" but type "%s" was requiered',
              [
                aToken,
                W.TypeName,
                aSymbolClass.TypeName
              ]
            )
          else
            RaiseErrorFmt(
              '%s ?',
              [
                aToken
              ]
            );
        end;
      end;
    end;

    function    TKnobsForth.MustFindSymbol( const aToken: string): TSymbol;
    begin
      Result := MustFindToken( aToken, TSymbol);
    end;

    procedure   TKnobsForth.MustBeCompilingFor( const aToken: string);
    begin
      if not Compiling
      then
        RaiseErrorFmt(
          'The symbol "%s" can''t be used outside a colon definition',
          [ aToken]
        );
    end;

    procedure   TKnobsForth.MustNotBeCompilingFor( const aToken: string);
    begin
      if Compiling
      then
        RaiseErrorFmt(
          'The symbol "%s" can''t be used inside a colon definition',
          [ aToken]
        );
    end;

    function    TKnobsForth.FindToken( const aToken: string): TSymbol;
    var
      i : Integer;
    begin
      Result := nil;
      for i := Length( FContext) - 1 downto 0
      do begin
        Result := FContext[ i].FindToken( aToken);
        if Assigned( Result)
        then Break;
      end;
    end;

    procedure   TKnobsForth.Run;
    begin
      Interpreting := True;
      try
        try
          while Interpreting
          do begin
            AcceptToken( GetToken( ' ', False));     // Use GetToken here, not FetchToken
            Application.ProcessMessages;
          end;
        except
          // All errors are trapped here.
          on E: Exception
          do begin
            Log( E.Message);
            Interpreting := False;
          end;
        end;
      finally
        Interpreting := False;
      end;
    end;

    procedure   TKnobsForth.AcceptToken( const aToken: string);
    var
      S       : TSymbol;
      aNumber : Variant;
    begin
      try
        S := nil;
        // Search for global definitions
        if not Assigned( S)
        then S := FindToken( aToken);
        // When found let the symbol be interpreted or compiled depending
        // on immediateness and the current compile state.
        if Assigned( S)
        then begin
          if S.Immediate or not Compiling
          then S.Interpret
          else CompileSymbol( S);
        end
        // Otherwise see if the token is a floating point number
        else if Float( aToken, aNumber)
        then begin
          if Compiling
          then CompileLiteral( aNumber)                // compile literal number
          else DataPush( aNumber);                     // push number onto data stack
        end
        // Otherwise see if the token is an integer or a character constant
        else if Number( aToken, aNumber)
        then begin
          if Compiling
          then CompileLiteral( aNumber)                // compile literal number
          else DataPush( aNumber);                     // push number onto data stack
        end
        // None worked ...
        else begin
          if not HandleUnknowSymbol( aToken)           // give user code a shot
          then RaiseErrorFmt( '%s ?', [ aToken]);      // and if it doesnt handle it, fail.
        end;
      except
        // All errors are trapped here.
        on E: Exception
        do Say( E.Message);
      end;
    end;

    function    TKnobsForth.HandleUnknowSymbol( const aToken: string): Boolean;
    begin
      Result := False;
      if Assigned( FOnUnknownSymbol)
      then Result := FOnUnknownSymbol( Self, aToken);
    end;

    procedure   TKnobsForth.DataOut( const aChannel: string; const aData: TVariants);
    begin
      if Assigned( FOnDataOut)
      then FOnDataOut( Self, aChannel, aData);
    end;

    procedure   TKnobsForth.AddScale( const aName: string; const aScale: TMidiFilter);
    begin
      if Assigned( FOnAddScale)
      then FOnAddScale( Self, aname, aScale);
    end;

    function    TKnobsForth.GetScaleNames: string;
    begin
      if Assigned( FOnGetScaleNames)
      then Result := FOnGetScaleNames( Self)
      else Result := 'no scales were defined';
    end;

    procedure   TKnobsForth.Explain( aSymbol: TSymbol);
    begin
      Say( aSymbol.See);
    end;

    procedure   TKnobsForth.Call( aList: TForthList; anIP: TCFA); // override;
    begin
      ReturnPush( FIP);
      FIP := anIP;
      while ( FIP >= 0) and ( FIP < aList.Count)
      do begin
        FW := aList.Symbol[ FIP];
        if FIP >= 0
        then FIP := FIP + 1;
        FW.Interpret;
        Yield;
      end;
      if FReturnStack.Depth > 0
      then FIP := ReturnPop
      else FIP := -1;
    end;

    procedure   TKnobsForth.AddAlias( const anOldName, aNewName: string);
    var
      W       : TSymbol;
      anAlias : TAlias;
    begin
      W       := MustFindSymbol( anOldName);
      anAlias := AddSymbol( aNewName, TAlias) as TAlias;
      anAlias.Aliased := W;
      FinalizeCreation( W);
    end;

    procedure   TKnobsForth.CompileLiteral( const aValue: Variant);
    begin
      FCode.CompileLiteral( aValue);
    end;

    procedure   TKnobsForth.CompileSymbol( const aSymbol: TSymbol);
    begin
      FCode.CompileSymbol( aSymbol);
    end;

    procedure   TKnobsForth.CompileToken( const aName: string); // overload;
    begin
      FCode.CompileToken( aName);
    end;

    procedure   TKnobsForth.CompileToken( const aName: string; const aSymbolClass: TSymbolClass); // overload;
    begin
      FCode.CompileToken( aName, aSymbolClass);
    end;

    function    TKnobsForth.Here: TCFA;
    begin
      if Assigned( Fcode)
      then Result := FCode.Here
      else Result := 0;
    end;

    procedure   TKnobsForth.SetHere( aValue: TCFA);
    begin
      FCode.SetHere( aValue);
    end;

    procedure   TKnobsForth.DataPush( aValue: Variant);
    begin
      FDataStack.Push( aValue);
    end;

    procedure   TKnobsForth.DataPushSymbol( const aValue: TSymbol);
    begin
      FDataStack.PushSymbol( aValue);
    end;

    function    TKnobsForth.DataPop: Variant;
    begin
      Result := FDataStack.Pop;
    end;

    function    TKnobsForth.DataFetch: Variant;
    begin
      Result := FDataStack.Fetch;
    end;

    function    TKnobsForth.DataIndex( anIndex: Integer): Variant;
    begin
      Result := FDataStack.Index( anIndex);
    end;

    procedure   TKnobsForth.DataPut( anIndex: Integer; aValue: Variant);
    begin
      FDataStack.Put( anIndex, aValue);
    end;

    function    TKnobsForth.DataPopSymbol: TSymbol;
    begin
      Result := FDataStack.PopSymbol;
    end;

    procedure   TKnobsForth.DataSwap;
    begin
      FDataStack.Swap;
    end;

    procedure   TKnobsForth.ReturnPush( aValue: Variant);
    begin
      FReturnStack.Push( aValue);
    end;

    function    TKnobsForth.ReturnPop: Variant;
    begin
      Result := FReturnStack.Pop;
    end;

    procedure   TKnobsForth.ControlPush( aValue: Variant);
    begin
      FControlStack.Push( aValue);
    end;

    function    TKnobsForth.ControlPop: Variant;
    begin
      Result := FControlStack.Pop;
    end;

    function    TKnobsForth.AddSymbol( const aName: string; aSymbolClass: TSymbolClass): TSymbol;
    begin
      if Assigned( FindToken( aName))
      then ShowWarningFmt( 'Redefinition of ''%s''', [ aName]);
      Result := Current.AddSymbol( aName, aSymbolClass);
      if Result is TColon
      then TColon( Result).CFA := Here;
    end;

    procedure   TKnobsForth.Internal( const aName: string; aProc: TInternalProc; const aComments: string);
    var
      anInternal : TInternal;
    begin
      anInternal := TInternal( AddSymbol( aName, TInternal));
      anInternal.InternalProc := aProc;
      FinalizeCreation( anInternal, aComments);
    end;

    procedure   TKnobsForth.SetupInternals;
    begin
      // Construct initial search order manually

      SetLength( FVocs, 0);                                 // Clear the vocabulary system
      FOnly := AddVoc( TVocabulary.Create( Self, 'only'));  // Evaluate( 'vocabulary only  // create the pseudo vocabulary only')
      _only( Self); _definitions( Self);                    // Evaluate( 'only definitions // compile to pseudo vocabulary only')

      Internal( 'only'          , _only          , '( -- ) // root (pseudo) vocabulary containing basic vocabulary control symbols');
      Internal( 'also'          , _also          , '( -- ) // duplicate top of search list');
      Internal( 'vocabulary'    , _vocabulary    , '( -- ) // vocabulary XXXX defines a new vocabulary XXXX');
      Internal( 'definitions'   , _definitions   , '( -- ) // set the top of the context (lookup) stack as the current (compile to) vocabulary');
      Internal( 'dropvoc'       , _dropvoc       , '( -- ) // remove top of search order');
      Internal( '.vocs'         , _dotvocs       , '( -- ) // show a list of all defined vocs.');
      Internal( '.order'        , _dotorder      , '( -- ) // show search order');
      Internal( '.words'        , _dotwords      , '( -- ) // show the symbol list');
      Internal( '.help'         , _help          , '( -- ) // show very basic help');
      Internal( 'reset'         , _reset         , '( -- ) // clear the complete forth system and then reload the default file');
      Internal( endoffile       , _eof           , '( -- ) // To stop on end of file, can be used to break out of a source file');

      Evaluate( 'only vocabulary forth also forth definitions');

      Internal( 'constant'      , _constant      , '( n -- ) // define a constant :: 12 constant joop');
      Internal( 'variable'      , _variable      , '( -- ) // define a variable :: variable truus');
      Internal( '@'             , _fetch         , '( n1 -- n2 ) // fetch value from a variable');
      Internal( '!'             , _store         , '( n1 n2 -- ) // Store value n1 into variable n2');

      Evaluate( 'variable base &10 base !');

      Internal( 'included'      , _included      , '( "fname" -- ) // include the file fname, and add the name to the files list');
      Internal( 'files'         , _files         , '( -- {files}) // return a list of the names of all included files');
      Internal( 'included?'     , _includedq     , '( "fname" -- f) // return true when fname is in files');
      Internal( 'required'      , _required      , '( "fname" -- ) // when fname is not in the files list include the file named fname');
      Internal( 'fload'         , _fload         , '( "filename" -- "string" ) // load text data from a file');
      Internal( 'fsave'         , _fsave         , '( contents "filename" -- ) // save text data to a file');
      Internal( 'bye'           , _bye           , '( ? -- ) // shut down the program');
      Internal( 'cls'           , _cls           , '( -- ) clear the terminal screen');
      Internal( 'data_out'      , _dataOut       , '( n1 .. nn n channel -- ) // send out n entries of data on channel in outside world');
      Internal( 'anonymous'     , _anonymous     , '( -- "string" ) // get anonymous symbol info');
      Internal( 'dump'          , _dump          , '( -- ) // make acode dump');
      Internal( 'vocdump'       , _vocdump       , '( -- ) // make a voc based dump');
      Internal( '>name'         , _toname        , '( <exec-token> -- "string") // Find a name for an execution token');
      Internal( 'cfa>name'      , _cfatoname     , '( cfa -- "string") // find a name for a CFA. works on colon and create words only');
      Internal( '#if'           , _h_if          , '( f -- ) // conditional compilation : n1 #ff .. [[#else]] .. [[#fi]]');
        Immediate;
      Internal( '#else'         , _h_else        , '( -- ) // conditional compilation : n1 [[#if]] .. #else .. [[#fi]]');
        Immediate;
      Internal( '#fi'           , _h_fi          , '( -- ) // conditional compilation : n1 [[#if]] .. [[#else]] .. #fi');
        Immediate;
      Internal( '#define'       , _h_define      , '( -- ) // define a symbol that does nothing.');
        Immediate;
      Internal( '#defined'      , _h_defined     , '( -- f ) // Check if a symbol is defined, result goes to data stack');
        Immediate;
      Internal( '#ifdef'        , _h_ifdef       , '( -- ) // Combination : [[#defined]] XXXX [[#if]]');
        Immediate;
      Internal( '#ifndef'       , _h_ifndef      , '( -- ) // same as [[#IfDef]] but inverted condition');
        Immediate;
      Internal( '.s'            , _dots          , '( -- ) // print a data stack dump');
      Internal( '.r'            , _dotr          , '( -- ) // print a return stack dump');
      Internal( '.c'            , _dotc          , '( -- ) // print a control stack dump');
      Internal( '.stacks'       , _dotstacks     , '( -- ) // print a stack dump for all stacks');
      Internal( 'decimal'       , _decimal       , '( -- ) // &10 [[base!]]');
      Internal( 'hex'           , _hex           , '( -- ) // &16 [[base!]]');
      Internal( 'binary'        , _binary        , '( -- ) // &2  [[base!]]');
      Internal( '#doc'          , _doc           , '( -- ) // Add documentation to [[LastDefined]], ended with [[#cod]], also see [[#//]] and [[##//]]');
        Immediate;
      Internal( '#//'           , _doceol        , '( -- ) // Add documentation to [[LastDefined]], ended at end of current line, also see [[#Doc]] and [[#Document]]');
        Immediate;
      Internal( '#document'     , _document      , '( -- ) // Add documentation to next token in input buffer, ended with [[#cod]]');
      Internal( '##//'          , _documenteol   , '( -- ) // Add documentation to next token in input buffer, ended at end of current line, also see [[#//]] and [[##//]]');
      Internal( '#cod'          , _cod           , '( -- ) // Ends a [[#doc]] or [[#document]] section');
      Internal( '('             , _lparen        , '( -- ) // skip input till )');
        Immediate;
      Internal( '//'            , _slashslash    , '( -- ) // skip input till EOL');
        Immediate;
      Internal( '.('            , _dotlparen     , '( -- ) // print input until )');
        Immediate;
      Internal( '."'            , _dotquote      , '( -- ) // print input until "');
        Immediate;
      Internal( './/'           , _dotslashslash , '( -- ) // print input until EOL');
        Immediate;
      Internal( '{'             , _lcurly        , '( -- {list} ) // construct a space separated list up to }');
        Immediate;
      Internal( '}'             , _rcurly        , '( -- ) // a noop to keep the list parser happy');
      Internal( '"'             , _doubleQuote   , '( -- "string" ) // " a string"');
        Immediate;
      Internal( ''''''          , _singleQuote   , '( -- "string" ) // '''' a string''');
        Immediate;

      Internal( 'true'          , _true          , '( -- f ) // push boolean true');
      Internal( 'false'         , _false         , '( -- f ) // push boolean false');
      Internal( 'e'             , _e             , '( -- n ) // push e');
      Internal( 'pi'            , _pi            , '( -- n ) // push pi');
      Internal( 'bl'            , _space         , '( -- n ) // push a space character');

      Internal( '.'             , _dot           , '( n -- ) // print tos, removing it from the stack');
      Internal( '.debug'        , _dotdebug      , '( n -- ) // print tos to debug monitor only and remove it from the stack');
      Internal( '+'             , _plus          , '( n1 n2 -- n3 ) // adition');
      Internal( '-'             , _minus         , '( n1 n2 -- n3 ) // subtraction of n2 from n1');
      Internal( '*'             , _times         , '( n1 n2 -- n3 ) // multiplication');
      Internal( '/'             , _div           , '( n1 n2 -- n3 ) // division of n1 by n2');
      Internal( '/mod'          , _divMod        , '( n1 n2 -- n3 n4 ) // division of n1 by n2 with result n3 and remainder n4');

      Internal( 'sqrt'          , _sqrt          , '( n -- ) // square root of n');
      Internal( 'abs'           , _abs           , '( n1 -- n2 ) // absolute value of the top stack element');
      Internal( 'max'           , _max           , '( n1 n2 -- n3 ) // maximum of the top two stack elements');
      Internal( 'min'           , _min           , '( n1 n2 -- n3 ) // minimum of the top two stack elements');
      Internal( 'and'           , _and           , '( n1 n2 -- n3 ) // bitwise and');
      Internal( 'or'            , _or            , '( n1 n2 -- n3 ) // bitwise or');
      Internal( 'xor'           , _xor           , '( n1 n2 -- n3 ) // bitwise xor');
      Internal( 'negate'        , _negate        , '( n1 -- n2 ) // unary minus performed on top stack element');
      Internal( 'not'           , _not           , '( n1 -- n2 ) // logic inversion performed on top stack element');
      Internal( '<<'            , _lmshift       , '( n1 n2 -- n3 ) // shift the second stack element left over the number of bits indicated by the top stack element');
      Internal( '>>'            , _rmshift       , '( n1 n2 -- n3 ) // shift the second stack element right over the number of bits indicated by the top stack element');
      Internal( '='             , _equal         , '( n1 n2 -- f ) // true when the top two stack elements are equal, false otherwise');
      Internal( '<>'            , _unequal       , '( n1 n2 -- f ) // true when the top two stack elements are un-equal, false otherwise');
      Internal( '<'             , _smaller       , '( n1 n2 -- f ) // true when u1 < u2, false otherwise');
      Internal( '>'             , _greater       , '( n1 n2 -- f ) // true when u1 > u2, false otherwise');
      Internal( '<='            , _smallerEqual  , '( n1 n2 -- f ) // true when u1 <= u2, false otherwise');
      Internal( '>='            , _greaterEqual  , '( n1 n2 -- f ) // true when u1 >= u2, false otherwise');
      Internal( 'random'        , _random        , '( n1 n2 -- n3 ) // produce a random value [n1 .. n2]');

      Internal( 'getmem'        , _getmem        , '( size -- pointer ) // allocate memory from the system heap');
      Internal( 'freemem'       , _freemem       , '( pointer -- ) // free previously allocated memory');
      Internal( 'reallocmem'    , _reallocmem    , '( pointer size -- pointer ) // reallocate previously allocated memory');
      Internal( 'mem@'          , _memfetch      , '( pointer offset n -- n1 .. nn ) // read n cell sized items from an offset on pointer');
      Internal( 'mem!'          , _memstore      , '( n1 .. nn n pointer offset -- ) // store n cell sized items to an offset on pointer');
      Internal( '($)'           , _tostring      , '( n -- "string" ) // typecast n to a string');
      Internal( 'format'        , _format        , '( n "fmt" -- "string" ) // format a float ... decimal only');
      Internal( 'now'           , _now           , '( -- n ) // push current local time as datetime');
      Internal( 'utc'           , _utc           , '( -- n ) // push current UTC time as datetime');
      Internal( 'time>osc'      , _timetoosc     , '( -- n ) // convert datetime to OSC time');
      Internal( 'osc>time'      , _osctotime     , '( -- n ) // convert OSC time to datetime');
      Internal( 'oscnow'        , _oscnow        , '( -- n ) // push current UTC time as OSC time');
      Internal( 'format_time'   , _format_time   , '( nt "fmt" -- "nr" ) // format date_time nt according to fmt');
      Internal( 'replace'       , _replace       , '( "string" "search" "replace" -- "string" ) // replace all occurances of search in string with replace, case sensitive');
      Internal( 'replacei'      , _replacei      , '( "string" "search" "replace" -- "string" ) // replace all occurances of search in string with replace, case insensitive');
      Internal( 'replacefirst'  , _replacefirst  , '( "string" "search" "replace" -- "string" ) // replace the first occurance of search in string with replace, case sensitive');
      Internal( 'replacefirsti' , _replacefirsti , '( "string" "search" "replace" -- "string" ) // replace the first occurance of search in string with replace, case insensitive');

      Internal( 'scale'         , _scale         , '( {list} -- ) // { n1 n2 .. nn } scale XXXX // add a musical scale to the system');
      Internal( 'scalenames'    , _scalenames    , '( -- "string" ) // return str - a list of all scales');

      Internal( 'drop'          , _drop          , '( n1 -- ) // drop top stack element');
      Internal( 'dup'           , _dup           , '( n1 -- n1 n1 ) // duplicate top stack element');
      Internal( 'swap'          , _swap          , '( n1 n2 -- n2 n1 ) // swap top and sub top stack elements');
      Internal( '?dup'          , _questiondup   , '( n1 -- n1 n1 | 0 -- 0 ) // duplicate the top stack element when it is not equal to zero');
      Internal( '2drop'         , _2drop         , '( n1 n2 -- ) // drop the top two stack elements');
      Internal( 'over'          , _over          , '( n1 n2 -- n1 n2 n1 ) // duplicate the second stack element and make it the top item');
      Internal( '2dup'          , _2dup          , '( n1 n2 -- n1 n2 n1 n2 ) // duplicate the top two stack elements');
      Internal( '2swap'         , _2swap         , '( n1 n2 n3 n4 -- n3 n4 n1 n2 ) // exchange the top two sets of two stack items');
      Internal( '2over'         , _2over         , '( n1 n2 n3 n4 -- n1 n2 n3 n4 n1 n2 ) // duplicate the second set of two stack elements');
      Internal( 'rot'           , _rot           , '( n1 n2 n3 -- n2 n3 n1 ) // rotate the 3rd stack element to the top');
      Internal( '+rot'          , _plusrot       , '( n1 n2 n3 -- n2 n3 n1 ) // rotate the 3rd stack element to the top');
      Internal( '-rot'          , _minusrot      , '( n1 n2 n3 -- n3 n1 n2 ) // rotate the top stack element to the 3rd position');
      Internal( 'pick'          , _pick          , '( nn .. na .. n0 a -- nn .. na .. n0 na ) // pick stack item a and duplicate it to stack top');
      Internal( 'put'           , _put           , '( nn .. na .. n0 x a -- nn .. x .. n0 ) // write item a into position x of the stack, replacing what wwas there');
      Internal( '+!'            , _plusStore     , '( n1 n2 -- ) // increment value in variable n1 with n2');
      Internal( '-!'            , _minStore      , '( n1 n2 -- ) // Decrement value in variable n1 with n2');

      Internal( 'token'         , _token         , '( bSep -- "string" ) // read bSep separated token from input stream');
      Internal( 'untoken'       , _untoken       , '( "string" bSep -- ) // prepend token and separator to input stream');
      Internal( 'readLine'      , _readLine      , '( -- "string" ) // read a line from the input and push it as a string');
      Internal( '>mark'         , _fmark         , '( -- n ) // forward jump (for if .. fi etc), also see [[<mark]], [[>resolve]] and [[<resolve]]');
        SysSymbol;
      Internal( '>resolve'      , _fresolve      , '( n -- ) // resolve [[>mark]], also see [[<mark]] and [[<resolve]]');
        SysSymbol;
      Internal( '<mark'         , _bmark         , '( -- "string" | n ) // backward jump (for begin ... again etc, also see [[>mark]]');
        SysSymbol;
      Internal( '<resolve'      , _bresolve      , '( "string" } n -- ), resolve [[<mark]], also see [[>mark]] and [[>resolve]]');
        SysSymbol;
      Internal( 'branch'        , _branch        , '( -- ), unconditional branch');
        SysSymbol;
      Internal( '?branch'       , _fbranch       , '( f -- ), branch if TOS is zero');
        SysSymbol;
      Internal( ':'             , _colon         , '( -- ) // Start a colon definition : XXXX ... [[;]]');
      Internal( 'recurse'       , _recurse       , '( ? -- ? ) // Call into self at run time, explicit recursion');
        Immediate;
      Internal( ';'             , _semi          , '( -- ) // end a colon definition [[:]] XXXX ... ;');
        Immediate;
      Internal( 'create'        , _create        , '( -- ): XXXX create ... [[does>]] ... [[pfa@]] ...;');
      Internal( '(does)'        , _pdoes         , '( -- pfa ), helper for [[does>]]');
        SysSymbol;
      Internal( 'does>'         , _does          , '( -- ), see [[create]] and [[pfa@]]');
        Immediate;
      Internal( 'ip'            , _ip            , '( -- n ) // push the current IP value');
      Internal( 'pfa@'          , _pfafetch      , '( pfa -- <data-token> ) // pfa is an index in code space, pfa@ will fetch the TSymbol stored there :: to be used after does> usually');
      Internal( 'pfa@@'         , _pfafetchfetch  , '( pfa -- data pfa )');
        Crash;
      Internal( 'pfa@!'         , _pfafetchstore  , '( pfa data -- pfa )');
        Crash;
      Internal( ','             , _comma         , '( n -- ) // allots in code space');
      Internal( 'allot'         , _allot         , '( n -- ) // allots in code space');
      Internal( 'here'          , _here          , '( -- n ) // location in code space');
      Internal( 'crash'         , _crash         , '( -- ) // make latest symbol toggle its crash bit = when set it will rase an exception on interpretation');
      Internal( 'immediate'     , _immediate     , '( -- ) // make latest symbol toggle its immediate bit');
      Internal( 'sysword'       , _sysword       , '( -- ) // make latest symbol toggle its sysword bit');
      Internal( 'hide'          , _hide          , '( -- ) // make latest symbol set its visible bit');
      Internal( 'reveal'        , _reveal        , '( -- ) // make latest symbol clear its visible bit');

      // These are for the compiler
      Internal( '_begin'        , __begin        , '( -- ) // compiled begin');
        SysSymbol;
      Internal( '_until'        , __until        , '( n -- ) // compiled until');
        SysSymbol;
      Internal( '_while'        , __while        , '( n -- ) // compiled while');
        SysSymbol;
      Internal( '_again'        , __again        , '( -- ) // compiled again');
        SysSymbol;
      Internal( '_repeat'       , __repeat       , '( -- ) // compiled repeat');
        SysSymbol;
      Internal( '_do'           , __do           , '( n1 n2 -- ) // compiled do');
        SysSymbol;
      Internal( '_?do'          , __qdo          , '( n1 n2 -- ) // compiled ?do');
        SysSymbol;
      Internal( '_od'           , __od           , '( -- ) // compiled od');
        SysSymbol;
      Internal( '_pod'          , __pod          , '( n -- ) // compiled +od');
        SysSymbol;
      Internal( '_break'        , __break        , '( -- ) // compiled break');
        SysSymbol;
      Internal( '_leave'        , __leave        , '( -- ) // compliled leave');
        SysSymbol;
      Internal( '_if'           , __if           , '( f -- ) // compiled if');
        SysSymbol;
      Internal( '_else'         , __else         , '( -- ) // compiled else');
        SysSymbol;
      Internal( '_fi'           , __fi           , '( -- ) // compiled fi');
        SysSymbol;
      Internal( '_case'         , __case         , '( -- ) // compiled case');
        SysSymbol;
      Internal( '_default'      , __default      , '( -- ) // compiled default');
        SysSymbol;
      Internal( '_esac'         , __esac         , '( -- ) // compiled esac');
        SysSymbol;
      Internal( '_of'           , __of           , '( n1 n2 -- n1 | ) // compiled of');
        SysSymbol;
      Internal( '_fo'           , __fo           , '( -- ) // compiled fo');
        SysSymbol;
      Internal( '_in'           , __in           , '( n1 n2 -- n1 | ) // compiled in');
        SysSymbol;
      Internal( '_ni'           , __ni           , '( -- ) // compiled ni');
        SysSymbol;
      Internal( '_exit'         , __exit         , '( -- ) // compiled exit');
        SysSymbol;

      Internal( 'begin'         , _begin         , '( -- n n ) // begin .. [[again]], begin .. 0 [[until]], begin .. (f) [[while]] .. [[repeat]].');
        Immediate;
      Internal( 'again'         , _again         , '( n n -- ) // [[begin]] .. again');
        Immediate;
      Internal( 'until'         , _until         , '( n n -- ) // [[begin]] .. ( f ) until');
        Immediate;
      Internal( 'while'         , _while         , '( n n -- ) // [[begin]] .. while .. [[repeat]]');
        Immediate;
      Internal( 'repeat'        , _repeat        , '( n n -- ) // [[begin]] .. [[while]] .. repeat');
        Immediate;
      Internal( 'do'            , _do            , '( -- n n ) // n1 n2 do .. [[leave]] .. [[od]] or n [[+od]]');
        Immediate;
      Internal( '?do'           , _qdo           , '( -- n n ) // n1 n2 ?do .. [[leave]] .. [[od]] or n [[+od]]');
        Immediate;
      Internal( 'od'            , _od            , '( n n -- ) // n1 n2 [[do]] or [[?do]] .. [lleave]] .. od');
        Immediate;
      Internal( '+od'           , _pod           , '( n n -- ) // n1 n2 [[do]] or [[?do]] .. n +od');
        Immediate;
      Internal( 'break'         , _break         , '( -- ) // n1 n2 [[do]] or [[?do]] .. [[leave]] .. break .. [[od]]');
        Immediate;
      Internal( 'leave'         , _leave         , '( -- ) // leave a [[do]] .. leave .. [[od]]');
        Immediate;
      Internal( 'i'             , _i             , '( -- n ) // Get inner most [[do]] index');
      Internal( 'j'             , _j             , '( -- n ) // Get 2nd level [[do]] index');
      Internal( 'k'             , _k             , '( -- n ) // Get 3rd level [[do]] index');
      Internal( '>r'            , _tor           , '( n -- ) // move a value from the data to the loop stack');
      Internal( 'r>'            , _rto           , '( -- n ) // move a value from the loop to the data stack');
      Internal( 'if'            , _if            , '( -- n ) // n if .. [[else]] .. [[fi]]');
        Immediate;
      Internal( 'else'          , _else          , '( n -- n ) // n [[if]] .. else .. [[fi]]');
        Immediate;
      Internal( 'fi'            , _fi            , '( n -- ) // n [[if]] .. [[else]] .. fi');
        Immediate;
      Internal( 'case'          , _case          , '( -- 4 0 ) // ( n ) case .. [[of]] .. [[fo]] .. [[default]] .. [[esac]]');
        Immediate;
      Internal( 'default'       , _default       , '( -- ) // ( n ) [[case]] .. [[of]] .. [[fo]] .. default .. [[esac]]');
        Immediate;
      Internal( 'esac'          , _esac          , '( n0 n1 .. nn-1 0 4 -- ) // ( n ) [[case]] .. [[of]] .. [[fo]] .. [[default]] .. esac');
        Immediate;
      Internal( 'of'            , _of            , '( -- n 5 ) // ( n ) [[case]] .. of .. [[fo]] .. [[default]] .. [[esac]]');
        Immediate;
      Internal( 'fo'            , _fo            , '( n1 5 -- n1 n2 ) // ( n ) [[case]] .. [[of]] .. fo .. [[default]] .. [[esac]]');
        Immediate;
      Internal( 'in'            , _in            , '( -- n 5 ) // ( n ) [[case]] .. in .. [[ni]] .. [[default]] .. [[esac]]');
        Immediate;
      Internal( 'ni'            , _ni            , '( n1 5 -- n1 n2 ) // ( n ) [[case]] .. [[in]] .. ni .. [[default]] .. [[esac]]');
        Immediate;
      Internal( 'exit'          , _exit          , '( -- ) // immediately leave the current colon definition returning to it''s caller');
        Immediate;

      Internal( 'defer'         , _defer         , '( -- ) // defer XXXX, to be later resolved with [[is]] or [[implements]]');
      Internal( ''''            , _tick          , '( -- <exec-token> ) // search next token in input stream and return it''s execution token. Also see [[['']]]');
      Internal( '['']'          , _bracketedtick , '( -- ) // search next token in input stream and compile it''s execution token - immediate, see [['']]');
        Immediate;
      Internal( '['             , _lsbracket     , '( -- ) // Switch compiler off, also see ]');
        Immediate;
      Internal( ']'             , _rsbracket     , '( -- ) // Switch compiler on');
      Internal( 'literal'       , _literal       , '( n -- ) // Compile TOS as a literal');
        Immediate;
      Internal( '?comp'         , _qcomp         , '( -- ) // Generate error when not compiling');
      Internal( 'compiling?'    , _compilingq    , '( -- n ) // Return true when compiling');
      Internal( 'compile'       , _compile       , '( -- ) // Compile next inline token, also see [[[Compile]]]');
      Internal( '[compile]'     , _compile       , '( -- ) // Compile next inline token - immediate, also see [[Compile]]');
        Immediate;
      Internal( 'postpone'      , _compile       , '( -- ) // Compile next inline token, also see [[[Compile]]] and [[Compile]]');
        Immediate;
      Internal( 'alias'         , _alias         , '( -- ) // defines an alias for a symbol');
      Internal( 'is'            , _is            , '( <exec-token> -- ) // resolve a [[defer]]red XXXX');
      Internal( 'implements'    , _implements    , '( -- ) // <definition> [[implements]] XXXX');
      Internal( 'execute'       , _execute       , '( <exec-token> -- ) // execute exec-token');
      Internal( 'perform'       , _perform       , '( <var-token> -- ) // : perform @ execute ;');
      Internal( 'evaluate'      , _evaluate      , '( "string" -- ) // evaluate a string as forth');
      Internal( 'explain'       , _explain       , '( <exec-token> -- ) // explain exec-token');
      Internal( 'see'           , _see           , '( -- ) // eats the next token and explains it, defined as : see '' explain ;');

      Evaluate( 'only also forth definitions vocabulary math also math definitions');

      Internal( 'int'           , _int           , '( n -- n ) // integer part of n');
      Internal( 'frac'          , _frac          , '( n -- n ) // fractional part of n');
      Internal( '^'             , _power         , '( n1 n2 -- n3 ) // n1 to the power n2');
      Internal( 'exp'           , _exp           , '( n -- n ) // e to the power of n');
      Internal( 'ln'            , _ln            , '( n -- n ) // natural log of n');
      Internal( 'rad2deg'       , _rad2deg       , '( n -- n ) // radians to degrees');
      Internal( 'deg2rad'       , _deg2rad       , '( n -- n ) // degrees to radians');
      Internal( 'sin'           , _sin           , '( n -- n ) // sine of n');
      Internal( 'asin'          , _asin          , '( n -- n ) // arc sine of n');
      Internal( 'cos'           , _cos           , '( n -- n ) // cosine of n');
      Internal( 'acos'          , _acos          , '( n -- n ) // arc cosine of n');
      Internal( 'tan'           , _tan           , '( n -- n ) // tangent of n');
      Internal( 'atan'          , _atan          , '( n -- n ) // arc tangent of n');
      Internal( 'atan2'         , _atan2         , '( n1 n2 -- ) // arc tangent of n1/n2');
      Internal( 'cosh'          , _cosh          , '( n -- n ) // cosine hyperbolicus of n');
      Internal( 'sinh'          , _sinh          , '( n -- n ) // sine hyperbolicus of n');
      Internal( 'tanh'          , _tanh          , '( n -- n ) // tangent hyperbolicis of n');
      Internal( 'atanh'         , _atanh         , '( n -- n ) // arc taangent hyperbolicus of n');
      Internal( 'sech'          , _sech          , '( n -- n ) //');
      Internal( 'csh'           , _csch          , '( n -- n ) //');
      Internal( 'acot'          , _acot          , '( n -- n ) // arc cotangent of n');
      Internal( 'asec'          , _asec          , '( n -- n ) //');
      Internal( 'acsc'          , _acsc          , '( n -- n ) //');
      Internal( 'acosh'         , _acosh         , '( n -- n ) // arc cosine hyperbolicus of n');
      Internal( 'asinh'         , _asinh         , '( n -- n ) // arc sine hyperbolicus of n');
      Internal( 'acoth'         , _acoth         , '( n -- n ) // arc cotangent hyperbolicus of n');
      Internal( 'asech'         , _asech         , '( n -- n ) //');
      Internal( 'acsch'         , _acsch         , '( n -- n ) //');

      Evaluate( 'only also forth definitions vocabulary lists also lists definitions');

      Internal( '>list'         , _tolist        , '( <exec-token> -- {list} ) // convert execution token to a quotation');
      Internal( 'call'          , _call          , '( {list} -- ) // interpret the list as code, like the joy i combinator, or the factor call combinator');
      Internal( 'stack'         , _stack         , '( x y z -- x y z {xyz} ) // Push the current stack as a list on top of itself');
      Internal( 'unstack'       , _unstack       , '( {list}or"string" -- n1 ... nn )');
      Internal( 'swons'         , _swons         , '( {list}or"string" n -- {list}or"string" ) // Append n to list');
      Internal( 'first'         , _first         , '( {list}or"string" -- n ) // get head of list');
      Internal( 'rest'          , _rest          , '( {list}or"string" -- {list}or"string" ) // remove first element from list');
      Internal( 'concat'        , _concat        , '( {list}or"string" -- {list}or"string" ) // concatenate two lists');
      Internal( 'at'            , _at            , '( {list}or"string" index -- item ) // Get list or string item at index');
      Internal( 'count'         , _count         , '( {list}or"string" -- count ) // Get string length or nr of items in a list');
      Internal( 'map'           , _map           , '( {list} {quotation} -- {list} )');
      Internal( 'step'          , _step          , '( {list} {quotation} -- )');
      Internal( 'filter'        , _filter        , '( {list} {quotation} -- {list} )');
      Internal( 'reduce'        , _reduce        , '( {list} n {quotation} -- n )');
      Internal( 'explode'       , _explode       , '( "string" "splitter" -- {list} )');
      Internal( 'implode'       , _implode       , '( {list} "joiner" -- "string" )');

      Evaluate( 'only also forth definitions');
    end;

    procedure   TKnobsForth.Crash;
    begin
      if Assigned( FLastDefined)
      then begin
        with FLastDefined
        do Crash := not Crash;
      end;
    end;

    procedure   TKnobsForth.Immediate;
    begin
      if Assigned( FLastDefined)
      then begin
        with FLastDefined
        do Immediate := not Immediate;
      end;
    end;

    procedure   TKnobsForth.Hide;
    begin
      if Assigned( FLastDefined)
      then FLastDefined.Hide;
    end;

    procedure   TKnobsForth.Reveal;
    begin
      if Assigned( FLastDefined)
      then FLastDefined.Reveal;
    end;

    procedure   TKnobsForth.SysSymbol;
    begin
      if Assigned( FLastDefined)
      then begin
        with FLastDefined
        do SysSymbol := not SysSymbol;
      end;
    end;

    procedure   TKnobsForth.ForgetAll;
    begin
      while Assigned( FLastDefined)
      do FLastDefined.ReleaseRef;
    end;

    procedure   TKnobsForth.CheckControlStructure( aSeen: Cardinal);
    var
      anExpected : Cardinal;
      SExpected  : string;
      SSeen      : string;
      SIntro     : string;
    begin
      anExpected := ControlPop;
      if aSeen <> anExpected
      then begin
        if anExpected in [ Low( ControlNames) .. High( ControlNames)]
        then begin
          SExpected := ControlNames [ anExpected];
          SIntro    := ControlIntros[ anExpected];
        end
        else begin
          SExpected := Format( '<unkown %d>', [ anExpected], AppLocale);
          SIntro    := '<unknown>';
        end;
        if aSeen in [ Low( ControlNames) .. High( ControlNames)]
        then SSeen := ControlNames[ aSeen]
        else SSeen := Format( '<unknown %d>', [ aSeen], AppLocale);
        RaiseErrorFmt(
          'control structure error, expected "%s" but saw "%s", the structure was started with "%s"',
          [
            SExpected,
            SSeen,
            SIntro
          ]
        );
      end;
    end;

    procedure   TKnobsForth.CheckDataStack( const aMsg: string);
    begin
      if FDataStack.Depth <> 0
      then RaiseErrorFmt( 'Data stack not empty (%s)', [ aMsg]);
    end;

    procedure   TKnobsForth.CheckControlStack( const aMsg: string);
    begin
      if FControlStack.Depth <> 0
      then RaiseErrorFmt( 'Control stack not empty (%s)', [ aMsg]);
    end;

    procedure   TKnobsForth.CheckReturnStack( const aMsg: string);
    begin
      if FReturnStack.Depth <> 0
      then RaiseErrorFmt( 'Return stack not empty (%s)', [ aMsg]);
    end;

    function    TKnobsForth.Anonymous: string;
    var
      i : Integer;
      W : TSymbol;
    begin
      Result := '{ ';
      for i := 0 to FAnonymous.Count - 1
      do begin
        W :=FAnonymous[ i] as TSymbol;
        Result := Result + Format( '<%d> %s ', [ W.RefCount, W.See], AppLocale);
      end;
      Result := Result + '}';
    end;

    function    TKnobsForth.VocDump: string;
    // Make a vocabulary based dump
    var
      aStringList : TStringList;
      i           : Integer;
    begin
      aStringList := TStringList.Create;
      try
        aStringList.Add( 'vocs');
        for i := 0 to Length( FVocs) - 1
        do FVocs[ i].Dump( aStringList, 1);
        aStringList.Add( 'end');
        for i := 0 to aStringList.Count - 1
        do Say( aStringList[ i]);
      finally
        aStringList.Free;
      end;
    end;

    function    TKnobsForth.CodeDump: string;
    // Make a linear code dump
    var
      aStringList : TStringList;
      i           : Integer;
    begin
      aStringList := TStringList.Create;
      try
        aStringList.Add( 'code');
        FCode.Dump( aStringList, 1);
        aStringList.Add( 'end');
        for i := 0 to aStringList.Count - 1
        do Say( aStringList[ i]);
      finally
        aStringList.Free;
      end;
    end;

    procedure   TKnobsForth.AddFile( const aFileName: string);
    begin
      SetLength( FFiles, Length( FFiles) + 1);
      FFiles[ Length( FFiles) - 1] := aFileName;
    end;

    procedure   TKnobsForth.ClearFiles;
    begin
      SetLength( FFiles, 0);
    end;

    function    TKnobsForth.Files: TForthList;
    var
      i : Integer;
    begin
      Result := TForthList.Create( Self);
      for i := 0 to Length( FFiles) - 1
      do Result.CompileVariant( FFiles[ i]);
    end;

    function    TKnobsForth.IsIncluded( const aFileName: string): Boolean;
    var
      i : Integer;
    begin
      Result := False;
      for i := 0 to Length( FFiles) - 1
      do begin
        if SameText( aFileName, FFiles[ i])
        then begin
          Result := True;
          Break;
        end;
      end;
    end;

    procedure   TKnobsForth.Required( const aFileName: string);
    begin
      if not IsIncluded( aFileName)
      then PushFile( ExpandFileName( aFileName));
    end;

    function    TKnobsForth.CfaToSymbol( aCFA: TCFA): TSymbol;
    var
      i : Integer;
    begin
      Result := nil;
      for i := 0 to Length( FVocs) - 1
      do begin
        Result := FVocs[ i].CfaToSymbol( aCFA);
        if Assigned( Result)
        then Break;
      end;
    end;

    procedure   TKnobsForth.Bye;
    begin
      if Assigned( FOnBye)
      then FOnBye( Self);
    end;

// private
    function    TKnobsForth.GetSourceName: string;
    begin
      if Assigned( FFileStack)
      then Result := FFileStack.FileName
      else Result := 'unknown';
    end;

    function    TKnobsForth.GetSourceLineNr: Integer;
    begin
      if Assigned( FFileStack)
      then Result := FFileStack.LineNr
      else Result := 0;
    end;

    function    TKnobsForth.GetSourceOffset: Integer;
    begin
      if Assigned( FFileStack)
      then Result := FFileStack.Offset
      else Result := 0;
    end;

    function    TKnobsForth.GetContext: TVocabulary;
    begin
      if Length( FContext) > 0
      then Result := FContext[ Length( FContext) - 1]
      else Result := nil;
    end;

    procedure   TKnobsForth.SetContext( aValue: TVocabulary);
    begin
      if Length( FContext) > 0
      then FContext[ Length( FContext) - 1] := aValue;
    end;

    function    TKnobsForth.AddVoc( aVoc: TVocabulary): TVocabulary;
    begin
      SetLength( FVocs, Length( FVocs) + 1);
      FVocs[ Length( FVocs) - 1] := aVoc;
      Result       := aVoc;
      FinalizeCreation( aVoc);
    end;

    procedure   TKnobsForth.FSave( const aContents, aFileName: string);
    begin
      StringTofile( aContents, aFileName);
    end;

    function    TKnobsForth.FLoad( const aFileName: string): string;
    begin
      Result := FileToString( aFileName);
    end;

    procedure   TKnobsForth.Evaluate( const aText: string);
    var
      S        : string;
      T        : string;
      i        : Cardinal;
      aStrings : TStringList;
    begin
      S := aText + ' ';
      aStrings := TStringList.Create;
      try
        T := '';
        for i := 1 to Length( S)
        do begin
          case S[ i] of
            ^J : begin aStrings.Add( T); T := ''; end;
            ^M : ;
            else T := T + S[ i];
          end;
        end;
        if T <> ''
        then aStrings.Add( T);
        PushLines( aStrings, 'evaluate');
        if not Interpreting
        then Run;
      finally
        aStrings.Free;
      end;
    end;

// public

    procedure   TKnobsForth.DoLine( aSender: TObject; aLineNr, aTotalCount: Integer);
    begin
      if Assigned( FOnLine)
      then FOnLine( Self, aLineNr, aTotalCount);
    end;

    procedure   TKnobsForth.DoFileEnter( aSender: TObject; const aFileName: string);
    var
      S : string;
    begin
      if ( Length( aFileName) > 0) and ( aFileName[ 1] <> '+')
      then begin
        if Assigned( FOnFileEnter) and ( aFileName <> '')
        then FOnFileEnter( Self, aFileName);
        S := CleanupFilePath( ExtractFilePath( aFileName));
        if S <> ''
        then begin
          if not SetCurrentDir( S)
          then RaiseErrorFmt( 'Can''t change directory to "%s"', [ S]);
        end;
      end;
    end;

    procedure   TKnobsForth.DoFileLeave( aSender: TObject; const aFileName: string);
    var
      S : string;
    begin
      if ( aFileName <> '') and ( aFileName[ 1] <> '+')
      then begin
        if Assigned( FOnFileLeave)
        then FOnFileLeave( Self, aFileName);
        S := CleanupFilePath( ExtractFilePath( aFileName));
        if S <> ''
        then begin
          if not SetCurrentDir( S)
          then RaiseErrorFmt( 'Can''t change directory to "%s"', [ S]);
        end;
      end;
    end;

    procedure   TKnobsForth.DoFileResume( aSender: TObject; const aFileName: string);
    var
      S : string;
    begin
      if ( aFileName <> '') and ( aFileName[ 1] <> '+')
      then begin
        if Assigned( FOnFileResume)
        then FOnFileResume( Self, aFileName);
        S := CleanupFilePath( ExtractFilePath( aFileName));
        if S <> ''
        then begin
          if not SetCurrentDir( S)
          then RaiseErrorFmt( 'Can''t change directory to "%s"', [ S]);
        end;
      end;
    end;

//  public

    constructor TKnobsForth.Create( anOwner: TComponent);
    begin
      inherited;
      FAnonymous    := TObjectList.Create;
      FAnonymous.OwnsObjects := False;
      FDataStack    := TStack.Create( Self, 'data'   ); FDataStack   .OnUnderflow := DoStackUnderflow;
      FReturnStack  := TStack.Create( Self, 'return' ); FReturnStack .OnUnderflow := DoStackUnderflow;
      FControlStack := TStack.Create( Self, 'control'); FControlStack.OnUnderflow := DoStackUnderflow;
      FCode         := TForthList .Create( Self, 'code');
      FFileStack    := TFileStack .Create;
      with FFileStack
      do begin
        OnLine       := Doline;
        OnFileEnter  := DoFileEnter;
        OnFileLeave  := DoFileLEave;
        OnFileResume := DoFileResume;
      end;
      // do NOT call : Reset here
    end;

    destructor  TKnobsForth.Destroy; // override;
    begin
      FDataStack   .OnUnderflow := nil;
      FReturnStack .OnUnderflow := nil;
      FControlStack.OnUnderflow := nil;
      with FFileStack
      do begin
        OnLine       := nil;
        OnFileEnter  := nil;
        OnFileLeave  := nil;
        OnFileResume := nil;
      end;
      ForgetAll;                              // Forget everything ..
      FCode        .ReleaseRef;
      FControlStack.ReleaseRef;
      FReturnStack .ReleaseRef;
      FDataStack   .ReleaseRef;
      FreeAndNil( FFileStack);
      FreeAndNil( FAnonymous);
      inherited;
    end;

    procedure   TKnobsForth.Reset;
    begin
      ClearScreen;
      FInterpreting :=  0;
      FIP           := -1;
      FDataStack   .Clear;
      FReturnStack .Clear;
      FControlStack.Clear;
      ForgetAll;                              // Forget everything ..
      FCode        .Clear;
      FAnonymous   .Clear;

      ClearScales;
      SetupInternals;
      FFileStack.Reset;
      if Assigned( FOnReset)
      then FOnReset( Self);
      ClearFiles;
      if FileExists( StartupFile)
      then begin
        PushFile( ExpandFileName( StartupFile));
        if not Interpreting
        then Run;
        FFileStack.Reset;
      end;
    end;

    procedure   TKnobsForth.DoBreak;
    begin
      RaiseBreak( 'user break');
    end;

    procedure   TKnobsForth.ClearScreen;
    begin
      if Assigned( FOnClearScreen)
      then FOnClearScreen( Self);
    end;

    procedure   TKnobsForth.ClearScales;
    begin
      if Assigned( FOnClearScales)
      then FOnClearScales( Self);
    end;

    procedure   TKnobsForth.CheckStacks( const aMsg: string);
    begin
      CheckReturnStack ( aMsg);
      CheckControlStack( aMsg);
      CheckDataStack   ( aMsg);
    end;

    procedure   TKnobsForth.AcceptCode( const aText: string; const aData: TVariants = nil);
    // For interactive use - this one can use code fragments for evaluation.
    var
      i    : Integer;
      aLen : Integer;
    begin
      if Assigned( aData)
      then begin
        aLen  := Length( aData);
        for i := aLen - 1 downto 0
        do DataPush( aData[ i]);
        DataPush( aLen);
      end;
      Evaluate( aText);
    end;

    procedure   TKnobsForth.AcceptAction( const aWord: string; const aData: TVariants = nil);
    // For automated use - supposed to be called with one word
    var
      i    : Integer;
      aLen : Integer;
      S    : TSymbol;
    begin
      S := FindToken( aWord);
      if Assigned( S)
      then begin
        if Assigned( aData)
        then begin
          aLen  := Length( aData);
          for i := aLen - 1 downto 0
          do DataPush( aData[ i]);
          DataPush( aLen);
        end;

        try
          S.Interpret;
        except
          // Kill exceptions for automated calls - stuff should be tested
          // interactively, exceptions in real timish stuff is a bit of a
          // PITA.
          // Anyway, can always set a breakpoint here on KilledException()
          on E: Exception
          do KilledException( E);
        end;

      end;
    end;

    function    TKnobsForth.LeaseStackImage( const aData: TVariants): TForthList;
    var
      i : Integer;
    begin
      Result := TForthList.Create( Self);
      if Assigned( aData)
      then begin
        for i := 0 to Length( aData)- 1
        do Result.CompileVariant( aData[ i]);
        Result.CompileVariant( Length( aData));
      end;
    end;

    procedure   TKnobsForth.ReleaseStackImage( aFrame: TForthList);
    begin
      if Assigned( aFrame)
      then aFrame.ReleaseRef;
    end;

    function    TKnobsForth.LeaseSymbol( const aName: string): TSymbol;
    begin
      Result := FindToken( aName);
      if Assigned( Result)
      then Result.AddRef;
    end;

    procedure   TKnobsForth.ReleaseSymbol( const aSymbol: TSymbol);
    begin
      if Assigned( aSymbol)
      then aSymbol.ReleaseRef;
    end;

    procedure   TKnobsForth.AcceptWithStackImage( const aSymbol: TSymbol; const aFrame: TForthList);
    begin
      if Assigned( aSymbol) and Assigned( aFrame)
      then begin
        aFrame.Call;
        aSymbol.Interpret;
      end;
    end;



var

  Tmp : Integer;

initialization


finalization

{$ifdef DEBUG}
  Tmp := AllSymbolsCounter;
  if Tmp <> 0
  then Nop;
  Tmp := AnonymousCounter;
  if Tmp <> 0
  then Nop;
{$endif}

end.
